 /* C Source:	     gc_basic_call_model_voice_video_3g_cnf_nbup.c
 * Description:   This is a demonstration program for a basic call model which performs,
 *					  Basic make call and DropCall from Outbound end.It also shows the case 
 *					  of simultaneous disconnect.
 *					  which executes the program continuously until it encounters signal ctrl+c.
 * General Notes: 1. The tab setting used for this file is 3. 
 *					  2. Pagination is done with Courier New, 10 point, 64 lines per page
 *					  3. For purposes of the demo, the outbound side waits 
 *						  1-2 seconds in the connected state, then calls gc_DropCall()
 *					  4. PDK as shipped has an issue where occasionally the first makecall may
 *						  fail with an invalid state error. The demo program has a recovery
 *						  mechanism built in. When connected to a switch, this should not be 
 *						  a problem.
 *					  5. gc_demo_open_isdn_channel may need to be modified for your environment,
 *						  especially if you are using T1.

 *
 *	%created_by:	fiedorom %
 *	%date_created: Tue May 25 11:21:03 2004 %
 * %modified by: SKS %
 *	%date_modified: Mon Aug 17 17:35:21 2009 %
 *
 ***************************************************************************/
/**********************************************************************
 *  Copyright (C) 2000-2002 Intel Corporation.
 *  All Rights Reserved
 *
 *  All names, products, and services mentioned herein are the 
 *  trademarks or registered trademarks of their respective 
 *  organizations and are the sole property of their respective owners.
 **********************************************************************/
#ifndef lint
static char *_csrc = "@(#) %filespec: gc_basic_call_model_voice_video_3g_cnf_nbup.c-28 %  (%full_filespec: gc_basic_call_model_voice_video_3g_cnf_nbup.c-28:csrc:gc#1 %)";
#endif

/* OS Header Files */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/timeb.h>
#include <errno.h>
#include <curses.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/select.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <stropts.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>



/* OS Socket API Headers */
#ifdef _WIN32
#include <winsock2.h>
#else
#include <unistd.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#endif

/* Dialogic Header Files */
#include <gcip.h>
#include <gclib.h>
#include <gcisdn.h>
#include <srllib.h>
// SKS
#include <cc_s7.h>
#include <dxxxlib.h>
#include <ipmlib.h>
#include <gcipmlib.h>
#include <pthread.h>
#include <fcntl.h>

/* Macro Definitions */
#define MAXCHAN			120		/* Maximum number of channels that can be opened - 2 E1 spans	*/
#define MAX_DEVNAME		64		/* Maximum length of device name											*/
#define MAX_CODECNAME	20		/* Maximum length of codec name											*/
#define MAX_STRING_SIZE 1000	/* Maximum length of string that to write on log file			*/

#ifdef USE_VIDEO

#include <m3glib.h>
#include <m3gevts.h>
#include <m3gerrs.h>
#include <port_connect.h>
#include <mmlib.h>
#include <cnflib.h>
#include <lb_mtklib.h>


#define MAX_REGION_COUNT      9
#define MAX_CONFERENCE_COUNT  9


#define ROUTE		1
#define UNROUTE	0
#define START_PLAY 2

#define AUDIO_LCN 			1
#define VIDEO_LCN 			2
#define AUDIO_VIDEO_LCN 	3

static int g_bDtmfSend=true;
static const char* portMediaTypeStr[] =
{
				"NONE",
				"AUDIO",
				"VIDEO",
				"H223/NBUP"
};

	typedef struct
	{
		char device_name[MAX_DEVNAME];
		SRL_DEVICE_HANDLE devH;
		DM_PORT_INFO               audioTxPortInfo;
  	DM_PORT_INFO               audioRxPortInfo;
  	DM_PORT_INFO               videoTxPortInfo;
  	DM_PORT_INFO               videoRxPortInfo;
		DM_PORT_INFO							h223TxPortInfo;
		DM_PORT_INFO							h223RxPortInfo;
	}DeviceInfo;

enum audio_video_codec
{
				PCMA							= 0,
				PCMU							= 1,
				G711U_5MS					=	2,
				G711U_20MS				=	3,
				G711A_5MS					=	4,
				G711A_20MS				=	5,
				AMR								= 6,
				H263							= 7,
				MPEG4							= 8,
        UNKNOWN           = 9
};

#else 
	typedef struct
	{
  	char device_name[MAX_DEVNAME];
  	int devH;
	}DeviceInfo;
#endif

// Device & API call type
// Do not change the order/value of device types (0-5)
#define M3G					0
#define M3GAUD			1
#define M3GVID			2
#define MM					3
#define IPM					4
#define CNFPTY			5

// apicall/event type
#define DM					6
#define CNF					7
#define CNFEV				8
#define M3GEV				9
#define MMEV				10
#define IPMEV				11
#define DMEV				12

/* Logging defines */
#define DX							13
#define MISC						14	/* None of the above															*/
#define EVENT						15	/* An event																		*/
#define STATE						16	/* STATE message																*/
#define GC_APIERR				17	/* Had error on GC API function											*/
#define GC_APICALL			18	/* GC API funtion call														*/
#define MISC_WITH_NL		19	/* None of the above with new line codes								*/
#define NON_GC_APIERR		20	/* Had error on non-GC API function										*/
#define GC_RESULT_INFO	21	/* want GC result information												*/

static const char *devEvtTypeStr[] =
{
	"M3G",
	"M3GAUD",
	"M3GVID",
	"MM",
	"IPM",
	"CNFPTY",
	"DM",
	"CNF",
	"CNFEV",
	"M3GEV",
	"MMEV",
	"IPMEV",
	"DMEV",
	"DX",
	"MISC",
	"EVENT",
	"STATE",
	"GC_APIERR",
	"GC_APICALL",
	"MISC_WITH_NL",
	"NON_GC_APIERR",
	"GC_RESULT_INFO",
	"TYPE_NOT_DEFINED"
};
/* Keep large																	*/
#define MAX_CALLS			2		/* Maximum number of calls per channel */
#define MIN_CONNECTED_TIME	1 /* Minimum connected time, artifact of test program				*/
/* and most likely will not be needed for a production program */

#define	MIN_MAKECALL_RETRY_TIME	5	/* min time to wait before retrying makecall under		*/
/* error conditions															*/
#define MAXCODECS			10	/* Maximum number of codecs per VoIP channel */
#define MAX_VADSTR		10	/* Maximum length of VAD string, e.g. "ON", "OFF", "N/A" */


/* Inbound or Outbound Direction of device */
#define DIR_IN				1
#define DIR_OUT			2
/* Miscellaneous declarations */

#define YES					1
#define NO					0


// SKS
#define BOARD_DEVICE  999
FILE					*boardLogFd;			
#define ALL_DEVICES	  -1		/* This value is used to write the */
/* miscellaneous statements into all log files */ 
#define OUT_OF_RANGE	  -1		/* callindex is out of range */

/*
 * Technology Types
 */
#define E1ORT1_CAS		1
#define ISDN					2							/* includes both E1 and T1 ISDN */
#define ANALOG				3
#define SS7						4
#define H323					5
#define SIP						6
#define BICC					7


#define MAX_QUEUE_SIZE 64
#define CALL_TYPE_VIDEO 1
#define CALL_TYPE_AUDIO 0

/* Global variables */
static char		logfile[] = "gc_basic_call_model_voice_video_3g_cnf_nbup";		/* log filename without the .log	 */
static char		cfgfile[] = "gc_basic_call_model_voice_video_3g_cnf_nbup.cfg"; /* config filename */

static int		num_devices;		/* Number of devices loaded from cfg file */
static int		num_sip_devices;		/* Number of devices loaded from cfg file */
static int		num_bicc_devices;		/* Number of devices loaded from cfg file */
static int		max_ip_channel_to_start;
static int		interrupted = NO;	/* Flag for user interrupted signals		 */
static int		reset_linedevs_left_to_complete = 0; 


/* Data structure which stores call information for each line device */
typedef struct {
				CRN					crn;							/* GlobalCall call handle */
				int					call_state;					/* state of first layer state machine */
				int					dropcall_active;			/* Flag for simultaneous disconnect check	*/
} CALL;

/* TX and RX codecs for each line device although only used for H.323 or SIP */
typedef struct {
				char					txcodec[MAX_CODECNAME];	/* TX codec type (H.323 or SIP only)					*/
				int					txrate;						/* TX codec rate (H.323 or SIP only)					*/
				char					txVAD[MAX_VADSTR];		/* TX VAD enabled/disabled (H.323 or SIP only)		*/
				char					rxcodec[MAX_CODECNAME];	/* RX codec type (H.323 or SIP only)					*/
				int					rxrate;						/* RX codec rate (H.323 or SIP only)					*/
				char					rxVAD[MAX_VADSTR];		/* RX VAD enabled/disabled (H.323 or SIP only)		*/
} CODEC;

/* Data structure which stores all information for each line device */
static struct channel {
#ifdef USE_VIDEO
				M3G_SIMULTANEOUS_CAP_SET simultaneousCapabilityList;

   	 		M3G_CAPS_LIST  preferredH223Caps;
   			M3G_CAPS_LIST    preferredAudioCaps;
   			M3G_CAPS_LIST    preferredVideoCaps;


				M3G_CAPS_LIST h223CapsList;
				M3G_CAPS_LIST audioCapsList;
				M3G_CAPS_LIST videoCapsList;
				M3G_BOOL 	bVideoCall;
				bool bPartyIntoCnf;
				int remoteAudioLogicalChannel;
				int remoteVideoLogicalChannel;
				bool bPlayActive;
				bool bRecordActive;
				bool bMediaStarted;
				bool bIpmMediaStarted;
				int txPortCtr;
				int rxPortCtr;
				int localCapCtr;
				long int olcEvtMask;
				int modifyMediaCtr;
				int modifyMediaIdleCtr;

				int dmevPortDisconnectCtr;
				int dmevPortConnectCtr;
				int dmevConnectCtr;
				int dmevDisconnectCtr;

				int openAudVidOlcCmplt;
				int respondAudVidOlcCmplt;

				bool bDxAndDtiConnected;
				bool bM3gAndDtiConnected;
				bool bM3gAndMmConnected;
				bool bM3gAndCnfPartyConnected;
				bool bIpmAndMmConnected;
				bool bIpmAndCnfPartyConnected;
				bool bIpmAndM3gConnected;
				bool bH245Started;
				bool isCallDisconnectInProgress;


//				DeviceInfo deviceInfo[7];	// SKS index 0=m3g, 1=m3gAud, 2=m3gVid, 3=mm, 4=ipm and 5=cnfpty
				IPM_MEDIA_INFO      ipmLocalRemoteMediaInfo;
				IPM_VIDEO_CODER_INFO_EX localVideoInfoEx;

				char		biccApmMessage[MAX_STRING_SIZE];
				char		biccIamMessage[MAX_STRING_SIZE];
				
				char                remoteEPIpAddress[MAX_DEVNAME];
				char                localEPIpAddress[MAX_DEVNAME];
				int                 remoteEPAudioPortNum;
				int                 localEPAudioPortNum;
				int                 remoteEPVideoPortNum;
				int                 localEPVideoPortNum;
				int                 localEPNbUPPortNum;
				int                 remoteEPNbUPPortNum;
				int		remoteEPVideoPayloadID;
				int		remoteEPAudioPayloadID;
				int 	remoteEPAmrModeSet;
				int		remoteEPNbUPPayloadID;

				int	 videoCodec;
				int  audioCodec;

				char                sdpText[MAX_STRING_SIZE];

#endif
				DeviceInfo deviceInfo[7];	// SKS index 0=m3g, 1=m3gAud, 2=m3gVid, 3=mm, 4=ipm and 5=cnfpty

				int callType;
				bool bBusy;

				int 			networkH;
				int			dxH;
				LINEDEV				ldev;							/* GlobalCall line device handle							*/
				LINEDEV				mediah;						/* Media handle - only for H323                    */
				CALL					call[MAX_CALLS];
				int					index;						/* Device index as loaded from cfg file				*/
				int					direction;					/* Inbound(WaitCall will be issued)
																					 or outbound(MakeCall will be issued)				*/
				int					blocked;						/* YES or NO													*/
				char					netdevice[MAX_DEVNAME]; /* Network device name										*/
				char					protname[MAX_DEVNAME];	/* Protocol name												*/
				char					dxDeviceName[MAX_DEVNAME]; /* Voice device name										*/ 
				char					devname[MAX_DEVNAME];	/* Argument to gc_OpenEx() function						*/
				char					destination_num[GC_ADDRSIZE]; /* Phone													*/
				int					num_codecs;					/* The No. of codecs specified (VoIP device only) 	*/
				CODEC					codec[MAXCODECS];			/* Codecs (VoIP device only)								*/
				int					numb_calls;					/* The No of calls on the device							*/
				int					waitcall_active;			/* Flag for waitcall function for issuing only once*/
				int					makecall_active;
				int					resetlinedev_active;		/* Necessary since must ignore call related GC		*/
				/* events with gc_ResetLineDev() active				*/
				int					resetlinedev_required_after_unblocked; /* for ISDN									*/
				/* and some error conditions								*/

				int					makecall_timeout;			/* necessary since not all technologies support		*/
				/* async makecall timeout and they might be			*/
				/* different for different technologies				*/
				GC_MAKECALL_BLK	makecallblk;				/* Variable for MAKECALL block							*/
				GC_PARM_BLKP gcParmBlk;
				FILE					*log_fp;						/* logfile Pointer											*/
				int					techtype;					/* technology type: E1ORT1_CAS, ISDN, ANALOG,		*/
				int			bSdpReceived;

				time_t				makecall_time;				/* time before which makecall is not allowed       */
				/* this is to add robustness to the app				*/
				/* if makecall fails which it might under some		*/
				/* circumstances, but will work if try later			*/
				/* if this value is 0, then OK to do now           */

				/* this variable is in so know when to drop a call */
				/* it is an artifact of the demo program and 	   */
				/* and should not be used in production level 		*/
				/* programs														*/
				time_t				dropcall_time;				/* time to drop the call on the outbound side      */
				/* format is result of time() function call        */
				/* 0 <=> no dropcall required                      */
				time_t				intercall_time;			/* Time when the call was released. The next call  */
				/* would be made after intercall_time +            */
				/* intercall_delay seconds.                        */
				int               intercall_delay;        /* Inter-call delay in milliseconds                */
} port[MAXCHAN];

/* codec translation table */
typedef struct {
				char*			codecname;
				int			codecvalue;
				int			isVADsupported;
} CODECXLAT;

static const CODECXLAT codecxlat[] = {
				{"G.711Alaw",	GCCAP_AUDIO_g711Alaw64k,			NO},
				{"G.711ulaw",	GCCAP_AUDIO_g711Ulaw64k,			NO},
				{"G.723_5.3k",	GCCAP_AUDIO_g7231_5_3k,				YES},
				{"G.723_6.3k",	GCCAP_AUDIO_g7231_6_3k,				YES},
				{"G.726_16k",	GCCAP_AUDIO_g726_16k,				YES},
				{"G.726_24k",	GCCAP_AUDIO_g726_24k,				YES},
				{"G.726_32k",	GCCAP_AUDIO_g726_32k,				YES},
				{"G.726_40k",	GCCAP_AUDIO_g726_40k,				YES},
				{"G.729",		GCCAP_AUDIO_g729		,				NO},
				{"G.729A",		GCCAP_AUDIO_g729AnnexA,				NO},
				{"G.729B",		GCCAP_AUDIO_g729wAnnexB,			YES},
				{"G.729AB",		GCCAP_AUDIO_g729AnnexAwAnnexB,	YES},
				{"GSM_FULL",	GCCAP_AUDIO_gsmFullRate,			YES},
				{"T.38UDP",		GCCAP_DATA_t38UDPFax,				NO},
				{"DONT_CARE",	GCCAP_dontCare,						NO}
};

static int		in_calls, out_calls;					/* Counters for Inbound calls and Outbound calls	*/
static int		gc_started = NO;						/* Flag is set if gc_Start() is Success				*/
time_t			start_time;								/* To note down the start time for the test			*/

/* Function prototype(s) */

void  *createEventHandlerThread(void *context);
static pthread_t  gEventHandlerThreadID;
static void dxPlayAudioPrompt(int index);

static void getNetworkHandle(int index);
#ifndef USE_VIDEO
static void getIpmHandle(int index);
#endif
static void openMediaDevice(int index);
static void unRouteDtiAndDx(int index);
static void routeDtiAndDx(int index);
static void routeIpmAndDx(int index);
static void unRouteIpmAndDx(int index);

static void		init_srl_mode(void);							/* inits SRL mode */
static void		load_config_file(void);						/* Loads Configuration File									*/
static void		drop_outbound_calls_if_required(void); /* used by demo to make repeated calls */
static void		make_calls_if_required(void);				/* used by demo to handle makecall error conditions 	*/
static void		printandlog(int index, int log_type,	/* Prints and logs												*/ 
								METAEVENT *metavent, char *msg_str, int callindex);
static void		gc_demo_makecall(int index);				/* Function for gc_MakeCall()									*/ 
static void		exitdemo(int exitlevel);					/* Exit routine													*/
static void		initiate_exit(void);
static void		intr_hdlr(void);								/* For handling user interrupted signals					*/
int 					_kbhit();
static void		set_calling_num(struct channel *pline, char *phone_num);
static void		enable_alarm_notification(struct channel *pline);/* enable alarm notification                */
static void		gc_demo_open_isdn_channel(int index);	/* Technology specific initialization for ISDN			*/
static void		gc_demo_open_E1orT1_channel(int index);/* Technology specific initialization for PDK			*/ 
static void		gc_demo_open_ss7_channel(int index);	/* Technology specific initialization for SS7			*/
static void		gc_demo_open_analog_channel(int index);/* Technology specific initialization for Analog		*/
static void		gc_demo_open_H323_channel(int index);	/* Technology specific initialization for H323  		*/
static void		gc_demo_open_SIP_channel(int index);	/* Technology specific initialization for SIP   		*/
static void		gc_set_channel_codecs(int index);		/* VoIP specific initialization of codecs					*/
static void		open_all_devices(void);						/* Function invokes Technology specific initilization 
																									 functions to open all devices								*/
static void		open_device(int index);						/* Function invokes Technology specific initilization 
																									 functions to open specific device						*/
static void		inbound_application(struct channel *pline); /* PLACE HOLDER FOR YOUR APPLICATION CODE			*/
static void		outbound_application(struct channel *pline); /* PLACE HOLDER FOR YOUR APPLICATION CODE			*/
static void		print_alarm_info(METAEVENTP metaeventp, struct channel *pline);
static void		process_event(void);							/* Function handles the GlobalCall Events					*/
static void		process_connected_event(struct channel *pline, METAEVENT *metaeventp);
static void		process_disconnected_event(struct channel *pline, METAEVENT *metaeventp);
static void		process_offered_event(METAEVENT *metaeventp, struct channel *pline);
static void		process_call_related_inbound_event(struct channel *pline, METAEVENT *metaeventp); /* Function handles the GlobalCall Events for inbound */
static void		process_call_related_outbound_event(struct channel *pline, METAEVENT *metaeventp); /* Function handles the GlobalCall Events for Outbound */
static void 	process_extension_event(struct channel *pline, METAEVENT *metaeventp);
static int		open_all_logfiles(void);					/* Function to open the logfiles								*/
static void		append_call_state_name(char *buffer, int index, int callindex);	/* append call state name to buffer	*/
static void		format_error_info(char *dest_buffer);		/* formats error information									*/
static void		format_result_info(int index, METAEVENT *metaeventp, /* formats event result information									*/
								char *dest_buffer);
static int		existCRN(CRN crn, struct channel *pline);	/* Retrieve CRN if available */
static void  	process_non_gc_event(METAEVENT *metaevent);
static void resetVariables(int index);
#ifdef USE_VIDEO

int cnfBrdH;
int cnfH;

LB_FRAME_HANDLE layoutTemplateH;
LB_FRAME_HANDLE cnfLayoutH;
CNF_VIDEO_LAYOUT_INFO cnfLayoutInfo;
LB_FRAME_HANDLE cnfRegionList[MAX_REGION_COUNT];
unsigned int cnfRegionListCount;

eLB_LAYOUT_SIZE cnfLayoutSize;
LB_FRAME_HANDLE templateRegionList[MAX_REGION_COUNT];

CNF_VISIBLE_PARTY_LIST cnfVisiblePartyList;
CNF_VISIBLE_PARTY_INFO cnfVisiblePartyInfo[MAX_REGION_COUNT];

int m3gBoardH;
int m3gBoardParmCtr;
static const char *strCoderType[]={"M3G_E_G7231","M3G_E_GSM_AMR_NB","M3G_E_H263","M3G_E_MPEG4"};
static const char *callDirectionStr[] = { "UNKNOWN", "INBOUND", "OUTBOUND" };
static const char *sigStr[] = { "UNKNOWN", "E1ORT1_CAS", "ISDN", "ANALOG", "SS7", "H323", "SIP", "BICC" };

enum sdp_parse_ret_code
{
				SDP_PARSE_SUCCESS		= 0,
				SDP_OFFERED					=	1,
				SDP_NOT_OFFERED			= 2,
				NO_MATCHING_VIDEO		=	3,	
				NO_MATCHING_AUDIO		= 4,	
				SDP_ALREADY_OFFERED	=	5,
				SDP_ERROR						= -1
};

M3G_BOOL 	bM3gStarted;
M3G_BOOL 	bM3gTraceStarted;
static pthread_t  gDtmfEventHandlerThreadID;
static pthread_t  gMakeCallOnRequestHandlerThreadID;
static pthread_t  gDisconnectCallOnRequestHandlerThreadID;
static void 	startGCLib(enumIPCCLIBMediaMode mediaMode);

void  				*createDtmfSendHandlerThread(void *context);
void 					*createMakeCallOnRequestHandlerThread(void *context);
void 					*createDisconnectCallOnRequestHandlerThread(void *context);
static int 		buildApmMessage(int index);

static void 	printMenu();
static void printChannelStatus();

static void 	sendH245UII(int index);
static void 	sendH245VFU(int index);
static void 	setupAudioCallMediaInformation(int index);
static void 	setupVideoCallMediaInformation(int index);
static void 	setupH223CallMediaInformation(int index);
static void 	initResponseSendIpm(int index);
static void 	initSendIpm(int index);
static void 	startIpmMedia(int index);
static void 	stopIpmMedia (int index);
static void processIpmevStartMedia(int index);
static void processIpmevTelephonyEvent(int index,int dig);
static void processIpmevStopped(int index);
static void processIpmevOpen(int index);
static void 	disconnect_video_call(int index);

static void 	getLocalMediaInfo(int index);
static void 	storeLocalMediaInfo(int index,IPM_MEDIA_INFO *pMediaInfo);
static void 	sendIFrameRequest(int index);
static void 	setDTMFMode(int index);
static void 	enableIpmEvents(int index);

static int 		storeSdpValues(char *sdpVal,int index);
static int 		parseOfferedSdp(METAEVENT *metaeventp,struct channel *pline);
static void 	buildOfferAnswerSdp(int index);

static int 	parseApmMessage(METAEVENT *metaevent,struct channel *pline);
static int 		getCallType(METAEVENT *metaevent,struct channel *pline);

#ifdef USE_CNF
// SKS cnf function defination start here
static void 	cnfRemoveParty(int index);
static void 	cnfAddParty(int index);
static void 	gc_demo_open_cnf_board();
static void 	gc_demo_open_conference();
static void 	SetCnfLayout(eLB_LAYOUT_TYPE layoutType,eLB_LAYOUT_SIZE layoutSize);
static void 	GetCnfLayout();
static void 	gc_demo_open_cnf_party();
static void 	enable_disable_cnf_dev_events(bool bFlag);
static void 	enable_disable_cnf_brd_dev_events(bool bFlag);
static void 	SetPartyAttributes(int index, bool bCoach);
static void 	SetCnfPartyVisible(unsigned int partyCount,SRL_DEVICE_HANDLE *partyList);
static void 	GetCnfPartyList();
static void 	GetCnfVisiblePartyList();
#endif

static char    dirStr[5][15]={"M3G_E_IDLE","M3G_E_TX","M3G_E_RX","M3G_E_TXRX"};
const long int eventsReqiredBefOlc = (M3GEV_MSD_ESTABLISHED+M3GEV_LOCAL_TCS_ACKD+M3GEV_REMOTE_TCS_RCVD);

static void 	printLocalCapability(int index);
static void 	cleanResourcesBeforeExit();
static void 	m3gEnableDisableEvents(int index,int handle);
static void 	stopPlayRecord(int index,bool rec);
static void	 	stopPlayAudioPrompt(int index);
static void   stopMedia(int index);
static void processStopMediaCmplt(int index, SRL_DEVICE_HANDLE devH);
static void processModifyMediaCmplt(int index);
static void processDmevPortDisconnect(int index);
static void processDmevPortConnect(int index);
static void 	mmPlayRecordAudioVideoPrompt(int index,bool rec);
static void 	processInitSendComplete(int index);
static int 		InitM3g(int index, int noOfEndPoints);
static void		stopM3g(int index);
static void		resetM3gDevice(int index,int handle);
static void		gc_demo_open_m3g_channel(int index);	/* Technology specific initialization for ISDN			*/
static void		gc_demo_open_mm_channel(int index);	/* Technology specific initialization for ISDN			*/
static void 	m3gSetParameter(int index,int handle);
static void 	getLocalCapability(int index,int handle);
static void 	storeLocalCapability(int index,int handle,M3G_CAPS_LIST *pDefaultCapsList);
static void 	setLocalCapability(int index);
static void 	connectM3gAndDti(int index);
static void 	disconnectM3gAndDti(int index);
static void 	getTxRxPortInfo(int index,int handle);
static void 	storeTxPortInfo(int index,METAEVENT *metaevent);
static void 	storeRxPortInfo(int index,METAEVENT *metaevent);
static void 	routeUnrouteDevices(int index, int devType1, int devType2, int routeFlag,int xcodeFlag);
static void 	connectDevices(int index, int txDevType, int rxDevType, DM_PORT_CONNECT_INFO_LIST *portConnectInfoList1);
static void 	disconnectDevices(int index, int txDevType, int rxDevType, DM_PORT_CONNECT_INFO_LIST *portConnectInfoList1);
static int 		connectDevicesForMakeCall(METAEVENT *metaeventp,struct channel *pline);
static void 	printPortInfoList(int index,DM_PORT_INFO_LIST* pPortInfoList);
//static void printPortInfo(int index,DM_PORT_INFO* pPortInfo,char str[5]);
static void 	printPortInfo(int index,int devType);
static void 	startH245(int index);
static void 	processStopH245Complete(int index);
static void 	stopH245(int index);
static void 	startM3gTrace(int index,int handle);
static void 	stopM3gTrace(int index,int handle);
static void 	openM3gMmChannel();
static void 	getMatchedCapabilityAndOpenLC(int index);
static void 	modifyMedia(int index,M3G_E_DIRECTION direction);
static void 	getMatchedCapability(int index,int handle,M3G_CAPS_LIST *capList);
static void 	sendOpenLC(int index);
static void 	sendOLCAck(int index,M3G_REMOTE_OLC_REQ* remoteOlcReq);
static void 	initiate_exit_video_call();
#endif
/********************************************************************************
 *			NAME: main ()
 *DESCRIPTION: This is the entry point. 
 *					It loads the configuration file.
 *					Take cares the	 Event Handling process.
 *					Starts GC Library.
 *					Opens Configured Devices.
 *					Waits for user interrupted signal (ctrl+c).
 *					Whenever an event is received on any of the opened 
 *					devices, Depending upon the event received the proper 
 *					action will be taken.
 *		  INPUT: None
 *		RETURNS: 0 if success else -1
 ********************************************************************************/
int main()
{
				int						i;
				char						str[MAX_STRING_SIZE], str1[MAX_STRING_SIZE];
				GC_CCLIB_STATUSALL	cclib_status_all;

				/* Note down the start time for the test */
				time(&start_time);

				/* Give control to Event Handler */ 
#ifdef _WIN32
				signal(SIGINT, (void (__cdecl*)(int)) intr_hdlr);
				signal(SIGTERM, (void (__cdecl*)(int)) intr_hdlr);
#else
				signal(SIGHUP, (void (*)()) intr_hdlr);
				signal(SIGQUIT, (void (*)()) intr_hdlr);
				signal(SIGINT, (void (*)()) intr_hdlr);
				signal(SIGTERM, (void (*)()) intr_hdlr);
#endif

#ifdef USE_VIDEO
				printf("\t\t\t\t********** Using Video Mode ***********\n");
#endif
				init_srl_mode();									/* set SRL mode to ASYNC, polled */

				/* Load Configuration file */
				load_config_file();

				/* Function to open all the logfiles */
				open_all_logfiles();
				if (num_devices == 0) {
								printf("No devices enabled - please modify gc_basic_call_model_voice_video_3g_cnf_nbup.cfg to enable desired technology\n");
								exitdemo(1);
				}
				printandlog(ALL_DEVICES, MISC, NULL, "********** GC DEMO - BASIC CALL MODEL ***********\n", 0);


#ifdef USE_VIDEO
				bM3gTraceStarted=M3G_FALSE;
				bM3gStarted=M3G_FALSE;
				startGCLib(MEDIA_OPERATIONAL_MODE_3PCC);
#else 
				/* Start GlobalCall */
				if (gc_Start(NULL) != GC_SUCCESS) { 
								sprintf(str, "gc_Start(startp = NULL) Failed");
								printandlog(ALL_DEVICES, GC_APIERR, NULL, str, 0);
								exitdemo(1);
				}
#endif
				/* Set the flag gc_started to YES */
				gc_started = YES;
				sprintf(str, "gc_Start(startp = NULL) Success");
				printandlog(ALL_DEVICES, GC_APICALL, NULL, str, 0);

				/* this code is usefule for installation problem debugging */
				/* you may not wish to use this code in production systems */
				if (gc_CCLibStatusEx("GC_ALL_LIB", &cclib_status_all) != GC_SUCCESS) {
								sprintf(str, "gc_CCLibStatusEx(GC_ALL_LIB, &cclib_status_all) Failed");
								printandlog(ALL_DEVICES, GC_APIERR, NULL, str, 0);
								exitdemo(1);
				}	
				strcpy(str, " Call Control Library Status:\n");
				for (i = 0; i < GC_TOTAL_CCLIBS; i++) {
								switch (cclib_status_all.cclib_state[i].state) {
												case GC_CCLIB_CONFIGURED:
																sprintf(str1, "		  %s - configured\n", cclib_status_all.cclib_state[i].name);
																break;

												case GC_CCLIB_AVAILABLE:
																sprintf(str1, "		  %s - available\n", cclib_status_all.cclib_state[i].name);
																break;

												case GC_CCLIB_FAILED:
																sprintf(str1, "		  %s - is not available for use\n", cclib_status_all.cclib_state[i].name);
																break;

												default:
																sprintf(str1, "		  %s - unknown CCLIB status\n", cclib_status_all.cclib_state[i].name);
																break;


								}
								strcat(str, str1);
				}
				printandlog(ALL_DEVICES, MISC, NULL, str, 0);

				// SKS create separate thread for handling events.
				if (pthread_create(&gEventHandlerThreadID, NULL,createEventHandlerThread, NULL)!=0 )
				{
								printf("pthread_create failure \n");
								exit(-1);
				}

				/* open all the GC devices */
				open_all_devices(); 

#ifdef USE_VIDEO
				int isFirstDTMFHit=true;
#endif

				char ch;
				int isHit = false;

				while ( interrupted==NO ) {
								if ( _kbhit() ) {
												ch = getchar();
												isHit = true;
								} else {
												isHit = false;
								}
								if(isHit) {
									if( (ch>=65 && ch<=90) || (ch>=97 && ch<=122)) {
										switch ( tolower(ch) ) {
												case 'q':
												{
															interrupted=YES;
															break;
												}
#ifdef USE_VIDEO
												case 's':
												{
															if(isFirstDTMFHit==true) {
																		isFirstDTMFHit=false;
																		// SKS create separate thread to send periodic (every 40ms) dtmf 
																		// It will be stopped if value reaches to 9 or user key-in 's' to stop.
																		if (pthread_create(&gDtmfEventHandlerThreadID, NULL,createDtmfSendHandlerThread, NULL)!=0 )
																		{
																					printf("pthread_create failure createDtmfSendHandlerThread()\n");
																		}
															} else {
																		g_bDtmfSend=false;
															}
															break;
												}
												case 'd':
												{
															if (pthread_create(&gDisconnectCallOnRequestHandlerThreadID, NULL,createDisconnectCallOnRequestHandlerThread, NULL)!=0 )
                              {
                                 printf("pthread_create failure createDisconnectCallOnRequestHandlerThread()\n");
                              }
															break;
												}
												case 'v':
												{
															if (pthread_create(&gMakeCallOnRequestHandlerThreadID, NULL,createMakeCallOnRequestHandlerThread, (void *)CALL_TYPE_VIDEO)!=0 )
                              {
                                 printf("pthread_create failure createMakeCallOnRequestHandlerThread()\n");
                              }
															break;
												}
                        case 'a':
                        {
                              if (pthread_create(&gMakeCallOnRequestHandlerThreadID, NULL,createMakeCallOnRequestHandlerThread, (void *)CALL_TYPE_AUDIO)!=0 )
                              {
                                 printf("pthread_create failure createMakeCallOnRequestHandlerThread()\n");
                              }
                              break;
                        }
												case 'p':
												{
															printChannelStatus();
															break;
												}
												case 'h':
												{
														printMenu();
														break;
												}
												default:
												{
														printMenu();
														break;
												}
#else 
												default:
												{
														printf("\nPress 'q' or 'Q' to Quit Application");
														break;
												}
#endif	
										}
									}	// ch value check
								}
								fflush(stdout);
								usleep(1000);
				} //end of while loop


				pthread_join(gEventHandlerThreadID,0);

				return 0;
} /* End of Main */

/****************************************************************
 *			NAME: void init_srl_mode(void)
 * DESCRIPTION: Inits SR mode
 *		RETURNS: None
 *	  CAUTIONS: NA
 ****************************************************************/
static void init_srl_mode(void)
{
				/* 
					 SRL MODEL TYPE - Single Threaded Async, event handler will be called in the Main thread 
				 */
#ifdef _WIN32
				int	mode = SR_STASYNC | SR_POLLMODE; 
#else
				int	mode = SR_POLLMODE;
#endif


				/* Set SRL mode */
#ifdef _WIN32
				if (sr_setparm(SRL_DEVICE, SR_MODELTYPE, &mode) == -1) {
								printandlog(ALL_DEVICES, NON_GC_APIERR, NULL, "Unable to set to SR_STASYNC | SR_POLLMODE", 0);
								exitdemo(1);
				}
				printandlog(ALL_DEVICES, MISC, NULL, "SRL Model Set to SR_STASYNC | SR_POLLMODE", 0);
#else
				if (sr_setparm(SRL_DEVICE, SR_MODEID, &mode) == -1) {
								printandlog(ALL_DEVICES, NON_GC_APIERR, NULL, "Unable to set mode ID to SR_POLLMODE ", 0);
								exitdemo(1);
				}

				printandlog(ALL_DEVICES, MISC, NULL, "SRL mode ID set to SR_POLLMODE", 0);
#endif
}

/****************************************************************
 *			NAME: open_all_logfiles(void)
 * DESCRIPTION: Function to Open the Logfiles and put file pointers in file 
 *					pointer field of port array. Later depending upon the index the
 *					particular file can be accessed to dump the data.
 *					If the network device name is present, then the log file
 *					uses the network device name
 *					else it uses the voice device name
 *		RETURNS: returns 1 if successful
 *	  CAUTIONS: Assumes device name includes network device name or a voice device name
 ****************************************************************/
static int open_all_logfiles(void)
{	
				char		temp_filename[MAX_DEVNAME], temp_str[MAX_STRING_SIZE];
				char		*i;
				int		index;

				for (index = 0; index < num_devices; index++) {
								/* build up temp_filename */
								strcpy(temp_filename, logfile);

								/* case 1: if dti name exists, use it */
								if (strcmp(port[index].netdevice, "NONE") != 0) {
												/* IP has different "dti" name than NON-IP */
												if (port[index].netdevice[0] == 'i') {
																//sscanf(port[index].devname, ":N_ipt%[^:]", temp_str); 
																sscanf(port[index].devname, ":N_%[^:]", temp_str); 
												} else {
																//sscanf(port[index].devname, ":N_dti%[^:]", temp_str); 
																sscanf(port[index].devname, ":N_%[^:]", temp_str); 
												}
												strcat(temp_filename,  "_");
												strcat(temp_filename, temp_str);
								}
								/* case 2: if voice name exists, use it */
								else if (strcmp(port[index].dxDeviceName, "NONE") != 0) {
												i = (char *)strstr(port[index].dxDeviceName, "B");
												strcpy(temp_str, i);
												strcat(temp_filename,  "_");
												strcat(temp_filename, temp_str);
								} else {
												printf("Missing both DTI and Voice resource in config file device name (%s)\n",
																				port[index].devname);
												exitdemo(1);
								}		
								strcat(temp_filename, ".log");

								if ((port[index].log_fp = fopen(temp_filename, "w")) == NULL) {
												printf("\nCannot open log file %s for write\n", temp_filename);
												perror("Reason is: ");
												exitdemo(1);
								}
								printf("%s successfully opened\n", temp_filename); 

				} /* End of for loop */

				boardLogFd=fopen("gc_basic_call_model_voice_video_3g_cnf_nbup_board.log","w");
				return 1;
}
void *createEventHandlerThread(void *context)
{
				int ret;
				/* 
					 -- Forever loop where the main work is done - wait for an event or user requested exit
				 */
				for (;;) {
								ret = sr_waitevt(500);					/* 1/2 second */
								if (ret != -1) {							/* i.e. not timeout */
												process_event();
								}

								if (interrupted == NO) {				/* flag set in intr_hdlr() */
												drop_outbound_calls_if_required();/* an artifact of the demo program */
												/* where the outbound side disconnects 1-2 */
												/* seconds after connected */
												make_calls_if_required();			/* handles error conditions on makecall */
								} else {
#ifdef USE_VIDEO
												initiate_exit_video_call();
#else
												initiate_exit();
#endif
												if (reset_linedevs_left_to_complete == 0) {
																exitdemo(1);
												}
								}
				}
}
#ifdef USE_VIDEO
void *createDtmfSendHandlerThread(void *context)
{
				// SKS send dtmf every 40ms
				while (g_bDtmfSend==true) {
								sendH245UII(0);
								usleep(40000000);
				}	// end of while loop
			return 0;
}
void *createMakeCallOnRequestHandlerThread(void *context)
{
		int ctr; 
		int retVal=-1;
		int availableChannels[MAXCHAN];
		int channel;
		int callType;
		int noOfChAvailable;
		char str[MAX_STRING_SIZE];

		callType=(int)context;
		strcpy(str,"\nOutgoing free channels:\n");
		for(ctr=0,noOfChAvailable=0;ctr<num_devices;ctr++) {
				if ((port[ctr].makecall_active == NO) && (port[ctr].direction == DIR_OUT)) {
						availableChannels[noOfChAvailable]=ctr;
						noOfChAvailable+=1;
						sprintf(str,"%s ch#%d to %s",str,ctr,port[ctr].destination_num);
						if(noOfChAvailable%2==0) {
							sprintf(str,"%s\n",str);
						} else {
							sprintf(str,"%s\t",str);
						}
        }
		}
		if(noOfChAvailable==0) {
				printf("Congratulations!!!!! No Channel Available to process your request!!!!!!!!!!!\n");
				return 0;
		}
		printf("%s\n Enter Channel# to place outgoing call:",str);
		retVal=scanf("%d",&channel);
		if(retVal==1) {
			for(ctr=0;ctr<noOfChAvailable;ctr++)
			{
				if(availableChannels[ctr]==channel) {
						printf("Placing Outgoing call on request ch#=%d | Number=%s\n",channel,port[channel].destination_num);
						port[channel].callType=callType;
						gc_demo_makecall(channel);
						return 0;
				}
			}
		} else {
			printf("Channel Not Found, Enter correct channel from list\n");
		}
		return 0;
}
void *createDisconnectCallOnRequestHandlerThread(void *context)
{
    int ctr;
		int direction;
    int busyChannels[MAXCHAN];
    int channel;
		int retVal=-1;
    int noOfChBusy;
    char str[MAX_STRING_SIZE];

    strcpy(str,"\nActive Calls on channels:\n");
    for(ctr=0,noOfChBusy=0;ctr<num_devices;ctr++) {
//				printf("index=%d|noOfChBusy=%d|bBusy=%d\n",ctr,noOfChBusy,port[ctr].bBusy);
        if (port[ctr].bBusy == true) {
            busyChannels[noOfChBusy]=ctr;
						direction=port[ctr].direction;
            noOfChBusy+=1;
            sprintf(str,"%s \t ch#%d Direction=%s",str,ctr,callDirectionStr[direction]);
            if(noOfChBusy%2==0) {
              sprintf(str,"%s\n",str);
            }
        }
    }
    if(noOfChBusy==0) {
        printf("\nHurray!!!! No active call on the system !!!!!!!!!!!\n");
        return 0;
    }
    printf("%s\nEnter Channel# to disconnect call:",str);
    retVal=scanf("%d",&channel);
		if(retVal==1) {
    	for(ctr=0;ctr<noOfChBusy;ctr++)
    	{
        if(busyChannels[ctr]==channel) {
						direction=port[channel].direction;
            printf("Disconnecting call on request ch#=%d | Direction=%s\n",channel,callDirectionStr[direction]);
						port[channel].isCallDisconnectInProgress=true;
						if(port[channel].techtype==SIP) {
							if(port[channel].callType==CALL_TYPE_AUDIO) {
								stopPlayAudioPrompt(channel);
							}
            	stopIpmMedia(channel);
						} else if( (port[channel].techtype==ISDN) || (port[channel].techtype==SS7) ) {
							if(port[channel].callType==CALL_TYPE_AUDIO) {
								stopPlayAudioPrompt(channel);
								unRouteDtiAndDx(channel);
								disconnect_video_call(channel);
							} else {
            		stopMedia(channel);
							}
						} else if(port[channel].techtype==BICC){
							if(port[channel].callType==CALL_TYPE_VIDEO) {
            		stopMedia(channel);
							} else {
            		stopIpmMedia(channel);
							}
						}
            return 0;
        }
    	}
      printf("Channel Not Found, Enter correct channel from list\n");
		} else {
      printf("Channel Not Found, Enter correct channel from list\n");
    }
    return 0;
}

static void printChannelStatus()
{
    int ctr;
    char str[MAX_STRING_SIZE];
		char netDevice[MAX_STRING_SIZE];
		char mediaDevice[MAX_STRING_SIZE];
		
		printf("\n=========================================================================================================================");
    printf("\nChannel Status:\nIndex\t\t   NetDevice\tSignalling\t\t\t     MediaDevice\t  Direction\t    Status");
		printf("\n=========================================================================================================================\n");
    for(ctr=0;ctr<num_devices;ctr++) {

				switch(port[ctr].techtype)
				{
						case SIP:
						{
								sprintf(netDevice,"%s",port[ctr].netdevice);
								sprintf(mediaDevice,"%s/%s/%s",port[ctr].dxDeviceName,port[ctr].deviceInfo[IPM].device_name,
												port[ctr].deviceInfo[MM].device_name);
								break;
						}
						case ISDN:
						case SS7:
						{
								sprintf(netDevice,"%s/%s",port[ctr].netdevice,port[ctr].deviceInfo[M3G].device_name);
								sprintf(mediaDevice,"%s/%s",port[ctr].dxDeviceName,port[ctr].deviceInfo[MM].device_name);
								break;
						}
						case BICC:
						{
								sprintf(netDevice,"%s/%s",port[ctr].netdevice,port[ctr].deviceInfo[M3G].device_name);
								sprintf(mediaDevice,"%s/%s/%s",port[ctr].dxDeviceName,port[ctr].deviceInfo[IPM].device_name,
												port[ctr].deviceInfo[MM].device_name);
								break;
						}
						default:
								break;
				}
#ifdef USE_CNF
								sprintf(mediaDevice,"%s/%s",mediaDevice,port[ctr].deviceInfo[CNFPTY].device_name);
#endif

      	printf("%5d\t%20s\t%5s\t%40s\t%11s\t%10s\n",ctr,netDevice,sigStr[port[ctr].techtype],
								mediaDevice,callDirectionStr[port[ctr].direction],(port[ctr].bBusy==true?"BUSY":"IDLE"));
    }
}

#endif

/****************************************************************
 *   NAME: void stopPlayAudioPrompt(int index)
 *   DESCRIPTION: stop playing audio prompt for audio call
 *   INPUT: index
 *   RETURNS: None
 *   CAUTIONS: NA
 ****************************************************************/

static void stopPlayAudioPrompt(int index)
{
/*
		if(port[index].bPlayActive==false) {
        return;
    }
    port[index].bPlayActive=false;
*/
		if(dx_stopch(port[index].dxH,EV_SYNC)==-1) {
				printandlog(index,MISC,NULL,"dx_stopch() failed",0);
		}
		else {
				printandlog(index,MISC,NULL,"dx_stopch() success",0);
		}
}

/****************************************************************
 *   NAME:  void dxPlayAudioPrompt(int index)
 *   DESCRIPTION: start play audio prompt for audio call
 *   INPUT: index
 *   RETURNS: None
 *   CAUTIONS: NA
 ****************************************************************/


static void dxPlayAudioPrompt(int index)
{

/*
    if(port[index].bPlayActive==true) {
        return;
    }
    port[index].bPlayActive=true;
*/


				int devH=port[index].dxH;
				int fdPrompt;	

				DX_IOTT iott;
				DV_TPT tpt;

				//fdPrompt=fopen("intro.vox","rb");
				fdPrompt=open("intro.vox",O_RDONLY);

				/* set up DX_IOTT */
				iott.io_type = IO_DEV|IO_EOT;
				iott.io_bufp = 0;
				iott.io_offset = 0;
				iott.io_length = -1; /* play till end of file */
				iott.io_fhandle=fdPrompt;

				/* set up DV_TPT */
				dx_clrtpt(&tpt,1);
				tpt.tp_type = IO_EOT; /* only entry in the table */
				tpt.tp_termno = DX_MAXDTMF; /* Maximum digits */
				tpt.tp_length = 4; /* terminate on four digits */
				tpt.tp_flags = TF_MAXDTMF; /* Use the default flags */

				/* clear previously entered digits */
				if (dx_clrdigbuf(devH) == -1) {
								/* process error */
				}

				/* Now play the file */
				if (dx_play(devH,&iott,&tpt,EV_ASYNC|PM_SR6) == -1) {
								printandlog(index,MISC,NULL,"dx_play() failed",0);
				}
				else {
								printandlog(index,MISC,NULL,"dx_play() success",0);
				}
}



/****************************************************************
 *			NAME: inbound_application(struct channel *pline)
 * DESCRIPTION: Place holder for your inbound application code in the connected state
 *		  INPUT: pline - pointer to port data structure
 *		RETURNS: NA
 *	  CAUTIONS: none
 ****************************************************************/
static void inbound_application(struct channel *pline)
{
				int index;
				index=pline->index;
#ifdef USE_VIDEO
				if(port[index].callType==CALL_TYPE_VIDEO && ((port[index].techtype==ISDN) || (port[index].techtype==SS7))) {
								startH245(index);
				} else if(port[index].callType==CALL_TYPE_AUDIO && ((port[index].techtype==ISDN) || (port[index].techtype==SS7)) ) {
								dxPlayAudioPrompt(index);
				} else if (port[index].techtype==SIP && port[index].callType==CALL_TYPE_AUDIO) {
// moved to DMEV_PORT_CONNECT for BICC else 
								startIpmMedia(index);
				}
#else
				dxPlayAudioPrompt(index);
#endif
}
/****************************************************************
 *			NAME: outbound_application(struct channel *pline)
 * DESCRIPTION: Place holder for your outbound application code in the connected state
 *		  INPUT: pline - pointer to port data structure
 *		RETURNS: NA
 *	  CAUTIONS: none
 ****************************************************************/
static void outbound_application(struct channel *pline)
{
				int index;
				index=pline->index;
#ifdef USE_VIDEO
				if(port[index].callType==CALL_TYPE_VIDEO && ((port[index].techtype==ISDN) || (port[index].techtype==SS7))) {
								startH245(index);
				}else if(port[index].callType==CALL_TYPE_AUDIO && ((port[index].techtype==ISDN) || (port[index].techtype==SS7))) {
								dxPlayAudioPrompt(index);
				} else if (port[index].techtype==SIP && port[index].callType==CALL_TYPE_AUDIO) {
// moved to DMEV_PORT_CONNECT for BICC else 
								startIpmMedia(index);
				}

#else
				char			str[MAX_STRING_SIZE];

				dxPlayAudioPrompt(index);

				time(&pline->dropcall_time);
				pline->dropcall_time += MIN_CONNECTED_TIME + 1; /* add 1 because the granularity is 1 sec */
				sprintf(str, "gc_DropCall() will be issued in %d-%d seconds", MIN_CONNECTED_TIME, MIN_CONNECTED_TIME + 1);
				printandlog(pline->index, MISC, NULL, str, 0);
#endif
}

/****************************************************************
 *			NAME: drop_outbound_calls_if_required(void)
 * DESCRIPTION: This drops outbound side calls if required
 *					For purposes of the demo, a call stays in the connected
 *					state state for a second or two, then is dropped by the outbound side
 *					Your application does not need this code
 *					and hence it is not optimized
 *		RETURNS: NA
 *	  CAUTIONS: none
 ****************************************************************/
static void drop_outbound_calls_if_required(void)
{
				int					index;
				struct channel		*pline;
				time_t				current_time;								/* in seconds */
				char					str[MAX_STRING_SIZE];
				int					callindex;

				/* get the current time in seconds */
				time(&current_time);

				/* check all devices for an "expired" drop call time */
				/* assumes caller of gc_ResetLinedev() set dropcall_time = 0 */
				for (index = 0; index < num_devices; index++) {
								pline = &port[index];				
								/* check if time tripped */
								if (pline->dropcall_time && (pline->dropcall_time <= current_time)) {
												/* drop call is no longer required */
												pline->dropcall_time = 0;

												/* Retrieve CRN  - assumes only 1 call is active */
												callindex = existCRN(-1, pline);
												if (callindex == OUT_OF_RANGE) {
																/* CRN not found, so log error and return from function. */
																sprintf(str, "CRN not found");
																printandlog(pline->index, MISC_WITH_NL, NULL, str, 0);
																return;
												}

												if (gc_DropCall(pline->call[callindex].crn, GC_NORMAL_CLEARING, EV_ASYNC) != GC_SUCCESS) {
																sprintf(str, "gc_DropCall(crn=0x%lx, cause=GC_NORMAL_CLEARING, mode=EV_ASYNC) Failed", pline->call[callindex].crn);
																printandlog(pline->index, GC_APIERR, NULL, str, 0);
																printandlog(pline->index, STATE, NULL, " ", callindex);
																exitdemo(1);
												}
												printandlog(index, MISC_WITH_NL, NULL, "********* Dropping outbound call from drop_outbound_calls_if_required() *********", 0);
												sprintf(str, "gc_DropCall(crn=0x%lx, cause=GC_NORMAL_CLEARING, mode=EV_ASYNC) Success", pline->call[callindex].crn);
												printandlog(pline->index, GC_APICALL, NULL, str, 0);

												/* Set the dropcall flag to YES on the related channel */
												pline->call[callindex].dropcall_active = YES;
								}
				}
}


/****************************************************************
 *			NAME: make_calls_if_required(void)
 * DESCRIPTION: This procedure sees if there are any calls that need to be made
 *					handles some makecall error conditions
 *		RETURNS: NA
 *	  CAUTIONS: none
 ****************************************************************/
static void make_calls_if_required(void)
{
				int					index;
				struct channel		*pline;
				time_t				current_time;								/* in seconds */

				time(&current_time);												/* get the current time in seconds */

				/* check all devices for an "expired" make call time */
				for (index = 0; index < num_devices; index++) {
								pline = &port[index];				
#ifdef USE_VIDEO
								if(pline->techtype==SIP) {
												continue;
								}
#endif
								if (pline->makecall_time && (pline->makecall_time <= current_time)) {			/* check if time tripped */
												printandlog(index, MISC, NULL, "makecall timer tripped - will retry the makecall", 0);
												pline->makecall_time = 0;								/* makecall request will now be met */
												gc_demo_makecall(index);								/* attempt to make the call again */
								}
								if (pline->intercall_time && (pline->intercall_time <= current_time)) {			/* check if time tripped */
												printandlog(index, MISC, NULL, "Inter-Call timer tripped - New call will now be attempted", 0);
												pline->intercall_time = 0;								/* makecall request will now be met */
												gc_demo_makecall(index);								/* attempt to make the call again */
								}

				}	// end of for loop
}

/****************************************************************
 *			NAME: process_event(void)
 * DESCRIPTION: Function to handle the GlobalCall Events 
 *		RETURNS: NA
 *	  CAUTIONS: none
 ****************************************************************/
static void process_event(void)
{
				METAEVENT			metaevent;
				struct channel		*pline;
				char					str[MAX_STRING_SIZE];
				int					index;				/* Device index */
				int					evttype;
				GC_INFO				t_info;	  			/* Structure which stores information about GC related errors */ 
				int					callindex;			/* CAUTION: once initialized, do not change */
				/* the value in this subroutine! */
				int					temp_callindex;	/* use the temporary callindex instead */

				/* Populate the metaEvent structure */
				if(gc_GetMetaEvent(&metaevent) != GC_SUCCESS)
				{
								/* serious problem - should never fail */
								printandlog(ALL_DEVICES, GC_APIERR, NULL, "gc_GetMetaEvent() failed", 0);
								exitdemo(1);
				}

				/* process GlobalCall events */
				if ((metaevent.flags & GCME_GC_EVENT) == 0)	{
								/*********************************************************************/
								/* Your application might have non-GC events , e.g. voice or fax		*/
								/* events that need to be processed. That would be done here.			*/
								/* However for the purposes of the demo, non_GC events are only		*/
								/* logged. Most likely you will need to do some sort of a				*/
								/* search of the port array to find the associated data structure		*/
								/* for this event. Since the event is not a GC event,						*/
								/* gc_GetUsrAttr() cannot help find the associated GC data structure */
								/*********************************************************************/
								//		sprintf(str, "Received a non-GC Event 0x%lx\n", metaevent.evttype);
								//		printandlog(ALL_DEVICES, MISC, NULL, str, 0);
								process_non_gc_event(&metaevent);
								return;												/* RETURN POINT */
				}
				/**************************************/
				/* when here => processing a GC event */
				/**************************************/
				/* initialize pline to port[index],
					 The pointer to this structure was stored in GC
					 during gc_OpenEx() */
				pline = (struct channel *) metaevent.usrattr;

				/* Retrieve device index */
				index = pline->index;

				evttype = metaevent.evttype;

				/* skip checks if no crn for this event */

				if ((metaevent.crn != 0) && (pline->resetlinedev_active == NO)) {
								/* Assumption: Only GCEV_DETECTED/GCEV_OFFERED will require the CRN
									 to be inserted into the array. */
								callindex = existCRN(metaevent.crn, pline);
								if (callindex == OUT_OF_RANGE) {
												/* CRN was not found, so insert CRN into the array.
												 * Should only occurs when GCEV_DETECTED or GCEV_OFFERED event is received.
												 * Note: GCEV_DETECTED is not enabled, but this code will support it if
												 * if you enable this event */
												if ((evttype == GCEV_DETECTED) || (evttype == GCEV_OFFERED)) {
																for (callindex = 0; callindex < MAX_CALLS; callindex++) {
																				if (pline->call[callindex].crn == 0) {
																								/* crn is not found, insert into crn array. */
																								pline->call[callindex].crn = metaevent.crn;
																								break;
																				}
																}
												} else {
																/* Error case: should not happen */
																sprintf(str, " Received an event(0x%x) with an invalid CRN(0x%lx)", evttype, metaevent.crn);
																printandlog(index, MISC_WITH_NL, NULL, str, 0);
																return;
												}
								}
				} else {
								callindex = 0;						/* some printing routines use callindex */
				}
				/**********************************************************************/
				/* CAUTION: at this point callindex is initialized, do not change it! */
				/**********************************************************************/

				printandlog(index, MISC_WITH_NL, NULL, "********* Received a GC event *********", 0);
				printandlog(index, EVENT, NULL, GCEV_MSG(evttype), 0);
				printandlog(index, STATE, NULL, " is the current GC call state ", callindex);

				/* 1st handle the generic events */
				switch (evttype) 
				{
								case GCEV_EXTENSION:
								{
												printandlog(index, MISC, NULL, "GCEV_EXTENSION received", 0);
												process_extension_event(pline, &metaevent);
												break;
								}
								case GCEV_EXTENSIONCMPLT:
								{
												printandlog(index, MISC, NULL, "GCEV_EXTENSIONCMPLT received", 0);
										break;
								}
								case GCEV_OPENEX:
								{
												printandlog(index, MISC, NULL, "GCEV_OPENEX received", 0);
												/* Although this demo currently only opens H.323 and SIP devices asynchronously */
												if ((pline->techtype == H323) || (pline->techtype == SIP)) {

																/* Enable alarm notification for H.323 or SIP devices which were opened ASYNC */
																enable_alarm_notification(pline);

																getNetworkHandle(index);
																/* Set DTMF mode to inband for SIP as default of Alphanumeric mode
																 * valid only for H.323 devices
																 */
																if (pline->techtype == SIP) {
#ifndef USE_VIDEO
																				getIpmHandle(index);
																				GC_PARM_BLK			  *parmblkp = NULL;
																				gc_util_insert_parm_val(&parmblkp, IPSET_DTMF, IPPARM_SUPPORT_DTMF_BITMASK,
																												sizeof(char), IP_DTMF_TYPE_INBAND_RTP);

																				if (gc_SetUserInfo(GCTGT_GCLIB_CHAN, port[index].ldev, parmblkp, GC_ALLCALLS) != GC_SUCCESS) {
																								sprintf(str, "gc_SetUserInfo(linedev=%ld) Failed configuring DTMF mode", port[index].ldev);
																								printandlog(index, GC_APIERR, NULL, str, 0);
																								exitdemo(1);
																				}

																				sprintf(str, "gc_SetUserInfo(linedev=%ld) Success - DTMF mode is inband", port[index].ldev);
																				printandlog(index, GC_APICALL, NULL, str, 0);
																				gc_util_delete_parm_blk(parmblkp);
#else
												GC_PARM_BLK       *parmblkp = NULL;
          							long            request_id;
          							gc_util_insert_parm_val(&parmblkp, GCSET_CALLEVENT_MSK, GCACT_ADDMSK,
            						sizeof(long), GCMSK_200_OK);
          							if (gc_SetConfigData(GCTGT_GCLIB_CHAN, pline->ldev, parmblkp, 0, GCUPDATE_IMMEDIATE, &request_id, EV_SYNC) != GC_SUCCESS) {
							              sprintf(str, "gc_SetConfigData(linedev=%ld) Failed - GCEV_200OK", port[index].ldev);
							              printandlog(index, GC_APIERR, NULL, str, 0);
							              exitdemo(1);
							          }
							          sprintf(str, "gc_SetConfigData(linedev=%ld) Success - GCEV_200OK", port[index].ldev);
							          printandlog(index, GC_APICALL, NULL, str, 0);
							          gc_util_delete_parm_blk(parmblkp);

#endif
																}
												}
												break;
										}

												/* This demo only opens H.323 and SIP devices asynchronously */
								case GCEV_OPENEX_FAIL:
								{
												printandlog(index, MISC, NULL, "GCEV_OPENEX_FAIL received", 0);
												printandlog(index, GC_RESULT_INFO, &metaevent, "Reason for GCEV_OPENEX_FAIL ", 0);
												exitdemo(1);
												break;
								}

								case GCEV_ALARM:
												print_alarm_info(&metaevent, pline);
												break;

								case GCEV_FATALERROR:
								{
												printandlog(index, GC_RESULT_INFO, &metaevent, "Reason for fatalerror ", 0);
												printandlog(index, STATE, NULL, " ", callindex);

												if (gc_ResultInfo(&metaevent, &t_info) < 0) {
																sprintf(str, "gc_ResultInfo() call failed");
																printandlog(index, GC_APIERR, NULL, str, 0);
																exitdemo(1);
												}

												if (t_info.gcValue == GCRV_RESETABLE_FATALERROR) {
																pline->dropcall_time = 0;								/* in case on */
																pline->resetlinedev_active = YES;					/* reset linedevice is on as a background process */
																/* do nothing else= wait till the GCEV_RESETLINEDEV completion event arrives */
												} else if (t_info.gcValue == GCRV_RECOVERABLE_FATALERROR) {
																/* do a close and open on recoverable, fatal errors */
																printandlog(index, MISC, NULL, " Recoverable fatal error occurred - closing device and re-opening it", 0);
																if (gc_Close(port[index].ldev) < 0) {
																				printandlog(index, GC_APIERR, NULL, "gc_Close() failed", 0);
																				exitdemo(1);
																}
																open_device(index);

												} else {
																printandlog(index, MISC, NULL, " Non-recoverable fatal error occurred", 0);
																exitdemo(1);
												}
												break;				
									}
								case GCEV_BLOCKED:
												pline->blocked = YES;						/* Nothing to do when a blocked event occurs */
												break;

								case GCEV_UNBLOCKED:
								{
												/* blocked/unblocked do not cause state transitions */
												pline->blocked = NO;
												if (pline->resetlinedev_required_after_unblocked == YES) {
																pline->resetlinedev_required_after_unblocked = NO;

																if (gc_ResetLineDev(port[index].ldev, EV_ASYNC) != GC_SUCCESS) {
																				sprintf(str, "gc_ResetLineDev(linedev=%ld, mode=EV_ASYNC) Failed", port[index].ldev);
																				printandlog(index, GC_APIERR, NULL, str, 0);
																				exitdemo(1);
																}
																pline->resetlinedev_active = YES;
																sprintf(str, "gc_ResetLineDev(linedev=%ld, mode=EV_ASYNC) Success", port[index].ldev);
																printandlog(index, GC_APICALL, NULL, str, 0);
																/* nothing more can do until the completion event arrives */
																break;
												}

												if (pline->waitcall_active == YES) {
																/* waitcall already active - do not call it again */
																break;
												}
												if (pline->resetlinedev_active == YES) {
																/* The line device is being reset */
																break;
												}
												/* FALL THROUGH ON PURPOSE */
								}
								case GCEV_RESETLINEDEV:
								{
												if (evttype == GCEV_RESETLINEDEV) {
																/* Reset call related flags and variables */
																pline->waitcall_active = NO;
																pline->makecall_active = NO;
																pline->resetlinedev_active = NO;
																for (temp_callindex = 0; temp_callindex < MAX_CALLS; temp_callindex++) {
																				/* reset internal variables to that of beginning */
																				pline->call[temp_callindex].call_state = GCST_NULL;
																				pline->call[temp_callindex].crn = 0;
																				pline->call[temp_callindex].dropcall_active = NO;
																}
												}

												if ((evttype == GCEV_RESETLINEDEV) 
																				&& (reset_linedevs_left_to_complete != 0)) {		/* tests if processing shutdown request */
																reset_linedevs_left_to_complete--;
																sprintf(str, "One less GCEV_RESETLINEDEV to wait for, %d left",
																								reset_linedevs_left_to_complete);
																printandlog(ALL_DEVICES, MISC, NULL, str, 0);

												}

												/* must further process the event if unblocked */
												if ((interrupted == NO) && (pline->blocked == NO)) {
																if (pline->direction == DIR_IN) {
																				/* Note: by definition, waitcall is not active at this point */
																				if (pline->blocked == NO) {
																								if (gc_WaitCall(pline->ldev, NULL, NULL, -1, EV_ASYNC) != GC_SUCCESS) {
																												sprintf(str, "gc_WaitCall(linedev=%ld, crnp=NULL, waittime=0, mode=EV_ASYNC) Failed", pline->ldev);
																												printandlog(index, GC_APIERR, NULL, str, 0);
																												exitdemo(1);
																								}
																								pline->waitcall_active = YES;
																								sprintf(str, "gc_WaitCall(linedev=%ld, crnp=NULL, waittime=0, mode=EV_ASYNC) Success", pline->ldev);
																								printandlog(index, GC_APICALL, NULL, str, 0);
																				}
																} else {
#ifndef USE_VIDEO
																				/* Make another call if possible */
																				gc_demo_makecall(index); 
#endif
																}
												}
												break;
								}
								case GCEV_TASKFAIL:
								{
												printandlog(index, GC_RESULT_INFO, &metaevent, "Reason for taskfail ", 0);
												printandlog(index, STATE, NULL, " ", callindex);

												/* best can do is a reset linedevice */
												/* if reset lindevice is already active, then retry it */
												if (gc_ResetLineDev(port[index].ldev, EV_ASYNC) == GC_SUCCESS) {
																pline->dropcall_time = 0;								/* in case on */
																pline->call[callindex].dropcall_active = NO;		/* in case on */
																pline->makecall_active = NO;							/* in case on */
																pline->resetlinedev_active = YES;
																sprintf(str, "gc_ResetLineDev(linedev=%ld, mode=EV_ASYNC) Success", port[index].ldev);
																printandlog(index, GC_APICALL, NULL, str, 0);
												} else {
																sprintf(str, "gc_ResetLineDev(linedev=%ld, mode=EV_ASYNC) Failed", port[index].ldev);
																printandlog(index, GC_APIERR, NULL, str, 0);
																if (pline->blocked == NO) {							/* If no alarms are active */
																				exitdemo(1);											/* Then should never get here */
																}
																/* when here, resetlinedev failed and in the blocked state */
																/* most likely due to an alarm being active */
																if (pline->call[callindex].crn) {					/* is there an active call? */
																				if (pline->call[callindex].dropcall_active == NO) {					/* drop and release the call if can */
																								if (gc_DropCall(pline->call[callindex].crn, GC_NORMAL_CLEARING, EV_ASYNC) == GC_SUCCESS)
																								{
																												sprintf(str, "gc_DropCall(crn=0x%lx, cause=GC_NORMAL_CLEARING, mode=EV_ASYNC) Success", pline->call[callindex].crn);
																												printandlog(pline->index, GC_APICALL, NULL, str, 0);
																												pline->call[callindex].dropcall_active = YES;
																								} else {
																												sprintf(str, "gc_DropCall(crn=0x%lx, cause=GC_NORMAL_CLEARING, mode=EV_ASYNC) Failed", pline->call[callindex].crn);
																												printandlog(pline->index, GC_APIERR, NULL, str, 0);
																												printandlog(pline->index, STATE, NULL, " ", callindex);
																								}
																				}
																}
																pline->resetlinedev_required_after_unblocked = YES;		/* try again later */
												}
												break;
								}
								case GCEV_DISCONNECTED:
								{
												if (pline->resetlinedev_active == NO) {
#ifdef USE_VIDEO
//														process_video_mode_disconnect(pline, &metaevent);
																port[index].isCallDisconnectInProgress=true;
																if (port[index].callType==CALL_TYPE_AUDIO) {
																		if(port[index].techtype==BICC) {
																					stopPlayRecord(index,true);
																					stopPlayRecord(index,false);
																		} else {
																				stopPlayAudioPrompt(index);
																		}
																		stopIpmMedia(index);
																		if(port[index].techtype==ISDN || port[index].techtype==SS7) {
																					unRouteDtiAndDx(index);
																					process_disconnected_event(pline, &metaevent);	/* process the disconnected event */
																		} else if(port[index].techtype==SIP) {
																					unRouteIpmAndDx(index);
																		}
																} else {				// else video call
																		if(port[index].techtype!=SIP) {
                                         if(port[index].bMediaStarted==true) {
                                              stopMedia(index);
                                          } else if(port[index].bH245Started==true) {
                                              stopH245(index);
                                          } else {
                                              disconnect_video_call(index);
                                          }
																			} else {
																					stopPlayRecord(index,false);
																					stopPlayRecord(index,true);
																					stopIpmMedia(index);
																			}
																}
#else 
																stopPlayAudioPrompt(index);
																if( (port[index].techtype==ISDN) || (port[index].techtype==SS7)) {
																	unRouteDtiAndDx(index);
																} else if(port[index].techtype==SIP){
																		unRouteIpmAndDx(index);
																}
																process_disconnected_event(pline, &metaevent);	/* process the disconnected event */
#endif
												}
												break;
								}
								case GCEV_SIP_200OK:
								{
										if(pline->call[callindex].call_state == GCST_CONNECTED) {
											if ( gc_SipAck(pline->call[callindex].crn, NULL, EV_ASYNC) != 0 ) {
													printandlog(pline->index, GC_APIERR, NULL,"Error calling gc_SIPAck",0);
											} else {
													printandlog(pline->index, GC_APICALL, NULL,"gc_SIPAck() success",0);
											}
										}
										break;
								}
								default:
								{
												if (pline->resetlinedev_active == NO) {	/* ignore all call related events with reset linedevice active */
																if (pline->direction == DIR_IN) {
																				process_call_related_inbound_event(pline, &metaevent);
																} else {
																				process_call_related_outbound_event(pline, &metaevent);
																}
												}
												break;
								}
				}	// SKS end of event parsing switch

				printandlog(index, STATE, NULL, " is the new GC call state after processing the event", callindex);
}

/****************************************************************
 *			NAME: void	print_alarm_info(METAEVENTP metaeventp, struct channel *pline)
 * DESCRIPTION: Prings alarm information
 *		 INPUTS: metaeventp - pointer to the alarm event
 *              pline - pointer to the channel data structure
 *		RETURNS: NA
 *	  CAUTIONS: Assumes already known to be an alarm event
 ****************************************************************/
static void	print_alarm_info(METAEVENTP metaeventp, struct channel *pline)
{
				long					alarm_number;
				char					*alarm_name;
				unsigned long		alarm_source_objectID;
				char					*alarm_source_object_name;
				char					str[MAX_STRING_SIZE];

				if (gc_AlarmNumber(metaeventp, &alarm_number) != GC_SUCCESS)
				{
								sprintf(str, "gc_AlarmNumber(...) FAILED");
								printandlog(pline->index, GC_APIERR, NULL, str, 0);
								printandlog(pline->index, STATE, NULL, " ", 0);
								exitdemo(1);
				}
				if (gc_AlarmName(metaeventp, &alarm_name) != GC_SUCCESS)
				{
								sprintf(str, "gc_AlarmName(...) FAILED");
								printandlog(pline->index, GC_APIERR, NULL, str, 0);
								printandlog(pline->index, STATE, NULL, " ", 0);
								exitdemo(1);
				}

				if (gc_AlarmSourceObjectID(metaeventp, &alarm_source_objectID) != GC_SUCCESS)
				{
								sprintf(str, "gc_AlarmSourceObjectID(...) FAILED");
								printandlog(pline->index, GC_APIERR, NULL, str, 0);
								printandlog(pline->index, STATE, NULL, " ", 0);
								exitdemo(1);
				}

				if (gc_AlarmSourceObjectName(metaeventp, &alarm_source_object_name) != GC_SUCCESS)
				{
								sprintf(str, "gc_AlarmSourceObjectName(...) FAILED");
								printandlog(pline->index, GC_APIERR, NULL, str, 0);
								printandlog(pline->index, STATE, NULL, " ", 0);
								exitdemo(1);
				}

				sprintf(str, "Alarm %s (0x%lx) occurred on ASO %s (%d)",
												alarm_name, alarm_number, alarm_source_object_name, (int) alarm_source_objectID);

				printandlog(pline->index, MISC, NULL, str, 0);
}



/****************************************************************
 *			NAME: process_disconnected_event(struct channel *pline, METAEVENT *metaeventp)
 * DESCRIPTION: Function to process a disconnected event
 *		 INPUTS: pline - pointer to entry in port table
 *					metaeventp - pointer to the current metaevent
 *		RETURNS: NA
 *	  CAUTIONS: Assumes event logging already done
 ****************************************************************/
static void process_disconnected_event(struct channel *pline, METAEVENT *metaeventp)
{
				char		str[MAX_STRING_SIZE];
				int		callindex;

				/* Retrieve CRN */
				callindex = existCRN(metaeventp->crn, pline);
				if (callindex == OUT_OF_RANGE) {
								/* CRN not found, so log error and return from function. */
								sprintf(str, "CRN not found");
								printandlog(pline->index, MISC_WITH_NL, NULL, str, 0);
								return;
				}

				switch (pline->call[callindex].call_state)
				{
								case GCST_IDLE:
												/*************************************************************/
												/* this represents a simultaneous disconnect						 */
												/* do nothing as gc_DropCall() must have already been issued */
												/* to reach this point!													 */
												/*************************************************************/
												break;

								case GCST_NULL:
												/* In case have a disconnect before dialing is reached */
												if (pline->makecall_active == NO) {
																/* this represents an error */
																printandlog(pline->index, MISC, NULL, "Received GCEV_DISCONNECTED in NULL call state", 0); 

																/* however, there is a PT - 24402 - which indicates that the inbound side */
																/* may under alarm conditions receive a disconnected event */
																/* even though the call has not been offered */
																/* the other work-around is to enable the detected state, but */
																/* for purposes of showing a second work-around, we will "create" the call */
																/* and then drop and release it */
																pline->call[callindex].crn = metaeventp->crn;					/* "create the call */
																/* the call state will be init in default */

																/* break; - when PT24402 is fixed, put this back in and pull the call "creation" */
												}
												/* fall through on purpose - need to do a dropcall and release call since there is a call active */
												/* may have reached this condition if there is a problem before dialing */
												/* or the inbound side has not enabled detection of events and the disconnected */
												/* event arrives (usually due to alarm conditions) without the app previously */
												/* being informed the call exists */

								default:
												/* GCEV_DISCONNECTED is allowed in all other call states */
												/* If dropcall is already active, don't reissue it */
												/* this represents a simultaneous disconnect */
												if(pline->call[callindex].dropcall_active == NO)
												{
																if (gc_DropCall(pline->call[callindex].crn, GC_NORMAL_CLEARING, EV_ASYNC) != GC_SUCCESS)
																{
																				sprintf(str, "gc_DropCall(crn=0x%lx, cause=GC_NORMAL_CLEARING, mode=EV_ASYNC) Failed", pline->call[callindex].crn);
																				printandlog(pline->index, GC_APIERR, NULL, str, 0);
																				printandlog(pline->index, STATE, NULL, " ", callindex);
																				exitdemo(1);
																}
																sprintf(str, "gc_DropCall(crn=0x%lx, cause=GC_NORMAL_CLEARING, mode=EV_ASYNC) Success", pline->call[callindex].crn);
																printandlog(pline->index, GC_APICALL, NULL, str, 0);
																pline->call[callindex].dropcall_active = YES;
												}
												pline->call[callindex].call_state = GCST_DISCONNECTED;
												pline->dropcall_time = 0;								/* dropcall is no longer required */
#ifdef USE_VIDEO
												pline->isCallDisconnectInProgress=false;
#endif
												break;
				}
}

/****************************************************************
 *			NAME: process_call_related_inbound_event(struct channel *pline, METAEVENT *metaeventp)
 * DESCRIPTION: Function to process inbound GlobalCall event
 *		 INPUTS: pline - pointer to entry in port table
 *					metaeventp - pointer to the metaevent structure
 *		RETURNS: NA
 *	  CAUTIONS: Tightly coupled with process_event() for pre-processing
 *					Assumes event logging already done
 *					Disconnected event already handled
 ****************************************************************/
static void process_call_related_inbound_event(struct channel *pline, METAEVENT *metaeventp)
{
				int		evttype, index;
				char		str[MAX_STRING_SIZE];
				int		unexpected_event = 0;					/*assume not till proven otherwise */
				/* May be overridden for GCEV_UNBLOCKED */
				int		callindex;

				evttype = metaeventp->evttype;
				index = pline->index;

				/* Retrieve CRN */
				callindex = existCRN(metaeventp->crn, pline);
				if (callindex == OUT_OF_RANGE) {
								/* CRN not found, so log error and return from function. */
								sprintf(str, "CRN not found");
								printandlog(pline->index, MISC_WITH_NL, NULL, str, 0);
								return;
				}
				sprintf(str, "SKS Rx Event Type = 0x%x",evttype);
				printandlog(index, MISC_WITH_NL, NULL, str, 0);

				/* Implement the state machine */
				switch (pline->call[callindex].call_state)
				{
						case GCST_NULL:
						{
								switch (evttype)
								{
										case GCEV_OFFERED:
										{
											sprintf(str, "GCEV_OFFERED received in GCST_NULL state on device=%s",port[index].netdevice);
											printandlog(pline->index, MISC_WITH_NL, NULL, str, 0);
											process_offered_event(metaeventp, pline);
											break;
										}
										case GCEV_CALLINFO:
										{
											sprintf(str, "GCEV_CALLINFO received in GCST_NULL state");
											printandlog(pline->index, MISC_WITH_NL, NULL, str, 0);
											break;
										}
										default:
											unexpected_event = 1;
											break;
									}
									break;
									/* end of case GCST_NULL */
								}

						case GCST_OFFERED:
						{
																if(evttype==GCEV_CALLINFO) {
																				sprintf(str, "GCEV_CALLINFO received in GCST_OFFERED state");
																				printandlog(pline->index, MISC_WITH_NL, NULL, str, 0);
																}
																if (evttype == GCEV_ACCEPT) {
																				pline->call[callindex].call_state = GCST_ACCEPTED;


#ifndef USE_VIDEO
																				/* Microsoft Phoenix Version 5.0 Soft SIP phone requires SDP field in OK event so we */
																				/* specify coders prior to calling gc_AnswerCall( ) to ensure OK event has SDP field.*/
																				/* Though done specifically for Phoenix interopability, this will not affect         */
																				/* interoperability with other SIP clients.                                          */
																				if(pline->techtype == SIP) {
																								gc_set_channel_codecs(pline->index);
																				}
#endif				
																				/* Note: if both answercall and acceptcall use 0 rings  */
																				/* then the outbound side may not get an alerting event */
																				/* this behavior is technology/protocol dependent		  */
																				if (gc_AnswerCall(pline->call[callindex].crn, 0, EV_ASYNC) != GC_SUCCESS) {
																								sprintf(str, "gc_AnswerCall(crn=0x%lx, # of rings=0, mode=EV_ASYNC) Failed", pline->call[callindex].crn);
																								printandlog(index, GC_APIERR, NULL, str, 0);
																								exitdemo(1);
																				}
																				sprintf(str, "gc_AnswerCall(crn=0x%lx, mode=EV_ASYNC) Success", pline->call[callindex].crn);
																				printandlog(index, GC_APICALL, NULL, str, 0);
																} else {
																				unexpected_event = 1;
																}
																break;
												}

								case GCST_ACCEPTED:
												{
																if (evttype == GCEV_ANSWERED) {
																				/* Do nothing but change the state to connected. The call will be cleared from
																					 outbound side. Just wait for Disconnect */
																				/* Your application specific code will go here */
																				pline->call[callindex].call_state = GCST_CONNECTED;

																				/* Increment the total calls on this device */
																				pline->numb_calls++;
																				/* Increment the total inbound calls on all devices */
																				in_calls++;
																				inbound_application(pline);			/* YOUR APPLICATION CODE IS CALLED HERE */
																} else {
																				unexpected_event = 1;
																}
																break;
												}

								case GCST_CONNECTED:
												{
																/*************************************************/
																/* Depending upon your application, you may or	 */
																/* may not have application specific event		 */
																/* handling here for when in the connected state */
																/* You could also have your app specific code	 */
																/* in process_event()									 */
																/*************************************************/
															if (evttype == GCEV_SIP_ACK_OK) {
															} else {
																unexpected_event = 1;						/* recall that GCEV_DISCONNECTED is handled elsewhere */
															}
																break;
												}

												/*************************************/
												/* now come the call teardown states */
												/*************************************/
								case GCST_DISCONNECTED:
												{
																/* If the GCEV_OFFERED event is observed in this state,
																	 then add the code that in located in GCST_IDLE state
																	 to resolved the issue. */
																if (evttype == GCEV_DROPCALL) {
																				pline->call[callindex].dropcall_active = NO;

																				/* Call ReleaseCall */
																				pline->call[callindex].call_state = GCST_IDLE;

																				if (gc_ReleaseCallEx(pline->call[callindex].crn, EV_ASYNC) != GC_SUCCESS) {
																								sprintf(str, "gc_ReleaseCallEx(crn=0x%lx, EV_ASYNC) Failed", pline->call[callindex].crn);
																								printandlog(index, GC_APIERR, NULL, str, 0);
																								exitdemo(1);
																				}
																				sprintf(str, "gc_ReleaseCallEx(crn=0x%lx, EV_ASYNC) Success", pline->call[callindex].crn);
																				printandlog(index, GC_APICALL, NULL, str, 0);
																				fflush(stdout);
																} else {
																				unexpected_event = 1;
																}
																break;
												}

								case GCST_IDLE:
												{
																/* In most cases, only GCEV_RELEASECALL will be seen in this state,
																	 but GCEV_OFFERED can be received after gc_ReleaseCallEx() and
																	 before GCEV_RELEASECALL. */
																if (evttype == GCEV_RELEASECALL) {
																				pline->call[callindex].call_state = GCST_NULL;
																				pline->call[callindex].crn = 0;
																				port[index].bBusy=false;
																				resetVariables(index);
																} else if (evttype == GCEV_OFFERED) {
																				process_offered_event(metaeventp, pline);
																} else {
																				unexpected_event = 1;
																}
																break;
												}

								default:
												{
																/* Should never get here */
																printandlog(index, MISC, NULL, "Invalid state in process_call_related_inbound_event()", 0); 
																printandlog(index, STATE, NULL, " ", callindex);
																break;
												}
				}

				if (unexpected_event && (evttype != GCEV_UNBLOCKED)) {
								printandlog(index, GC_RESULT_INFO, metaeventp, "Event was unexpected in the current state", 0);
								printandlog(index, STATE, NULL, " ", callindex);
				}
}

// SKS get type(audio/video) of incoming call
/****************************************************************
 *   NAME: int getCallType(METAEVENT *metaevent,struct channel *pline)
 *   DESCRIPTION: upon GCEV_OFFERED this function will determine
 whether incomig call is an audio or a video call
 *   INPUT: METAEVENT, struct channel
 *   RETURNS: CALL_TYPE_AUDIO | CALL_TYPE_VIDEO
 *   CAUTIONS: NA
 ****************************************************************/
static int getCallType(METAEVENT *metaevent,struct channel *pline)
{
				int callType=CALL_TYPE_AUDIO;
				int         nCtr;
				S7_IE_BLK      ie_blk;
				int         ieType,ieLen;
				int         x;
				int         index=pline->index;
				int         nAllIELen=0;
				char        str[128];

				if(port[index].techtype==ISDN) {
								if (gc_GetSigInfo(pline->ldev, (char *)&ie_blk, U_IES, metaevent) != GC_SUCCESS)  {
												// process error
												printandlog(pline->index, MISC, NULL, "gc_GetSigInfo() call failed", 0);
								}
				} else if(port[index].techtype==SS7) {
								if (gc_GetSigInfo(pline->ldev,(char *)&ie_blk, 0, metaevent) != GC_SUCCESS)  {
												// process error
												printandlog(pline->index, MISC, NULL, "gc_GetSigInfo() call failed", 0);
								}
				}
				nAllIELen=(int)ie_blk.length;
				for (nCtr=0;nCtr<(nAllIELen-1);)  {
								ieType=(int)(unsigned char)ie_blk.data[nCtr];
								ieLen=(int)(unsigned char)ie_blk.data[nCtr+1];
								sprintf(str,"IE Type=0x%x| IE Len=0x%x ==> |",ieType ,ieLen);
								nCtr+=2;
								if(ieType==4 || ieType==124 || ieType==29 || ieType==3) {
												int val1, val2, val3;
												val1=(int)(unsigned char)ie_blk.data[nCtr];
												val2=(int)(unsigned char)ie_blk.data[nCtr+1];
												val3=(int)(unsigned char)ie_blk.data[nCtr+2];
												if(val1==136 && val2==144 && val3==166) {
																callType=CALL_TYPE_VIDEO;
																return CALL_TYPE_VIDEO;
												}
												if(ieType==3) {
																val1=(int)(unsigned char)ie_blk.data[nCtr+2];
																val2=(int)(unsigned char)ie_blk.data[nCtr+3];
																val3=(int)(unsigned char)ie_blk.data[nCtr+4];
																if(val1==136 && val2==144 && val3==166) {
																				callType=CALL_TYPE_VIDEO;
																				return CALL_TYPE_VIDEO;
																}
												}
								}
								for(x=0;x<ieLen;x++,nCtr++) {
												sprintf(str,"%s0x%x|",str, (unsigned char)ie_blk.data[nCtr]);
								}
								sprintf(str,"%s",str);
				}
				return callType;
}

/****************************************************************
 *			NAME: process_offered_event(struct channel *pline)
 * DESCRIPTION: Function to process an offered event while in the NULL state
 *		 INPUTS: pline - pointer to entry in port table
 *		RETURNS: NA
 *	  CAUTIONS: Assumes event logging already done
 ****************************************************************/
static void process_offered_event(METAEVENT *metaeventp, struct channel *pline)
{
				char		dnis[GC_ADDRSIZE];
				char		ani[GC_ADDRSIZE];
				char		caller_name[GC_ADDRSIZE];
				char		str[MAX_STRING_SIZE];
				int		callindex;
				int index;

				/* Retrieve CRN */
				callindex = existCRN(metaeventp->crn, pline);
				if (callindex == OUT_OF_RANGE) {
								printandlog(pline->index, MISC, NULL, "metaeventp->crn unexpectedly not found for a GCEV_OFFERED event", 0);
								exitdemo(1);
				}
				index=pline->index;
				/* Save the CRN in port[index] and call Acceptcall */
				pline->call[callindex].call_state = GCST_OFFERED;
				pline->call[callindex].crn = metaeventp->crn;
			
				port[index].bBusy=true;

				if ((pline->techtype ==ISDN) || (pline->techtype == SS7)) {
								pline->callType=getCallType(metaeventp,pline);

								if(pline->callType==CALL_TYPE_VIDEO) {
												printandlog(index, MISC, NULL, "Video Call",0);
								} else {
												printandlog(index, MISC, NULL, "Audio Call",0);
								}
				} else if(pline->techtype == SIP) {
#ifdef USE_VIDEO
					int retVal=-1;
					retVal=parseOfferedSdp(metaeventp,pline);
					switch(retVal) 
					{
							case SDP_PARSE_SUCCESS:
										printandlog(index, MISC, NULL, "Successfully Parsed SDP",0);
										break;
							case SDP_NOT_OFFERED:
										printandlog(index, MISC, NULL, "SDP not offered in INVITE",0);
										break;
							case NO_MATCHING_VIDEO:
							case NO_MATCHING_AUDIO:
										disconnect_video_call(index);
										return;
										break;
					}
#endif
				} else if (pline->techtype == BICC) {
				#ifdef USE_VIDEO
					if(parseApmMessage(metaeventp,pline)!=SDP_PARSE_SUCCESS) {
							printandlog(index, MISC, NULL, "Droping Call, no media information found in APM message",0);
							disconnect_video_call(index);
							return;
					}
				#endif
				}

#ifdef USE_VIDEO
				GC_PARM_BLKP        gcParmBlk = NULL;
				if( (pline->callType==CALL_TYPE_VIDEO) && ( (pline->techtype == ISDN) || (pline->techtype == SS7) ) ) {
						connectM3gAndDti(index);
						routeUnrouteDevices(index,M3G,MM,ROUTE,DMFL_TRANSCODE_ON);
				} else if(pline->techtype == SIP) {
						if (pline->callType==CALL_TYPE_VIDEO) {
								routeUnrouteDevices(index,IPM,MM,ROUTE,DMFL_TRANSCODE_ON);
						}
						buildOfferAnswerSdp(index);
						if( gc_util_insert_parm_ref_ex(&gcParmBlk, IPSET_SDP, IPPARM_SDP_ANSWER, strlen(pline->sdpText), pline->sdpText)!=GC_SUCCESS){
								printandlog(pline->index, MISC, NULL,"process_offered_event::gc_util_insert_parm_ref_ex failed",0);
								return;
						}
						if( gc_SetUserInfo(GCTGT_GCLIB_CRN,pline->call[callindex].crn, gcParmBlk, GC_SINGLECALL)!=GC_SUCCESS) {
								printandlog(pline->index, MISC, NULL,"process_offered_event::gc_SetUserInfo failed",0);
								return;
						}
						printandlog(pline->index, MISC, NULL, pline->sdpText,0);
				} else if (pline->techtype == BICC) {
						if (pline->callType==CALL_TYPE_VIDEO) {
                routeUnrouteDevices(index,M3G,IPM,ROUTE,DMFL_TRANSCODE_NATIVE);	// SKS TODO H223 port routing
        		}
        		buildOfferAnswerSdp(index);
						// SKS TODO
						S7_IE_BLK      ie_blk;
						GC_IE_BLK 	gc_ie_blk;
						int length=0;

						length=buildApmMessage(index);
						
						ie_blk.length=length;

						memcpy(ie_blk.data,pline->biccApmMessage,length);
						int i;
						printf("\n In Hex\n");
						for(i=0;i<ie_blk.length;i++) {
								printf("%02x ",ie_blk.data[i]);
						}
						printf("\n\n\n\nnow character:\n");
						for(i=0;i<ie_blk.length;i++) {
								printf("%c",ie_blk.data[i]);
						}

						gc_ie_blk.gclib = NULL;
						gc_ie_blk.cclib = &ie_blk;
						if (gc_SndMsg(pline->ldev, pline->call[callindex].crn, 0x41 /*APM*/, &gc_ie_blk) != GC_SUCCESS) {
							printandlog(pline->index, GC_APIERR, NULL, "gc_SndMsg(APM) failed",0);
						}
        }		// SKS BICC ends here

#endif
				if(port[index].callType==CALL_TYPE_AUDIO) {
						switch(pline->techtype)
						{
								case ISDN:
								case SS7:
									routeDtiAndDx(index);
									break;
								case SIP:
									routeIpmAndDx(index);
									break;
#ifdef USE_VIDEO
								case BICC:
                	routeUnrouteDevices(index,IPM,MM,ROUTE,DMFL_TRANSCODE_NATIVE);
									break;
#endif
								default:
										break;
						}
				}

				/* show some technology specific functionality */
				/* get destination and origination  if not ANALOG */
				if (pline->techtype != ANALOG) {
								/* first, get called party number */
								if (gc_GetCallInfo(pline->call[callindex].crn, DESTINATION_ADDRESS, dnis) == GC_SUCCESS) {
												sprintf(str, "gc_GetCallInfo(crn=0x%lx) Success - called party = %s", pline->call[callindex].crn, dnis);
								} else {
												sprintf(str, "gc_GetCallInfo(crn=0x%lx) Failure - called party not available ", pline->call[callindex].crn);
								}
								printandlog(pline->index, GC_APICALL, NULL, str, 0);
								/* do application specific dnis event processing here */
								/* For simplicity's sake, max DDI logic is not shown here */

								/* next get the calling party here */
								if (gc_GetCallInfo(pline->call[callindex].crn, ORIGINATION_ADDRESS, ani) == GC_SUCCESS) {
												sprintf(str, "gc_GetCallInfo(crn=0x%lx) Success - calling party = %s", pline->call[callindex].crn, ani);
								} else {
												sprintf(str, "gc_GetCallInfo(crn=0x%lx) Failure - calling party not available", pline->call[callindex].crn);
								}
								printandlog(pline->index, GC_APICALL, NULL, str, 0);
				} else {
								/* get calling party, if Analog */
								if (gc_GetCallInfo(pline->call[callindex].crn, ORIGINATION_ADDRESS, ani) == GC_SUCCESS) {
												/* get ANI */		
												sprintf(str, "gc_GetCallInfo(crn=0x%lx) Success - calling party  = %s", pline->call[callindex].crn, ani);
								} else {
												sprintf(str, "gc_GetCallInfo(crn=0x%lx) Failure - calling party not available", pline->call[callindex].crn);
								}
								printandlog(pline->index, GC_APICALL, NULL, str, 0);

								/* get the call info */
								if (gc_GetCallInfo(pline->call[callindex].crn, CALLNAME, caller_name) == GC_SUCCESS) {
												sprintf(str, "gc_GetCallInfo(crn=0x%lx, CALLNAME) Success, name = %s",
																				pline->call[callindex].crn, caller_name);
								} else {
												sprintf(str, "gc_GetCallInfo(crn=0x%lx, CALLNAME) Failure, name not available",
																				pline->call[callindex].crn);
								}
								printandlog(pline->index, GC_APICALL, NULL, str, 0);
				}

				/* Note: if # of rings is == 0 then the outbound side */
				/* may not get the alerting event, depending upon the protocol */ 
				/* if the number of rings is 1, there is an occasional problem in */
				/* in the outbound side's ability to recognize the tone */
				if (gc_AcceptCall(pline->call[callindex].crn, 3, EV_ASYNC) != GC_SUCCESS) {
								sprintf(str, "gc_AcceptCall(crn=0x%lx, # of rings=3, mode=EV_ASYNC) Failed", pline->call[callindex].crn);
								printandlog(pline->index, GC_APIERR, NULL, str, 0);
								exitdemo(1);
				}
				sprintf(str, "gc_AcceptCall(crn=0x%lx, mode=EV_ASYNC) Success", pline->call[callindex].crn);
				printandlog(pline->index, GC_APICALL, NULL, str, 0);

}

#ifdef USE_VIDEO
static int connectDevicesForMakeCall(METAEVENT *metaeventp, struct channel *pline)
{
				int index=pline->index;
				int		callindex;
				int retVal=false;

				callindex = existCRN(metaeventp->crn, pline);

				if(port[index].callType==CALL_TYPE_VIDEO) {
								if(port[index].techtype==SIP) {
												retVal=parseOfferedSdp(metaeventp,pline);
												switch(retVal)
												{
																case SDP_ALREADY_OFFERED:
																				printandlog(index, MISC, NULL, "SDP again offered by remote EP, parsed already ignoring current one",0);
																				break;
																case SDP_PARSE_SUCCESS:
																{
																				printandlog(index, MISC, NULL, "Successfully Parsed SDP",0);
																				routeUnrouteDevices(index,IPM,MM,ROUTE,DMFL_TRANSCODE_ON);
																				break;
																}
																case SDP_NOT_OFFERED:
																				printandlog(index, MISC, NULL, "SDP not offered yet",0);
																				break;
																case NO_MATCHING_VIDEO:
																case NO_MATCHING_AUDIO:
																{
																				printandlog(index, MISC, NULL, "Dropping Call .........",0);
																				// SKS JACK (Kapanga do not send ACK to BYE unless HMP sends ACK to its 200_OK) TODO correct procedures
													              if ( gc_SipAck(port[index].call[callindex].crn, NULL, EV_ASYNC) != 0 )
												                {
												                        printandlog(pline->index, GC_APIERR, NULL,"Error calling gc_SIPAck",0);
												                }

																				disconnect_video_call(index);
																				return SDP_ERROR;
																				break;
																}
																default:
																		break;
												}
								}	else if (port[index].techtype==ISDN || port[index].techtype==SS7){
												connectM3gAndDti(index);
												routeUnrouteDevices(index,M3G,MM,ROUTE,DMFL_TRANSCODE_ON);
								} else if (port[index].techtype==BICC) {
												routeUnrouteDevices(index,M3G,IPM,ROUTE,DMFL_TRANSCODE_NATIVE);
								}
				} else if(port[index].callType==CALL_TYPE_AUDIO) {
            switch(port[index].techtype)
            {
                case ISDN:
                case SS7:
                  routeDtiAndDx(index);
                  break;
                case SIP:
								{
									retVal=parseOfferedSdp(metaeventp,pline);
                  switch(retVal)
                  {
                     	case SDP_ALREADY_OFFERED:
                          printandlog(index, MISC, NULL, "SDP again offered by remote EP, parsed already ignoring current one",0);
                          break;
                     	case SDP_PARSE_SUCCESS:
                          printandlog(index, MISC, NULL, "Successfully Parsed SDP",0);
                          break;
                     	case SDP_NOT_OFFERED:
                          printandlog(index, MISC, NULL, "SDP not offered yet",0);
                          break;
                    	case NO_MATCHING_VIDEO:
                    	case NO_MATCHING_AUDIO:
											{
                             printandlog(index, MISC, NULL, "Dropping Call .........",0);
														 // SKS JACK (Kapanga do not send ACK to BYE unless HMP sends ACK to its 200_OK) TODO correct procedures
													   if ( gc_SipAck(port[index].call[callindex].crn, NULL, EV_ASYNC) != 0 )
												     {
												          printandlog(pline->index, GC_APIERR, NULL,"Error calling gc_SIPAck",0);
												     }
                             disconnect_video_call(index);
                             return SDP_ERROR;
                             break;
												}
												default:
													break;
                    }
                  	routeIpmAndDx(index);
                  	break;
								}
                case BICC:
								{
                  	routeUnrouteDevices(index,IPM,MM,ROUTE,DMFL_TRANSCODE_NATIVE);
                  	break;
								}
                default:
                    break;
				}
			}	// audio call 
			return 0;
}
#endif

static void process_extension_event(struct channel *pline, METAEVENT *metaeventp)
{
#ifdef USE_VIDEO
		if (pline->techtype == BICC && pline->direction==DIR_OUT) {
          if(parseApmMessage(metaeventp,pline)!=SDP_PARSE_SUCCESS) {
              printandlog(pline->index, MISC, NULL, "Droping Call, no media information found in APM message",0);
							disconnect_video_call(pline->index);
          }
     }
#endif
}
/****************************************************************
 *			NAME: process_call_related_outbound_event(struct channel *pline, METAEVENT *metaeventp)
 * DESCRIPTION: Function to process outbound GlobalCall event
 *		 INPUTS: pline - pointer to entry in port table
 *					metaeventp - pointer to the metaevent structure
 *		RETURNS: NA
 *	  CAUTIONS: Tightly coupled with rocess_event() for pre-processing
 *					Assumes event logging already done
 *					Disconnected event already handled
 ****************************************************************/
static void process_call_related_outbound_event(struct channel *pline, METAEVENT *metaeventp)
{
				int		evttype, index;
				int		unexpected_event = 0;					/* till proven otherwise */
				char		str[MAX_STRING_SIZE];
				int		callindex;

				evttype = metaeventp->evttype;
				index = pline->index;

				/* Retrieve CRN */
				callindex = existCRN(metaeventp->crn, pline);
				if (callindex == OUT_OF_RANGE) {
								/* CRN not found, so log error and return from function. */
								sprintf(str, "CRN not found");
								printandlog(pline->index, MISC_WITH_NL, NULL, str, 0);
								return;
				}
				sprintf(str,"Rx Event Type 0x%x",evttype);	
				printandlog(index, MISC_WITH_NL, NULL, str, 0);

				switch (pline->call[callindex].call_state)
				{
								/************************************/
								/* first come the call setup states */
								/************************************/
								case GCST_NULL:
												{
																switch (evttype)
																{
																				case GCEV_ALERTING:						/* just change the state to alerting - no other action is necessary */
																								pline->call[callindex].call_state = GCST_ALERTING;
#ifdef USE_VIDEO
																								if(connectDevicesForMakeCall(metaeventp,pline)==SDP_ERROR) {
																												return;
																								}
#endif
																								break;

																				case GCEV_PROCEEDING:					/* just change the state to proceeding - no other action is necessary */
																								pline->call[callindex].call_state = GCST_PROCEEDING;
#ifdef USE_VIDEO
																								if(connectDevicesForMakeCall(metaeventp,pline)==SDP_ERROR) {
																												return;
																								}
#endif
																								break;

																								/* This case can happen in some technologies/protocols if the inbound side */
																								/* answered the call without ringing (i.e. if acceptcall was issued, */
																								/* then the rings parameter in both acceptcall and anwercall was 0 */
																								/* if only answercall was issued, then the # of rings was 0 */
																								/* The internal GC call state should be Dialing, but the dialing event */
																								/* is disabled by default and we did not enable it */
																				case GCEV_CONNECTED:
																				{
#ifdef USE_VIDEO
																								if(connectDevicesForMakeCall(metaeventp,pline)==-1) {
																												return;
																								}
#endif
																								process_connected_event(pline, metaeventp);
																								break;
																				}
																				default:
																								unexpected_event = 1;
																								break;
																}
																break;
												}

								case GCST_ALERTING:
												{
																if (evttype == GCEV_CONNECTED) {
															#ifdef USE_VIDEO
																				if(connectDevicesForMakeCall(metaeventp,pline)==-1) {
																								return;
																				}
															#endif
																				process_connected_event(pline, metaeventp);
																} else {
																				unexpected_event = 1;
																}
																break;
												}

								case GCST_PROCEEDING:
								{
																switch (evttype)
																{
																				case GCEV_ALERTING:						/* just change the state to alerting - no other action is necessary */
																				{
																						#ifdef USE_VIDEO
																								if(connectDevicesForMakeCall(metaeventp,pline)==-1) {
																												return;
																								}
																						#endif
																								pline->call[callindex].call_state = GCST_ALERTING;
																								break;
																				}
																				case GCEV_CONNECTED:
																				{
																						#ifdef USE_VIDEO
																								if(connectDevicesForMakeCall(metaeventp,pline)==-1) {
																												return;
																								}
																						#endif
																								process_connected_event(pline, metaeventp);
																								break;
																				}
																				default:
																								unexpected_event = 1;
																								break;
																}
																break;
												}

												/*************************************/
												/* now come the call teardown states */
												/*************************************/
								case GCST_CONNECTED:
												{
																/*************************************************/
																/* Depending upon your application, you may or	 */
																/* may not have application specific event		 */
																/* handling here for when in the connected state */
																/* You could also have your app specific code	 */
																/* in process_event()									 */
																/*************************************************/
																if (evttype == GCEV_SIP_ACK_OK) {
																			break;
																}
																if (evttype != GCEV_DROPCALL) {			/* recall the outbound side issues a gc_DropCall() */
																				unexpected_event = 1;					/* for the purposes of this demo */
																				break;
																}
																/* FALL THROUGH ON PURPOSE since this is one of the 2 ways of reaching Idle */
																/* the other being a disconnected event */
												}

								case GCST_DISCONNECTED:
												{
																if (evttype == GCEV_DROPCALL) {
																				pline->call[callindex].dropcall_active = NO;

																				/* Call ReleaseCall */
																				pline->call[callindex].call_state = GCST_IDLE;

																				if (gc_ReleaseCallEx(pline->call[callindex].crn, EV_ASYNC) != GC_SUCCESS) {
																								sprintf(str, "gc_ReleaseCallEx(crn=0x%lx, EV_ASYNC) Failed", pline->call[callindex].crn);
																								printandlog(index, GC_APIERR, NULL, str, 0);
																								exitdemo(1);
																				}
																				sprintf(str, "gc_ReleaseCallEx(crn=0x%lx, EV_ASYNC) Success", pline->call[callindex].crn);
																				printandlog(index, GC_APICALL, NULL, str, 0);
																				fflush(stdout);
																				break;
																} else {
																				unexpected_event = 1;
																}
																break;
												}

								case GCST_IDLE:
												{
																if (evttype == GCEV_RELEASECALL) {
																				pline->call[callindex].call_state = GCST_NULL;
																				pline->call[callindex].crn = 0;
																				pline->makecall_active = NO;					/* makecall is no longer active */
																				port[index].bBusy=false;
																				resetVariables(index);
																				if(pline->intercall_delay)  /* Set Inter-call Delay if specified */
																				{
																								time(&pline->intercall_time);
																								pline->intercall_time += pline->intercall_delay;
																				}
#ifndef USE_VIDEO
																				gc_demo_makecall(index);
#endif
																} else {
																				unexpected_event = 1;
																}
																break;
												}

								default:
												{
																/* Should never get here */
																printandlog(index, MISC, NULL, "Invalid state in process_call_related_outbound_event()", 0); 
																break;
												}
				}

				if (unexpected_event && (evttype != GCEV_UNBLOCKED)) {
								printandlog(index, GC_RESULT_INFO, metaeventp, "Event was unexpected in the current state", 0);
								printandlog(index, EVENT, NULL, GCEV_MSG(evttype), 0);
								printandlog(index, STATE, NULL, " ", callindex);
				}
}

/****************************************************************
 *			NAME: process_connected_event(struct channel *pline, METAEVENT *metaeventp)
 * DESCRIPTION: Function to process a connected event
 *		 INPUTS: pline - pointer to entry in port table
 *		RETURNS: NA
 *	  CAUTIONS: Assumes event logging already done
 *					Recall that this is only done for the outbound side
 *					as the GCEV_CONNECTED event is an outbound side event only
 ****************************************************************/
static void process_connected_event(struct channel *pline, METAEVENT *metaeventp)
{
				char		str[MAX_STRING_SIZE];
				char		value;										/* for connect type */
				int		rc;
				int		callindex;

				/* Retrieve CRN */
				callindex = existCRN(metaeventp->crn, pline);
				if (callindex == OUT_OF_RANGE) {
								/* CRN not found, so log error and return from function. */
								sprintf(str, "CRN not found");
								printandlog(pline->index, MISC_WITH_NL, NULL, str, 0);
								return;
				}

				/* Obtain connect type reason if not H323 nor SIP */
				if (	(pline->techtype != H323) && (pline->techtype != SIP) && 
							(pline->techtype != SS7) && (pline->techtype != BICC) ) {
								/* print connect type */
								rc = gc_GetCallInfo(pline->call[callindex].crn, CONNECT_TYPE, &value);
								if (rc != GC_SUCCESS) {
												if (rc == EGC_UNSUPPORTED) {
																strcpy(str, "call connected - gc_GetCallInfo(CONNECT_TYPE) not supported");
												} else {
																strcpy(str, "call connected - gc_GetCallInfo(CONNECT_TYPE) error\n");
																sprintf(str, "gc_GetCallInfo(crn=0x%lx, CONNECT_TYPE, &value) failed", pline->call[callindex].crn);
																printandlog(pline->index, GC_APIERR, NULL, str, 0);
												}
								} else {
												switch (value)
												{
																case GCCT_CAD :
																				strcpy(str, "call connected - cadence break");
																				break;

																case GCCT_LPC :
																				strcpy(str, "call connected - loop current drop");
																				break;

																case GCCT_PVD :
																				strcpy(str, "call connected - voice detection");
																				break;

																case GCCT_PAMD :
																				strcpy(str, "call connected - answering machine");
																				break;

																case GCCT_FAX :
																				strcpy(str, "call connected - FAX machine");
																				break;

																case GCCT_NA :
																				strcpy(str, "call connected - call progress not applicable");
																				break;

																default :
																				strcpy(str, "call connected - unknown connect type");
																				break;
												}
								}
								printandlog(pline->index, MISC, NULL, str, 0); 
				}
				pline->call[callindex].call_state = GCST_CONNECTED;
				pline->numb_calls++;							/* Increment the total calls on this device */
				out_calls++;									/* Increment the total outbound calls on all devices */
				outbound_application(pline);				/* call your application specific code here */
}

/****************************************************************
 *			NAME: enable_alarm_notification(struct channel *pline)
 * DESCRIPTION: Enables all alarms notification for pline
 *					Also fills in pline->mediah
 *		  INPUT: pline - pointer to channel data structure
 *		RETURNS: None - exits if error
 *	  CAUTIONS: Does not sanity checking as to whether or not the technology supports
 *				   alarms - assumes caller has done that already
 ****************************************************************/
static void	enable_alarm_notification(struct channel *pline)
{
				char		str[MAX_STRING_SIZE];
				int		alarm_ldev;						/* linedevice that alarms come on */


				alarm_ldev = pline->ldev;				/* until proven otherwise */
				if (pline->techtype == H323) {
								/* Recall that the alarms for IP come on the media device, not the network device */
								if (gc_GetResourceH(pline->ldev, &alarm_ldev, GC_MEDIADEVICE) != GC_SUCCESS) {
												sprintf(str, "gc_GetResourceH(linedev=%ld, &alarm_ldev, GC_MEDIADEVICE) Failed", pline->ldev);
												printandlog(pline->index, GC_APIERR, NULL, str, 0);
												exitdemo(1);
								}
								pline->mediah = alarm_ldev;			/* save for later use */
				}

				if (gc_SetAlarmNotifyAll(alarm_ldev, ALARM_SOURCE_ID_NETWORK_ID, ALARM_NOTIFY) != GC_SUCCESS) {
								sprintf(str, "gc_SetAlarmNotifyAll(linedev=%ld, ALARM_SOURCE_ID_NETWORK_ID, \
												ALARM_NOTIFY) Failed", pline->ldev);
								printandlog(pline->index, GC_APIERR, NULL, str, 0);
								printandlog(pline->index, MISC, NULL, "You may ignore this error - \
																most likely the underlying technology does not support alarms yet", 0);
				} else {
								sprintf(str, "gc_SetAlarmNotifyAll(linedev=%ld, ALARM_SOURCE_ID_NETWORK_ID, ALARM_NOTIFY) PASSED", pline->ldev);
								printandlog(pline->index, MISC, NULL, str, 0);
				}
}


/****************************************************************
 *			NAME: open_all_devices(void)
 * DESCRIPTION: Routine calls the specific technology function, which
 *					opens all the line devices. 
 *		  INPUT: None
 *		RETURNS: None - exits if error
 *	  CAUTIONS: none
 ****************************************************************/
static void open_all_devices(void)
{
	int	index;

	for (index = 0; index < num_devices; index++) { 
	 		resetVariables(index);
			open_device(index);
			//sr_setparm(port[index].ldev, SR_USERCONTEXT, (void *)&port[index].index);
	}
#ifdef USE_VIDEO
	if(InitM3g(BOARD_DEVICE,(num_devices-num_sip_devices))<0) {
							printf("Can't continue there is an error while initializing video devices \n");
							exit(-1);	
	}
	#ifdef USE_CNF
				gc_demo_open_cnf_board();
	#endif
#endif
}
/****************************************************************
 *			NAME: open_devices(int index)
 * DESCRIPTION: Routine calls the specific technology function, which
 *					opens all specific line device. 
 *		  INPUT: index - index into port data structure
 *		RETURNS: None - exits if error
 *	  CAUTIONS: Since multiple paths here, will init run-time fields of port data structure
 ****************************************************************/
static void open_device(int index)
{
				int		callindex;

				/* reset run-time variables that might now be wrong */
				port[index].ldev = 0;

				/* reset the call related information */
				for (callindex = 0; callindex < MAX_CALLS; callindex++) {
								port[index].call[callindex].crn = 0;
								port[index].call[callindex].call_state = GCST_NULL;
								port[index].call[callindex].dropcall_active = NO;	
				}
#ifdef USE_VIDEO
				port[index].callType=CALL_TYPE_VIDEO;
				port[index].bSdpReceived=false;
#else
				port[index].callType=CALL_TYPE_AUDIO;
#endif
				port[index].index = index;
				port[index].blocked = YES;
				port[index].waitcall_active = NO;	
				port[index].makecall_active = NO;	
				port[index].resetlinedev_active = NO;	
				port[index].resetlinedev_required_after_unblocked = NO;	
				port[index].dropcall_time = 0;
				port[index].makecall_time = 0;
				port[index].intercall_time = 0;

				if ((strcmp(port[index].protname, "isdn") == 0) ||
												(strcmp(port[index].protname, "ISDN") == 0)) {
								gc_demo_open_isdn_channel(index);
				} else if ( (strstr(port[index].protname, "ss7") != NULL) ||
												(strstr(port[index].protname, "SS7") != NULL) ||
												(strstr(port[index].protname, "bicc") != NULL) ||
                        (strstr(port[index].protname, "BICC") != NULL) ) {
								gc_demo_open_ss7_channel(index);
				} else if ((strstr(port[index].protname, "_AN_") != NULL) ||
												(strstr(port[index].protname, "_an_") != NULL)) {
								gc_demo_open_analog_channel(index);
				} else if (strstr(port[index].protname, "H323") != NULL) {
								gc_demo_open_H323_channel(index);
				} else if (strstr(port[index].protname, "SIP") != NULL) {
								gc_demo_open_SIP_channel(index);
				} else {								/* assume everything else is E1 or T1 */
								/* if wrong, then gc_OpenEx() will fail */
								gc_demo_open_E1orT1_channel(index);
				}

				openMediaDevice(index);

}

/****************************************************************
 *   NAME: void getNetworkHandle(int index)
 *   DESCRIPTION: get the network handle from line device
 *   INPUT: index
 *   RETURNS: None
 *   CAUTIONS: NA
 ****************************************************************/
static void getNetworkHandle(int index)
{
				char str[MAX_STRING_SIZE];

				if ( gc_GetResourceH(port[index].ldev, &port[index].networkH,GC_NETWORKDEVICE) != GC_SUCCESS ) {
								sprintf(str,"gc_GetResourceH(%s) failed",port[index].netdevice);
								// Handle Error
				} else {
								sprintf(str,"gc_GetResourceH(%s) successful",port[index].netdevice);
				}
				printandlog(index,GC_APICALL,NULL,str,0);
				sr_setparm(port[index].networkH, SR_USERCONTEXT, (void *)&port[index].index);
}
#ifndef USE_VIDEO
/****************************************************************
 *   NAME: void getNetworkHandle(int index)
 *   DESCRIPTION: get the network handle from line device
 *   INPUT: index
 *   RETURNS: None
 *   CAUTIONS: NA
 ****************************************************************/
static void getIpmHandle(int index)
{
				char str[MAX_STRING_SIZE];

				if ( gc_GetResourceH(port[index].ldev, &port[index].deviceInfo[IPM].devH,GC_MEDIADEVICE) != GC_SUCCESS ) {
								sprintf(str,"gc_GetResourceH(%s) failed",port[index].deviceInfo[IPM].device_name);
				} else {
								sprintf(str,"gc_GetResourceH(%s) successful",port[index].deviceInfo[IPM].device_name);
				}
				printandlog(index,GC_APICALL,NULL,str,0);
				sr_setparm(port[index].deviceInfo[IPM].devH, SR_USERCONTEXT, (void *)&port[index].index);
}
#endif

/****************************************************************
 *   NAME:   void openMediaDevice(int index)
 *   DESCRIPTION:  open dx device and get the handle to play audio
 prompt
 *   INPUT: index
 *   RETURNS: None
 *   CAUTIONS: NA
 ****************************************************************/
static void openMediaDevice(int index)
{
				char str[MAX_STRING_SIZE];
				port[index].dxH=dx_open(port[index].dxDeviceName,0);
				if(port[index].dxH>0) {
								sprintf(str,"dx_open(%s) successfull",port[index].dxDeviceName);
								sr_setparm(port[index].dxH, SR_USERCONTEXT, (void *)&port[index].index);
				}
				else {
								sprintf(str,"dx_open(%s) failed",port[index].dxDeviceName);
				}
				printandlog(index,DX,NULL,str,0);

#ifdef USE_VIDEO
				if( (port[index].techtype==SIP) || (port[index].techtype==H323) || (port[index].techtype==BICC)) {
								port[index].deviceInfo[IPM].devH=ipm_Open(port[index].deviceInfo[IPM].device_name,NULL,EV_ASYNC);
								if(port[index].deviceInfo[IPM].devH>0) {
										sprintf(str,"ipm_Open(%s) successfull",port[index].deviceInfo[IPM].device_name);
										sr_setparm(port[index].deviceInfo[IPM].devH, SR_USERCONTEXT, (void *)&port[index].index);
								}
								else {
										sprintf(str,"ipm_Open(%s) failed",port[index].deviceInfo[IPM].device_name);
								}
								printandlog(index,IPM,NULL,str,0);
				}
#endif
}
#ifdef USE_VIDEO
static void enableIpmEvents(int index)
{
	 char str[MAX_STRING_SIZE];
  const int nEvents = 5;
	 eIPM_EVENT event[nEvents];
		event[0]=EVT_RFC2833;
		event[1]=EVT_INIT_RECEIVED;
 		event[2]=EVT_PROCEDURE_DONE;
		event[3]=EVT_SEND_FAILED;
		event[4]=EVT_NACK_SENT;
   if ( ipm_EnableEvents(port[index].deviceInfo[IPM].devH, event, nEvents, EV_ASYNC) < 0 ) {
     sprintf(str, "ipm_EnableEvent() failed on %s", port[index].deviceInfo[IPM].device_name);
   } else {
     sprintf(str,"ipm_EnableEvent() Success on %s", port[index].deviceInfo[IPM].device_name);
   }
	 printandlog(index,IPM,NULL,str,0);
}
static void setDTMFMode(int index)
{
				char str[MAX_STRING_SIZE];
				IPM_PARM_INFO       parmInfo;
				eIPM_DTMFXFERMODE   value;
				int                 PLType  = 0;

				value               	= DTMFXFERMODE_RFC2833;
				parmInfo.eParm      	= PARMCH_DTMFXFERMODE;
				parmInfo.pvParmValue	= &value;
				if ( ipm_SetParm(port[index].deviceInfo[IPM].devH, &parmInfo, EV_SYNC) < 0 )
				{
								sprintf(str,"ERROR: ipm_SetParm() on %s",port[index].deviceInfo[IPM].device_name);
								printandlog(index, IPM, NULL, str, 0);
				}

				PLType                  = 101;
				parmInfo.eParm          = PARMCH_RFC2833EVT_TX_PLT;
				parmInfo.pvParmValue    = &PLType;
				if ( ipm_SetParm(port[index].deviceInfo[IPM].devH, &parmInfo, EV_SYNC) < 0 )
				{
								sprintf(str,"ERROR: ipm_SetParm() on %s", port[index].deviceInfo[IPM].device_name);
								printandlog(index, IPM, NULL, str, 0);
				}
				PLType              = 101;
				parmInfo.eParm      = PARMCH_RFC2833EVT_RX_PLT;
				parmInfo.pvParmValue = &PLType;
				if ( ipm_SetParm(port[index].deviceInfo[IPM].devH, &parmInfo, EV_SYNC) < 0 )
				{
								sprintf(str,"ERROR: ipm_SetParm() on %s", port[index].deviceInfo[IPM].device_name);
								printandlog(index, IPM, NULL, str, 0);
				}
}
#endif
/****************************************************************
 *   NAME:   void routeDtiAndDx(int index)
 *   DESCRIPTION: route dti and dx device, required in audio call
 *   INPUT: index
 *   RETURNS: None
 *   CAUTIONS: NA
 ****************************************************************/
static void routeDtiAndDx(int index)
{
				if(port[index].bDxAndDtiConnected==true) {
					return;
				}
				char str[MAX_STRING_SIZE];
				long         ts = 0;
				SC_TSINFO    scts;

				port[index].bDxAndDtiConnected=true;

				scts.sc_numts = 1;
				scts.sc_tsarrayp = &ts;
				if ( gc_GetXmitSlot(port[index].ldev, &scts) < 0 ) {
								sprintf(str,"gc_GetXmitSlot(%s) failed",port[index].netdevice);
								printandlog(index,GC_APICALL,NULL,str,0);
				}
				if ( dx_listen(port[index].dxH, &scts) < 0 ) {
								sprintf(str,"dx_listen(%s) failed",port[index].dxDeviceName);
								printandlog(index,DX,NULL,str,0);
				} else {
								sprintf(str,"dx_listen(%s) success",port[index].dxDeviceName);
								printandlog(index,DX,NULL,str,0);
				}
				ts = 0;
				scts.sc_numts = 1;
				scts.sc_tsarrayp = &ts;
				if ( dx_getxmitslot(port[index].dxH, &scts) < 0 ) {
								sprintf(str,"dx_getxmitslot(%s) failed",port[index].dxDeviceName);
								printandlog(index,DX,NULL,str,0);
				}
				if ( gc_Listen(port[index].ldev, &scts, EV_SYNC) < 0 ) {
								sprintf(str,"gc_Listen(%s) failed",port[index].netdevice);
								printandlog(index,GC_APICALL,NULL,str,0);
				}
				else {
								sprintf(str,"gc_Listen(%s) success",port[index].netdevice);
								printandlog(index,GC_APICALL,NULL,str,0);
				}
}
/****************************************************************
 *   NAME:   void unRouteDtiAndDx(int index)
 *   DESCRIPTION: unroute dti and dx device if route earlier for audio call
 *   INPUT: index
 *   RETURNS: None
 *   CAUTIONS: NA
 ****************************************************************/
static void unRouteDtiAndDx(int index)
{
        if(port[index].bDxAndDtiConnected==false) {
          return;
        }

				char str[MAX_STRING_SIZE];
				port[index].bDxAndDtiConnected=false;
				if ( dx_unlisten(port[index].dxH) < 0 ) {
								sprintf(str,"dx_unlisten(%s) failed",port[index].dxDeviceName);
								printandlog(index,DX,NULL,str,0);
				} else {
								sprintf(str,"dx_unlisten(%s) success",port[index].dxDeviceName);
								printandlog(index,GC_APICALL,NULL,str,0);
				}
				if ( gc_UnListen(port[index].ldev,EV_SYNC) < 0 ) {
								sprintf(str,"gc_UnListen(%s) failed",port[index].netdevice);
								printandlog(index,GC_APICALL,NULL,str,0);
				} else {
								sprintf(str,"gc_UnListen(%s) success",port[index].netdevice);
								printandlog(index,GC_APICALL,NULL,str,0);
				}
}

static void routeIpmAndDx(int index)
{
				char str[MAX_STRING_SIZE];
				long         ts = 0;
				SC_TSINFO    scts;

				scts.sc_numts = 1;
				scts.sc_tsarrayp = &ts;
				if ( ipm_GetXmitSlot(port[index].deviceInfo[IPM].devH, &scts, EV_ASYNC) < 0 ) {
								sprintf(str,"ipm_GetXmitSlot(%s) failed",port[index].deviceInfo[IPM].device_name);
								printandlog(index,IPM,NULL,str,0);
				}
				if ( dx_listen(port[index].dxH, &scts) < 0 ) {
								sprintf(str,"dx_listen(%s) failed",port[index].dxDeviceName);
								printandlog(index,DX,NULL,str,0);
				}
				ts = 0;
				scts.sc_numts = 1;
				scts.sc_tsarrayp = &ts;
				if ( dx_getxmitslot(port[index].dxH, &scts) < 0 ) {
								sprintf(str,"dx_getxmitslot(%s) failed",port[index].dxDeviceName);
								printandlog(index,DX,NULL,str,0);
				}
				if ( ipm_Listen(port[index].deviceInfo[IPM].devH, &scts, EV_SYNC) < 0 ) {
								sprintf(str,"ipm_Listen(%s) failed",port[index].deviceInfo[IPM].device_name);
								printandlog(index,IPM,NULL,str,0);
				}
				printandlog(index,DX,NULL,"dx_listen and ipm_Listen successful",0);
}

static void unRouteIpmAndDx(int index)
{
				char str[MAX_STRING_SIZE];

				if ( dx_unlisten(port[index].dxH) < 0 ) {
								sprintf(str,"dx_unlisten(%s) failed",port[index].dxDeviceName);
								printandlog(index,DX,NULL,str,0);
								exitdemo(1);
				}
				if ( ipm_UnListen(port[index].deviceInfo[IPM].devH,EV_ASYNC) < 0 ) {
								sprintf(str,"ipm_UnListen(%s) failed",port[index].deviceInfo[IPM].device_name);
								printandlog(index,IPM,NULL,str,0);
								exitdemo(1);
				}
}

/****************************************************************
 *			NAME: gc_demo_makecall(index)
 * DESCRIPTION: Routine calls the gc_Makecall() function if not blocked and there is not a call active
 *		  INPUT: index into port[] for this channel
 *		RETURNS: exits with exitdemo() if not successful
 *	  CAUTIONS: none
 ****************************************************************/
static void gc_demo_makecall(int index)
{
				char					str[MAX_STRING_SIZE];
				struct channel		*pline;
				GC_MAKECALL_BLK	*makecallblkp;				/* Variable for MAKECALL block */		
				int		callindex;
				char					*numberstr;

				pline = &port[index];
				if ( (pline->blocked == YES) || (pline->resetlinedev_active == YES)
												|| (interrupted == YES) ) {
								return;										/* cannot make a call until unblocked and completion of reset line device */
				}

				/* Find empty place in array to store CRN. */
				for (callindex = 0; callindex < MAX_CALLS; callindex++) {
								if (pline->call[callindex].crn == 0) {
												break;
								}
				}

				if (pline->makecall_active == YES) {
								printandlog(index, MISC, NULL, "Trying to make a call while makecall is already active\n", 0);
								return;
				}

				/* verify OK to do a makecall now */
				if ((pline->makecall_time) ||  (pline->intercall_time)){
								/* makecall or intercall timer active, cannot do just now */
								return;
				}

#ifdef USE_VIDEO
				if(pline->techtype != SIP && pline->callType==CALL_TYPE_VIDEO) {
								GC_IE_BLK   ie;
								IE_BLK      ieblk;

								ie.gclib = NULL;    // GlobalCall specific portion
								ie.cclib = &ieblk;  // CCLIB specific portion

								if(port[index].techtype==ISDN) {

												ieblk.length  = 10;							// increase length if using both LLC and BC

												ieblk.data[0] = (char)0x04;			// octet 1 IE type i.e. 0x04 = BC / 0x7c = LLC
												ieblk.data[1] = (char)0x03;			// octet 2 length of the IE
												ieblk.data[2] = (char)0x88;   	// octet 3 0x80 = speech / 0x88=unrestricted data
												ieblk.data[3] = (char)0x90;   	// octet 4 0x90 = 64 kbps packet mode
												ieblk.data[4] = (char)0xA6;   	// octet 5 0xA2 = uLaw / 0xA6 = video

												// if not able to place call with LLC setting then try using BC

												ieblk.data[5] = (char)0x7c;			// octet 1 IE type i.e. 0x04 = BC / 0x7c = LLC
												ieblk.data[6] = (char)0x03;			// octet 2 length of the IE
												ieblk.data[7] = (char)0x88;   	// octet 3 0x80 = speech  / 0x88=unrestricted data
												ieblk.data[8] = (char)0x90;   	// octet 4 0x90 = 64 kbps packet mode
												ieblk.data[9] = (char)0xA6;   	// octet 5 0xA2 = uLaw / 0xA6 = video

								} else if(port[index].techtype==SS7) {

												ieblk.length  = 12; // increase length if using both LLC and BC

												ieblk.data[0] = (char)0x03;   // octet 1 IE type i.e. 0x04 = BC / 0x7c = LLC
												ieblk.data[1] = (char)0x05;   // octet 2 length of the IE
												ieblk.data[2] = (char)0x7C;      // octet 4 0x90 = 64 kbps packet mode
												ieblk.data[3] = (char)0x03;      // octet 5 0xA2 = uLaw / 0xA6 = video
												ieblk.data[4] = (char)0x88;      // octet 3 0x80 = speech / 0x88=unrestricted data
												ieblk.data[5] = (char)0x90;      // octet 4 0x90 = 64 kbps packet mode
												ieblk.data[6] = (char)0xA6;      // octet 5 0xA2 = uLaw / 0xA6 = video

												// if not able to place call with LLC setting, then try using BC

												ieblk.data[7] = (char)0x1D;   // octet 1 IE type i.e. 0x04 = BC / 0x7c = LLC
												ieblk.data[8] = (char)0x03;   // octet 2 length of the IE
												ieblk.data[9] = (char)0x88;      // octet 3 0x80 = speech  / 0x88=unrestricted data
												ieblk.data[10] = (char)0x90;      // octet 4 0x90 = 64 kbps packet mode
												ieblk.data[11] = (char)0xA6;      // octet 5 0xA2 = uLaw / 0xA6 = video
								}

								if (gc_SetInfoElem(pline->ldev, &ie) != GC_SUCCESS) {
												printandlog(index, GC_APICALL, NULL, "can't make video call gc_SetInfoElem() for the channel failed", 0);
												return;
								} else {
												printandlog(index, GC_APICALL, NULL, "gc_SetInfoElem() for the channel success, Will make an outgoing Video call", 0);
								}
				}
// SKS if BICC	

        if (pline->techtype == BICC) 
        {
            buildOfferAnswerSdp(index);
            // SKS TODO
            S7_IE_BLK      ie_blk;
            GC_IE_BLK   gc_ie_blk;
            int length=0;

            length=buildApmMessage(index);

            ie_blk.length=length;

            memcpy(ie_blk.data,pline->biccApmMessage,length);
            int i;
            printf("\n In Hex\n");
            for(i=0;i<ie_blk.length;i++) {
                printf("%02x ",ie_blk.data[i]);
            }
            printf("\n\n\n\nnow character:\n");
            for(i=0;i<ie_blk.length;i++) {
                printf("%c",ie_blk.data[i]);
            }

            gc_ie_blk.gclib = NULL;
            gc_ie_blk.cclib = &ie_blk;


                if (gc_SetInfoElem(pline->ldev, &gc_ie_blk) != GC_SUCCESS) {
                        printandlog(index, GC_APICALL, NULL, "can't make BICC call gc_SetInfoElem() for the channel failed", 0);
                        return;
                } else {
                        printandlog(index, GC_APICALL, NULL, "gc_SetInfoElem() for the channel success, Will make an outgoing BICC call", 0);
                }


        }   // SKS BICC ends here			
#endif

				if(pline->techtype != SIP) {
								numberstr = pline->destination_num;
				} else {  /* else if SIP: */
#ifndef USE_VIDEO
								/* we are setting destination address in makecallblk */
								numberstr = NULL;
								/* and we are specifying coders on a per call basis */
								gc_set_channel_codecs(index);
#else
								numberstr = NULL;
								pline->gcParmBlk = NULL;
								buildOfferAnswerSdp(index);
								printandlog(index, GC_APICALL, NULL,pline->sdpText, 0);

								if( gc_util_insert_parm_ref_ex(&pline->gcParmBlk, IPSET_SDP, IPPARM_SDP_OFFER, strlen(pline->sdpText), (void *)pline->sdpText)!=0) {
												printandlog(index, GC_APIERR, NULL, "Error inserting SDP OFFER", 0);
												// TODO error handling
												return;
								}
								pline->makecallblk.gclib->ext_datap = pline->gcParmBlk;
#endif
				}

				/* if the makecall block has been initialized, use it */
				/* else use NULL for the makecall block pointer */
				makecallblkp = (pline->makecallblk.gclib != NULL) ? &pline->makecallblk : NULL;

				port[index].bBusy=true;

				if (gc_MakeCall(pline->ldev, &pline->call[callindex].crn, numberstr, makecallblkp, pline->makecall_timeout, EV_ASYNC) == GC_SUCCESS) {
								pline->makecall_active = YES;					/* indicate there is a call active */
								sprintf(str, "gc_MakeCall(linedev=%ld, numberstr=%s, mode=EV_ASYNC) Success", pline->ldev, pline->destination_num);
								printandlog(index, GC_APICALL, NULL, str, 0);
				} else {
								sprintf(str, "gc_MakeCall(linedev=%ld, numberstr=%s, mode=EV_ASYNC) Failed", pline->ldev, pline->destination_num);
								printandlog(index, GC_APIERR, NULL, str, 0);
								if (pline->techtype == ISDN) {
												sprintf(str, "Please verify setup in gc_demo_open_isdn_channel() is correct");
												printandlog(index, MISC, NULL, str, 0);
								}

								/* There are circumstances under which makecall will fail now, but will succeed later */
								/* so rather than exit the app, let's schedule a retry */
								time(&pline->makecall_time);
								pline->makecall_time += MIN_MAKECALL_RETRY_TIME + 1;
								sprintf(str, "Will retry the makecall in %d-%d seconds", MIN_MAKECALL_RETRY_TIME, MIN_MAKECALL_RETRY_TIME + 1);
								printandlog(index, MISC, NULL, str, 0);
				}
} /* End of Function */

static void process_non_gc_event(METAEVENT *metaevent)
{
				char str[MAX_STRING_SIZE];
				int index;
				int devH = (int)metaevent->evtdev;

				sr_getparm(devH, SR_USERCONTEXT, &index);

				sprintf(str,"SKS Rx Evt evttype=0x%lx|%ld on index=%d | devH=%d",metaevent->evttype,metaevent->evttype,index,devH);
//				printandlog(index,MISC,NULL,str,0);
				switch(metaevent->evttype)
				{
#ifdef USE_VIDEO
	#ifdef USE_CNF
								case CNFEV_CONF_CLOSED:
											printandlog(index,CNFEV,NULL,"CNFEV_CONF_CLOSED received",0);
											break;
								case CNFEV_OPEN:
								{
												printandlog(index,CNFEV,NULL,"CNFEV_OPEN received",0);
												break;
								}
								case CNFEV_OPEN_FAIL:
												printandlog(index,CNFEV,NULL,"CNFEV_OPEN_FAIL received",0);
												break;
								case CNFEV_RESET_DEVICES:
								{
												printandlog(index,CNFEV,NULL,"CNFEV_RESET_DEVICES received",0);
												if(devH==cnfBrdH) {
																enable_disable_cnf_brd_dev_events(true);
																gc_demo_open_conference();
												}
												break;
								}
								case CNFEV_RESET_DEVICES_FAIL:
												printandlog(index,CNFEV,NULL,"CNFEV_RESET_DEVICES_FAIL received",0);
												break;
								case CNFEV_OPEN_CONF:
								{
												printandlog(index,CNFEV,NULL,"CNFEV_OPEN_CONF received",0);
												enable_disable_cnf_dev_events(true);
												SetCnfLayout(eLB_LAYOUT_TYPE_9_1,eLB_LAYOUT_SIZE_QCIF);
												// SKS Open party for all the indexes
												gc_demo_open_cnf_party();
												break;
								}
								case CNFEV_OPEN_CONF_FAIL:
												printandlog(index,CNFEV,NULL,"CNFEV_OPEN_CONF_FAIL received",0);
												break;
								case CNFEV_CONF_OPENED:
											printandlog(index,CNFEV,NULL,"CNFEV_CONF_OPENED received",0);
											break;
								case CNFEV_OPEN_PARTY:
								{
												int partyIndex=(int) sr_getUserContext();
												CNF_OPEN_PARTY_RESULT *result=(CNF_OPEN_PARTY_RESULT*)metaevent->evtdatap;
												strcpy(port[partyIndex].deviceInfo[CNFPTY].device_name,result->szPartyName);
												sprintf(str,"CNFEV_OPEN_PARTY  received party=%s|cnfPartyH=%ld",result->szPartyName,result->PartyHandle);
												printandlog(partyIndex,CNFEV,NULL,str,0);
												SetPartyAttributes(partyIndex,false);
												getTxRxPortInfo(partyIndex,result->PartyHandle);
												break;
								}
								case CNFEV_OPEN_PARTY_FAIL:
											printandlog(index,CNFEV,NULL,"CNFEV_OPEN_PARTY_FAIL received",0);
											break;
								case CNFEV_SET_VIDEO_LAYOUT:
								{
												CNF_VIDEO_LAYOUT_INFO *cnfVideoLayoutInfo = (CNF_VIDEO_LAYOUT_INFO*)metaevent->evtdatap;
												cnfLayoutH=cnfVideoLayoutInfo->LayoutHandle;
												cnfLayoutSize=cnfVideoLayoutInfo->eLayoutSize;
												sprintf(str,"CNFEV_SET_VIDEO_LAYOUT received cnfLayoutH=%ld|layoutSize=%s",
																	cnfLayoutH,( cnfLayoutSize==0?"SQCIF":(cnfLayoutSize==1?"QCIF":"CIF") ) );
												printandlog(index,CNFEV,NULL,str,0);
												if (lb_GetRegionList(cnfLayoutH, cnfRegionList, &cnfRegionListCount) == MTK_ERROR) {
          											printandlog(0, MISC, NULL, "FAILURE: lb_GetRegionList()", 0);
								        } else {
											          printandlog(0, MISC, NULL, "SUCCESS: lb_GetRegionList()", 0);
        								}
												break;
								}
								case CNFEV_SET_VIDEO_LAYOUT_FAIL:
												printandlog(index,CNFEV,NULL,"CNFEV_SET_VIDEO_LAYOUT_FAIL received",0);
												break;
								case CNFEV_GET_VIDEO_LAYOUT:
								{
												CNF_VIDEO_LAYOUT_INFO *cnfLayout=(CNF_VIDEO_LAYOUT_INFO *)metaevent->evtdatap;
												cnfLayoutH=cnfLayout->LayoutHandle;
												cnfLayoutSize=cnfLayout->eLayoutSize;
												sprintf(str,"CNFEV_GET_VIDEO_LAYOUT received cnfLayoutH=%ld|layoutSize=%s",
																	cnfLayoutH,( cnfLayoutSize==0?"SQCIF":(cnfLayoutSize==1?"QCIF":"CIF") ) );
												printandlog(index,CNFEV,NULL,str,0);
												break;
								}
								case CNFEV_GET_VIDEO_LAYOUT_FAIL:
												printandlog(index,CNFEV,NULL,"CNFEV_GET_VIDEO_LAYOUT_FAIL received",0);
												break;
								case CNFEV_ADD_PARTY:
								{
												int partyIndex=(int) sr_getUserContext();
												printandlog(partyIndex,CNFEV,NULL,"CNFEV_ADD_PARTY received",0);
												GetCnfVisiblePartyList();
												break;
								}
								case CNFEV_ADD_PARTY_FAIL:
												printandlog(index,CNFEV,NULL,"CNFEV_ADD_PARTY_FAIL received",0);
												break;
								case CNFEV_PARTY_ADDED:
												printandlog(index,CNFEV,NULL,"CNFEV_PARTY_ADDED received",0);
												break;
								case CNFEV_REMOVE_PARTY:
								{
												int partyIndex=(int) sr_getUserContext();
												printandlog(partyIndex,CNFEV,NULL,"CNFEV_REMOVE_PARTY received",0);
												if(port[partyIndex].techtype==SIP)
												{
														routeUnrouteDevices(partyIndex,IPM,CNFPTY,UNROUTE,DMFL_TRANSCODE_ON);
												} else {
														stopH245(partyIndex);
												}
												// SKS refresh screen after removing party
												GetCnfVisiblePartyList();
												break;
								}
								case CNFEV_REMOVE_PARTY_FAIL:
												printandlog(index,CNFEV,NULL,"CNFEV_REMOVE_PARTY_FAIL received",0);
												break;
								case CNFEV_PARTY_REMOVED:
												printandlog(index,CNFEV,NULL,"CNFEV_PARTY_REMOVED received",0);
												break;
								case CNFEV_SET_ATTRIBUTE:
												printandlog(index,CNFEV,NULL,"CNFEV_SET_ATTRIBUTE received",0);
												break;
								case CNFEV_SET_ATTRIBUTE_FAIL:
												printandlog(index,CNFEV,NULL,"CNFEV_SET_ATTRIBUTE_FAIL received",0);
												break;
								case CNFEV_GET_ATTRIBUTE:
												printandlog(index,CNFEV,NULL,"CNFEV_GET_ATTRIBUTE received",0);
												break;
								case CNFEV_GET_ATTRIBUTE_FAIL:
												printandlog(index,CNFEV,NULL,"CNFEV_GET_ATTRIBUTE_FAIL received",0);
												break;
								case CNFEV_GET_PARTY_LIST:
								{
												CNF_PARTY_INFO *cnfPartyInfo=(CNF_PARTY_INFO*) metaevent->evtdatap;
												int i;
												for(i=0;i<cnfPartyInfo->unPartyCount;i++) {
													sprintf(str,"CNFEV_GET_PARTY_LIST received partyCount=%d|partyH[%d]=%ld",cnfPartyInfo->unPartyCount,i,cnfPartyInfo->pPartyList[i]);
													printandlog(index,CNFEV,NULL,str,0);
												}
												SetCnfPartyVisible(cnfPartyInfo->unPartyCount,cnfPartyInfo->pPartyList);	
												break;
								}
								case CNFEV_GET_DTMF_CONTROL:
												printandlog(index,CNFEV,NULL,"CNFEV_GET_DTMF_CONTROL received",0);
												break;
								case CNFEV_GET_DTMF_CONTROL_FAIL:
												printandlog(index,CNFEV,NULL,"CNFEV_GET_DTMF_CONTROL_FAIL received",0);
												break;
								case CNFEV_GET_DEVICE_COUNT:
												printandlog(index,CNFEV,NULL,"CNFEV_GET_DEVICE_COUNT received",0);
												break;
								case CNFEV_GET_DEVICE_COUNT_FAIL:
												printandlog(index,CNFEV,NULL,"CNFEV_GET_DEVICE_COUNT_FAIL received",0);
												break;
								case CNFEV_GET_ACTIVE_TALKER:
												printandlog(index,CNFEV,NULL,"CNFEV_GET_ACTIVE_TALKER received",0);
												break;
								case CNFEV_ENABLE_EVENT:
												printandlog(index,CNFEV,NULL,"CNFEV_ENABLE_EVENT received",0);
												break;
								case CNFEV_GET_VISIBLE_PARTY_LIST:
								{
												printandlog(index,CNFEV,NULL,"CNFEV_GET_VISIBLE_PARTY_LIST received",0);
												CNF_VISIBLE_PARTY_LIST *cnfVisiblePartyList=(CNF_VISIBLE_PARTY_LIST *)metaevent->evtdatap;
												int nCtr;
												for(nCtr=0;nCtr<cnfVisiblePartyList->unCount;nCtr++) {
															sprintf(str,"VisiblePartyCount=%d|PartyHandle=%ld|RegionHandle=%ld",cnfVisiblePartyList->unCount,
																cnfVisiblePartyList->pPartyList[nCtr].PartyHandle,cnfVisiblePartyList->pPartyList[nCtr].RegionHandle);
															printandlog(index,CNFEV,NULL,str,0);
												}
//												if(cnfVisiblePartyList->unCount < cnfRegionListCount) {
														GetCnfPartyList();
//                        } else {
//                            sprintf(str,"Can't make more than %d parties visible into conference, try to increase region count",cnfRegionListCount);
//                        }
//												printandlog(index,CNFEV,NULL,str,0);
												break;
								}
								case CNFEV_GET_VISIBLE_PARTY_LIST_FAIL:
												printandlog(index,CNFEV,NULL,"CNFEV_GET_VISIBLE_PARTY_LIST_FAIL received",0);
												break;
								case CNFEV_SET_VISIBLE_PARTY_LIST:
								{
												printandlog(index,CNFEV,NULL,"CNFEV_SET_VISIBLE_PARTY_LIST received",0);
												CNF_VISIBLE_PARTY_LIST *cnfVisiblePartyList=(CNF_VISIBLE_PARTY_LIST *)metaevent->evtdatap;
                        int nCtr;
                        for(nCtr=0;nCtr<cnfVisiblePartyList->unCount;nCtr++) {
                              sprintf(str,"VisiblePartyCount=%d|PartyHandle=%ld|RegionHandle=%ld",cnfVisiblePartyList->unCount,
                                cnfVisiblePartyList->pPartyList[nCtr].PartyHandle,cnfVisiblePartyList->pPartyList[nCtr].RegionHandle);
                              printandlog(index,CNFEV,NULL,str,0);
                        }
												break;
								}
								case CNFEV_SET_VISIBLE_PARTY_LIST_FAIL:
								{
												printandlog(index,CNFEV,NULL,"CNFEV_SET_VISIBLE_PARTY_LIST_FAIL received",0);
												break;
								}
	#endif								

								case M3GEV_REMOTE_VENDORID_RCVD:
								{
												M3G_VENDORID_INFO *vendorID=(M3G_VENDORID_INFO*)metaevent->evtdatap;
												sprintf(str,"M3GEV_REMOTE_VENDORID_RCVD evt on %d Vendor Prod=%s|Ver=%s",
																	devH,vendorID->productNumber.octet,vendorID->versionNumber.octet);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								}
								case M3GEV_H245_UII_RCVD:
								{
												M3G_H245_UII *h245Uii=(M3G_H245_UII*)metaevent->evtdatap;
												sprintf(str,"Rx M3GEV_H245_UII_RCVD evt on handle=%d NoOfDigit=%d Digit=%s",devH,h245Uii->numDigits,h245Uii->digitBuffer);
												printandlog(index,M3GEV,NULL,str,0);
#ifdef USE_CNF
												if(h245Uii->digitBuffer[0]==48 && port[index].bPartyIntoCnf==false) {
													stopPlayRecord(index,false);
													stopPlayRecord(index,true);
													modifyMedia(index,M3G_E_IDLE);
												}
#endif
												break;
								}
								case M3GEV_H245_MISC_CMD_RCVD:
												{
																sprintf(str,"M3GEV_H245_MISC_CMD_RCVD Event Received on handle=%d",devH);
																printandlog(index,M3GEV,NULL,str,0);
																break;
												}
								case M3GEV_OPEN_CMPLT:
												{
														sprintf(str,"M3GEV_OPEN_CMPLT Event Received on handle=%d",devH);
														printandlog(index,M3GEV,NULL,str,0);
														if(devH==port[index].deviceInfo[M3G].devH) {
								 								resetM3gDevice(index,devH);
														}
														break;
												}
								case M3GEV_OPEN_FAIL:
												sprintf(str,"M3GEV_OPEN_FAIL Event Received on Handle=%d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_RESET_CMPLT:
												{
														sprintf(str,"M3GEV_RESET_CMPLT Event Received on device Handle(%d)",devH);
														printandlog(index,M3GEV,NULL,str,0);
														getLocalCapability(index,port[index].deviceInfo[M3G].devH);
														getLocalCapability(index,port[index].deviceInfo[M3GVID].devH);
														getLocalCapability(index,port[index].deviceInfo[M3GAUD].devH);
														break;
												}
								case M3GEV_RESET_FAIL:
												sprintf(str,"M3GEV_RESET_FAIL Event Received on Handle=%d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_SET_PARM_CMPLT:
												{
																sprintf(str,"M3GEV_SET_PARM_CMPLT Event Received on Handle=%d",devH);
																printandlog(index,M3GEV,NULL,str,0);
																break;
												}
								case M3GEV_SET_PARM_FAIL:
												sprintf(str,"M3GEV_SET_PARM_FAIL Event Received on Handle=%d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_ENABLE_EVENTS_CMPLT:
												sprintf(str,"M3GEV_ENABLE_EVENTS_CMPLT Event Received on Handle=%d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_ENABLE_EVENTS_FAIL:
												sprintf(str,"M3GEV_ENABLE_EVENTS_FAIL Event Received on Handle=%d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_DISABLE_EVENTS_CMPLT:
												sprintf(str,"M3GEV_DISABLE_EVENTS_CMPLT Event Received on Handle=%d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_DISABLE_EVENTS_FAIL:
												sprintf(str,"M3GEV_DISABLE_EVENTS_FAIL Event Received on Handle=%d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;

								case M3GEV_GET_LOCAL_CAPS_CMPLT:
												{
																port[index].localCapCtr++;
																sprintf(str,"M3GEV_GET_LOCAL_CAPS_CMPLT_%d Evt Rcvd on Handle=%d", port[index].localCapCtr,devH);
																printandlog(index,M3GEV,NULL,str,0);
																storeLocalCapability(index,devH,(M3G_CAPS_LIST*)metaevent->evtdatap);
																if(port[index].localCapCtr==3) {
																				setLocalCapability(index);
																}
																break;
												}
								case M3GEV_GET_LOCAL_CAPS_FAIL:
												sprintf(str,"M3GEV_GET_LOCAL_CAPS_FAIL Event Received on Handle=%d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_SET_TCS_CMPLT:
								{
												sprintf(str,"M3GEV_SET_TCS_CMPLT Event Received on Handle=%d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												getTxRxPortInfo(index,port[index].deviceInfo[M3GAUD].devH);
												getTxRxPortInfo(index,port[index].deviceInfo[M3GVID].devH);
												getTxRxPortInfo(index,port[index].deviceInfo[M3G].devH);
												break;
								}
								case M3GEV_SET_TCS_FAIL:
												sprintf(str,"M3GEV_SET_TCS_FAIL Event Received on Handle=%d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case DMEV_CONNECT:
												sprintf(str,"DMEV_CONNECT Event Received on %d",devH);
												printandlog(index,DMEV,NULL,str,0);
												break;
								case DMEV_CONNECT_FAIL:
												sprintf(str,"DMEV_CONNECT_FAIL Event Received on %d",devH);
												printandlog(index,DMEV,NULL,str,0);
												break;
								case DMEV_DISCONNECT:
												sprintf(str,"DMEV_DISCONNECT Event Received on %d",devH);
												printandlog(index,DMEV,NULL,str,0);
												port[index].bM3gAndDtiConnected=false;
												port[index].dmevDisconnectCtr++;
												if(port[index].dmevDisconnectCtr==2) {
																disconnect_video_call(index);
												}
												break;
								case DMEV_DISCONNECT_FAIL:
												sprintf(str,"DMEV_DISCONNECT_FAIL Event Received on %d",devH);
												printandlog(index,DMEV,NULL,str,0);
												break;
								case DMEV_PORT_CONNECT:
								{
												port[index].dmevPortConnectCtr++;
												sprintf(str,"DMEV_PORT_CONNECT Event Received on %d totalCount=%d",devH,port[index].dmevPortConnectCtr);
												printandlog(index,DMEV,NULL,str,0);
												processDmevPortConnect(index);
												break;
								}
								case DMEV_PORT_CONNECT_FAIL:
												sprintf(str,"DMEV_PORT_CONNECT_FAIL Event Received on %d",devH);
												printandlog(index,DMEV,NULL,str,0);
												break;
								case DMEV_PORT_DISCONNECT:
								{
												port[index].dmevPortDisconnectCtr++;
												sprintf(str,"DMEV_PORT_DISCONNECT Event Received on %d totalCount=%d",devH,port[index].dmevPortDisconnectCtr);
												printandlog(index,DMEV,NULL,str,0);
												processDmevPortDisconnect(index);
												break;
								}
								case DMEV_PORT_DISCONNECT_FAIL:
												sprintf(str,"DMEV_PORT_DISCONNECT_FAIL Event Received on %d",devH);
												printandlog(index,DMEV,NULL,str,0);
												break;
								case DMEV_GET_TX_PORT_INFO:
												{
																sprintf(str,"DMEV_GET_TX_PORT_INFO Event Received on %d",devH);
																printandlog(index,DMEV,NULL,str,0);
																if(devH==port[index].deviceInfo[CNFPTY].devH) {
																				storeTxPortInfo(index,metaevent);
																				break;
																}
																port[index].txPortCtr++;
																storeTxPortInfo(index,metaevent);
																break;
												}
								case DMEV_GET_RX_PORT_INFO:
												{
																sprintf(str,"DMEV_GET_RX_PORT_INFO Event Received on %d",devH);
																printandlog(index,DMEV,NULL,str,0);
																if(devH==port[index].deviceInfo[CNFPTY].devH) {
																				storeRxPortInfo(index,metaevent);
																				break;
																}
																port[index].rxPortCtr++;
																storeRxPortInfo(index,metaevent);
																break;
												}
								case MMEV_OPEN:
												sprintf(str,"MMEV_OPEN Event Received on %d",devH);
												printandlog(index,MMEV,NULL,str,0);
												getTxRxPortInfo(index,port[index].deviceInfo[MM].devH);
												break;
								case MMEV_OPEN_FAIL:
												sprintf(str,"MMEV_OPEN_FAIL Event Received on %d",devH);
												printandlog(index,MMEV,NULL,str,0);
												break;
								case MMEV_RECORD_ACK:
								{
																sprintf(str,"MMEV_RECORD_ACK Event Received on %d",devH);
																printandlog(index,MMEV,NULL,str,0);
																if(port[index].bPlayActive==false) {
																		mmPlayRecordAudioVideoPrompt(index,false);
																}
																break;
								}
                case MMEV_RECORD_ACK_FAIL:
                {
                                sprintf(str,"MMEV_RECORD_ACK_FAIL Event Received on %d",devH);
                                printandlog(index,MMEV,NULL,str,0);
                                break;
                }
                case MMEV_RECORD:
                {
                                MM_PLAY_RECORD_CMPLT *mmRecordCmplt=(MM_PLAY_RECORD_CMPLT*)metaevent->evtdatap;
                                sprintf(str,"MMEV_RECORD Event Received on %d Reason=%s",devH,
                                    (mmRecordCmplt->details[0].Reason==EMM_TR_MAXTIME?"EMM_TR_MAXTIME":"EMM_TR_USERSTOP"));
                                printandlog(index,MMEV,NULL,str,0);
                                if(mmRecordCmplt->details[0].Reason==EMM_TR_MAXTIME) {
                                        mmPlayRecordAudioVideoPrompt(index,true);
                                }

                                break;
                }
                case MMEV_RECORD_FAIL:
                {
                                sprintf(str,"MMEV_RECORD_FAIL Event Received on %d",devH);
                                printandlog(index,MMEV,NULL,str,0);
                                break;
                }


								case MMEV_PLAY_ACK:
												{
																sprintf(str,"MMEV_PLAY_ACK Event Received on %d",devH);
																printandlog(index,MMEV,NULL,str,0);
																if(port[index].bRecordActive==false) {
																		mmPlayRecordAudioVideoPrompt(index,true);
																}
																break;
												}
								case MMEV_PLAY_ACK_FAIL:
												{
																sprintf(str,"MMEV_PLAY_ACK_FAIL Event Received on %d",devH);
																printandlog(index,MMEV,NULL,str,0);
																break;
												}
								case MMEV_PLAY:
												{
																MM_PLAY_RECORD_CMPLT *mmPlayCmplt=(MM_PLAY_RECORD_CMPLT*)metaevent->evtdatap;
																sprintf(str,"MMEV_PLAY Event Received on %d Reason=%s",devH,
																		(mmPlayCmplt->details[0].Reason==EMM_TR_EOF?"EMM_TR_EOF":"EMM_TR_USERSTOP"));
																printandlog(index,MMEV,NULL,str,0);
																if(mmPlayCmplt->details[0].Reason==EMM_TR_EOF) {
																				mmPlayRecordAudioVideoPrompt(index,false);				
																}
																break;
												}
								case MMEV_PLAY_FAIL:
												{
																sprintf(str,"MMEV_PLAY_FAIL Event Received on %d",devH);
																printandlog(index,MMEV,NULL,str,0);
																break;
												}
								case MMEV_STOP_ACK:
												sprintf(str,"MMEV_STOP_ACK Event Received on %d",devH);
												printandlog(index,MMEV,NULL,str,0);
												break;
								case MMEV_STOP_ACK_FAIL:
												sprintf(str,"MMEV_STOP_ACK_FAIL Event Received on %d",devH);
												printandlog(index,MMEV,NULL,str,0);
												break;
								case M3GEV_FRAMING_ESTABLISHED:
												sprintf(str,"M3GEV_FRAMING_ESTABLISHED on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_MSD_ESTABLISHED:
												{
																port[index].olcEvtMask+=M3GEV_MSD_ESTABLISHED;
																sprintf(str,"M3GEV_MSD_ESTABLISHED on %d (evtReqBefOlc=0x%lx|olcMask=0x%lx)",devH,eventsReqiredBefOlc,port[index].olcEvtMask);
																printandlog(index,M3GEV,NULL,str,0);
																if(port[index].olcEvtMask==eventsReqiredBefOlc) {
																				getMatchedCapabilityAndOpenLC(index);
																}
																break;
												}
								case M3GEV_MSD_FAILED:
												sprintf(str,"M3GEV_MSD_FAILED Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_REMOTE_TCS_RCVD:
												{
																port[index].olcEvtMask+=M3GEV_REMOTE_TCS_RCVD;
																sprintf(str,"M3GEV_REMOTE_TCS_RCVD on %d (evtReqBefOlc=0x%lx|olcMask=0x%lx)",devH,eventsReqiredBefOlc,port[index].olcEvtMask);
																printandlog(index,M3GEV,NULL,str,0);
																if(port[index].olcEvtMask==eventsReqiredBefOlc) {
																				getMatchedCapabilityAndOpenLC(index);
																}
																break;
												}
								case M3GEV_LOCAL_TCS_ACKD:
								{
											port[index].olcEvtMask+=M3GEV_LOCAL_TCS_ACKD;
											sprintf(str,"M3GEV_LOCAL_TCS_ACKD on %d (evtReqBefOlc=0x%lx|olcMask=0x%lx)",devH,eventsReqiredBefOlc,port[index].olcEvtMask);
											printandlog(index,M3GEV,NULL,str,0);
											if(port[index].olcEvtMask==eventsReqiredBefOlc) {
													getMatchedCapabilityAndOpenLC(index);
											}
											break;
								}
								case M3GEV_START_H245_CMPLT:
												sprintf(str,"M3GEV_START_H245_CMPLT Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_START_H245_FAIL:
												sprintf(str,"M3GEV_START_H245_FAIL Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_STOP_H245_CMPLT:
								{
												sprintf(str,"M3GEV_STOP_H245_CMPLT Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												processStopH245Complete(index);
												break;
								}
								case M3GEV_OPEN_LC_CMPLT:
								{
												sprintf(str,"M3GEV_OPEN_LC_CMPLT Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												M3G_REMOTE_OLCACK_RESP *remoteOlcAckResp=(M3G_REMOTE_OLCACK_RESP*)metaevent->evtdatap;
												if(remoteOlcAckResp->capabilityType==M3G_E_AUDIO_CAPABILITY) {
																port[index].openAudVidOlcCmplt+=AUDIO_LCN;
												}
												else if(remoteOlcAckResp->capabilityType==M3G_E_VIDEO_CAPABILITY) {
																port[index].openAudVidOlcCmplt+=VIDEO_LCN;
												}
												if(port[index].openAudVidOlcCmplt==AUDIO_VIDEO_LCN) {
																modifyMedia(index,M3G_E_TX);
												}
												break;
								}
								case M3GEV_OPEN_LC_FAIL:
												sprintf(str,"M3GEV_OPEN_LC_FAIL Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_REMOTE_OLC_RCVD:
												{
																sprintf(str,"M3GEV_REMOTE_OLC_RCVD Event Received on %d",devH);
																printandlog(index,M3GEV,NULL,str,0);
																sendOLCAck(index,(M3G_REMOTE_OLC_REQ*)metaevent->evtdatap);
																break;
												}
								case M3GEV_RESPOND_TO_LC_CMPLT:
												{
																M3G_LOGICAL_CHANNEL_NUMBER *logicalChannel=(M3G_LOGICAL_CHANNEL_NUMBER*)metaevent->evtdatap;
																sprintf(str,"M3GEV_RESPOND_TO_LC_CMPLT Event Received on %d lcn=%d",devH,*logicalChannel);
																printandlog(index,M3GEV,NULL,str,0);
																if(*logicalChannel==port[index].remoteAudioLogicalChannel) {
																				port[index].respondAudVidOlcCmplt+=AUDIO_LCN;
																}
																else if(*logicalChannel==port[index].remoteVideoLogicalChannel) {
																				port[index].respondAudVidOlcCmplt+=VIDEO_LCN;
																}
																if(port[index].respondAudVidOlcCmplt==AUDIO_VIDEO_LCN) {
																				modifyMedia(index,M3G_E_RX);
																}
																break;
												}
								case M3GEV_RESPOND_TO_LC_FAIL:
												sprintf(str,"M3GEV_RESPOND_TO_LC_FAIL Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_CLOSE_LC_CMPLT:
												sprintf(str,"M3GEV_CLOSE_LC_CMPLT Event Received on %d",devH);
												printandlog(index,M3G,NULL,str,0);
												break;
								case M3GEV_CLOSE_LC_FAIL:
												sprintf(str,"M3GEV_CLOSE_LC_FAIL Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_REMOTE_CLOSE_LC_RCVD:
												sprintf(str,"M3GEV_REMOTE_CLOSE_LC_RCVD Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_START_MEDIA_CMPLT:
												sprintf(str,"M3GEV_START_MEDIA_CMPLT Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_START_MEDIA_FAIL:
												sprintf(str,"M3GEV_START_MEDIA_FAIL Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_MODIFY_MEDIA_CMPLT:
								{
												M3G_E_DIRECTION *direction = (M3G_E_DIRECTION*)metaevent->evtdatap;
												if(*direction==M3G_E_IDLE) {
													port[index].modifyMediaIdleCtr+=1;
													port[index].modifyMediaCtr=0;
												} else {
													port[index].modifyMediaCtr+=1;
												}
												sprintf(str,"M3GEV_MODIFY_MEDIA_CMPLT Evt on %d dir=%s|devName=%s",devH,dirStr[*direction],ATDV_NAMEP(metaevent->evtdev));
												printandlog(index,M3GEV,NULL,str,0);
												processModifyMediaCmplt(index);
												break;
								}
								case M3GEV_MODIFY_MEDIA_FAIL:
												sprintf(str,"M3GEV_MODIFY_MEDIA_FAIL Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_STOP_MEDIA_CMPLT:
								{
												sprintf(str,"M3GEV_STOP_MEDIA_CMPLT Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												processStopMediaCmplt(index,devH);
												break;
								}
								case M3GEV_STOP_MEDIA_FAIL:
												sprintf(str,"M3GEV_STOP_MEDIA_FAIL Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_H245_MSD_EVT:
												sprintf(str,"M3GEV_H245_MSD_EVT Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_H245_MES_EVT:
												sprintf(str,"M3GEV_H245_MES_EVT Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_ENDSESSION_SENT:
												sprintf(str,"M3GEV_ENDSESSION_SENT Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												break;
								case M3GEV_ENDSESSION_RCVD:
												sprintf(str,"M3GEV_ENDSESSION_RCVD Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												stopMedia(index);
												stopH245(index);
												break;
								case M3GEV_FRAMING_LOST:
								{
												sprintf(str,"M3GEV_FRAMING_LOST Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												port[index].isCallDisconnectInProgress=true;
												stopMedia(index);
												stopH245(index);
												break;
								}
								case M3GEV_START_TRACE_CMPLT:
												sprintf(str,"M3GEV_START_TRACE_CMPLT Event Received on %d",devH);
												printandlog(index,M3GEV,NULL,str,0);
												bM3gTraceStarted=M3G_TRUE;
												break;
								case M3GEV_START_TRACE_FAIL:
												{
																M3G_E_ERROR_TYPE	*pError=(M3G_E_ERROR_TYPE *)metaevent->evtdatap;
																sprintf(str,"M3GEV_START_TRACE_FAIL Event Received on %d with err =%d",devH,(int)*pError);
																printandlog(index,M3GEV,NULL,str,0);
												}
												break;
// M3G event ends here
// IPM events
								case IPMEV_OPEN:
												{
																sprintf(str,"IPMEV_OPEN Event Received on %d",devH);
																printandlog(index,IPMEV,NULL,str,0);
																processIpmevOpen(index);
																break;
												}
								case IPMEV_GET_LOCAL_MEDIA_INFO:
												{
																sprintf(str,"IPMEV_GET_LOCAL_MEDIA_INFO Event Received on %d",devH);
																printandlog(index,IPMEV,NULL,str,0);
																IPM_MEDIA_INFO* pMediaInfo=(IPM_MEDIA_INFO*)metaevent->evtdatap;
																storeLocalMediaInfo(index,pMediaInfo);
																break;
												}
								case IPMEV_EVENT_ENABLED:
								{
												sprintf(str,"IPMEV_EVENT_ENABLED Event Received on %d",devH);
												printandlog(index,IPMEV,NULL,str,0);
												break;
								}
								case IPMEV_STOPPED:
								{
											sprintf(str,"IPMEV_STOPPED Event Received on %d",devH);
											printandlog(index,IPMEV,NULL,str,0);
											processIpmevStopped(index);
											break;
								}
								case IPMEV_ERROR:
								{ 
											sprintf(str,"IPMEV_ERROR Event Received on %d ErrNo=%ld|ErrMsg=%s",devH,ATDV_LASTERR(devH),ATDV_ERRMSGP(devH));
											printandlog(index,IPMEV,NULL,str,0);
											stopIpmMedia(index);
											break;
								}
								case IPMEV_STARTMEDIA:
								{
											sprintf(str,"IPMEV_STARTMEDIA Event Received on %d",devH);
											printandlog(index,IPMEV,NULL,str,0);
											processIpmevStartMedia(index);
											break;
								}
								case IPMEV_TELEPHONY_EVENT:
								{
											int dig;
											IPM_TELEPHONY_INFO *pDigit = (IPM_TELEPHONY_INFO *)metaevent->evtdatap;
											dig=pDigit->TelephonyInfo.TelEvtInfo.eTelephonyEventID;
											sprintf(str,"IPMEV_TELEPHONY_EVENT Event Received on %d digBuf=%d",devH,dig);
											printandlog(index,IPMEV,NULL,str,0);
											processIpmevTelephonyEvent(index,dig);
											break;
								}
							case IPMEV_NACK_SENT:
							{
      							sprintf(str,"IPMEV_NACK_SENT Event Received on %d",devH);
										break;
							}
    					case IPMEV_INIT_RESPONSE_SEND:
							{
      							sprintf(str,"IPMEV_INIT_RESPONSE_SEND Event Received on %d",devH);
      							printandlog(index,DX,NULL,str,0);
      							break;
							}
    					case IPMEV_INIT_SEND:
      				{
        						sprintf(str,"IPMEV_INIT_SEND Event Received on %d",devH);
        						printandlog(index,DX,NULL,str,0);
      							break;
      				}
    					case IPMEV_INIT_COMPLETE:
							{
      							sprintf(str,"IPMEV_INIT_COMPLETE Event Received on %d",devH);
     	 							printandlog(index,IPMEV,NULL,str,0);
										processInitSendComplete(index);
      							break;
							}
    					case IPMEV_INIT_FAILED:
      				{			
										IPM_INIT_FAILURE *pIpmFailue=(IPM_INIT_FAILURE *)metaevent->evtdatap;
//										pIpmFailue->data.pNBUP.unSubFlowCount
//										pIpmFailue->data.pNBUP.unRFCICount
//										pIpmFailue->data.pNBUP->pRFCIs.ucID

										sprintf(str,"IPMEV_INIT_FAILED Event Received on %d -> unVersion=%u|eProtocol=%d|unVersion=%u|unCause=%u",
												devH, pIpmFailue->unVersion, pIpmFailue->eProtocol,pIpmFailue->data.pNBUP->unVersion, 
												pIpmFailue->data.pNBUP->unCause);
      							printandlog(index,DX,NULL,str,0);
										
      							break;
							}
    					case IPMEV_INIT_RECEIVED:
      				{
          				sprintf(str,"IPMEV_INIT_RECEIVED Event Received on %d",devH);
          				printandlog(index,DX,NULL,str,0);
        					if(port[index].direction==DIR_IN) {
          						IPM_INIT_RECEIVED *pRecv = (IPM_INIT_RECEIVED *)metaevent->evtdatap;
          						if (!pRecv)
          						{
            						sprintf(str,"Invalid pointer returned from metaevent->evtdatap");
            						printandlog(index,DX,NULL,str,0);
          						}
          						if (pRecv->eProtocol == RTP_PROTOCOL_NBUP)
            							initResponseSendIpm(index);
          						else {
            							sprintf(str,"Unsupported protocol (%d) in init request.", pRecv->eProtocol);
            							printandlog(index,DX,NULL,str,0);
          						}
        					}
      						break;
      				}

// IPM events section ends here
#endif
								case TDX_PLAY:
												{
																sprintf(str,"TDX_PLAY Event Received on %d",devH);
																long termReason;
																termReason=ATDX_TERMMSK(devH);
																if(termReason & TM_USRSTOP) {
																				sprintf(str,"%s Reason=TM_USRSTOP",str);
																} else {
																				sprintf(str,"%s Reason=TM_EOD",str);
																				// Play Prompt Again until user disconnect the call
																				dxPlayAudioPrompt(index);
																}
																printandlog(index,DX,NULL,str,0);
																break;
												}

				}
}
#ifdef USE_VIDEO

static void startGCLib(enumIPCCLIBMediaMode mediaMode)
{
				IPCCLIB_START_DATA   ipccLibData;
				IP_VIRTBOARD         ipVirtualBoard;
				GC_START_STRUCT      gcLibStart;

				INIT_IP_VIRTBOARD(&ipVirtualBoard);

				//ipVirtualBoard.sip_max_calls = num_devices;
				//ipVirtualBoard.total_max_calls = num_devices;
				//ipVirtualBoard.sip_max_calls = (num_sip_devices+num_bicc_devices);
				//ipVirtualBoard.total_max_calls = (num_sip_devices+num_bicc_devices);
				
				if(max_ip_channel_to_start<=0) {
					max_ip_channel_to_start=1;
				}

				ipVirtualBoard.sip_max_calls = max_ip_channel_to_start;
				ipVirtualBoard.total_max_calls = max_ip_channel_to_start;
				ipccLibData.version          = 0x100;
				ipccLibData.num_boards       = 1;
				ipVirtualBoard.version        = IP_VIRTBOARD_VERSION;
				ipVirtualBoard.h323_max_calls    = IP_CFG_NO_CALLS;
				ipVirtualBoard.sip_msginfo_mask  = IP_SIP_MSGINFO_ENABLE | IP_SIP_MIME_ENABLE;

				ipVirtualBoard.sip_signaling_port   = 5060;  // SKS 5060 for the time being

				ipccLibData.board_list = &ipVirtualBoard;
				INIT_IPCCLIB_START_DATA(&ipccLibData, 1,&ipVirtualBoard);

				// Enable 3PCC mode
				ipccLibData.max_parm_data_size     = IP_CFG_PARM_DATA_MAXLEN;
				ipccLibData.media_operational_mode = mediaMode;

				memset(&gcLibStart, 0, sizeof(GC_START_STRUCT));
				CCLIB_START_STRUCT ccstart[] =
				{
								{"GC_IPM_LIB",NULL},
								{"GC_DM3CC_LIB",NULL},
								{"GC_SS7_LIB",NULL},
								{"GC_H3R_LIB",&ipccLibData}
				};
				gcLibStart.cclib_list = ccstart;
				gcLibStart.num_cclibs = 4;

				int rc = gc_Start(&gcLibStart);

				if ( rc != GC_SUCCESS ) {
								printandlog(ALL_DEVICES, GC_APIERR, NULL, "Failed to Start Global Call Library", 0);
								exitdemo(1);
				} else {
								printandlog(ALL_DEVICES,GC_APICALL, NULL, "Starting Global Call Library Successful", 0);
				}
}

static void disconnect_video_call(int index)
{
				printandlog(index, GC_APICALL, NULL, "disconnect_video_call()",0);
				struct channel    *pline;
				int callindex=0;
				char          str[MAX_STRING_SIZE];

				if(port[index].call[callindex].call_state==GCST_NULL || port[index].call[callindex].call_state==GCST_DISCONNECTED)
				{
								return;
				}

				// SKS assumed glare condition never happens subjected
				//     incoming and outgoing cic are in two different group
				pline = &port[index];

				if (pline->call[callindex].crn) {         /* is there an active call? */
					if (pline->call[callindex].dropcall_active == NO) {         /* drop and release the call if can */
						if (gc_DropCall(pline->call[callindex].crn, GC_NORMAL_CLEARING, EV_ASYNC) == GC_SUCCESS) {
								sprintf(str, "gc_DropCall(crn=0x%lx, cause=GC_NORMAL_CLEARING, mode=EV_ASYNC) Success", pline->call[callindex].crn);
								printandlog(pline->index, GC_APICALL, NULL, str, 0);
								pline->call[callindex].dropcall_active = YES;
						} else {
								sprintf(str, "gc_DropCall(crn=0x%lx, cause=GC_NORMAL_CLEARING, mode=EV_ASYNC) Failed", pline->call[callindex].crn);
								printandlog(pline->index, GC_APIERR, NULL, str, 0);
								printandlog(pline->index, STATE, NULL, " ", callindex);
						}
					} else {
							printandlog(pline->index, STATE, NULL, " There is one drop call already active\n",callindex);
					}
				} else {
								printandlog(pline->index, STATE, NULL, " There is no active call\n",callindex);
				}
				pline->resetlinedev_required_after_unblocked = YES;   /* try again later */
				pline->call[callindex].call_state = GCST_DISCONNECTED;
//				port[index].makecall_active = NO;
//				port[index].bBusy=false;
//				resetVariables(index);
}

static void cleanResourcesBeforeExit()
{
				int i;
				for(i=0;i<num_devices;i++) {
						if(port[i].techtype!=SIP) {
								if(m3g_Close(port[i].deviceInfo[M3G].devH)==-1) {
										printandlog(i, M3G, NULL, "m3g_Close(M3G) failed",0);
								}
								if(m3g_Close(port[i].deviceInfo[M3GAUD].devH)==-1) {
										printandlog(i, M3G, NULL, "m3g_Close(M3GAUD) failed",0);
								}
								if(m3g_Close(port[i].deviceInfo[M3GVID].devH)==-1) {
										printandlog(i, M3G, NULL, "m3g_Close(M3GVID) failed",0);
								}
								if(port[i].techtype==BICC) {
										if(ipm_Close(port[i].deviceInfo[IPM].devH,NULL)==-1) {
												printandlog(i, IPM, NULL, "ipm close failed",0);
										}
								}
						} else {
								if(ipm_Close(port[i].deviceInfo[IPM].devH,NULL)==-1) {
										printandlog(i, IPM, NULL, "ipm close failed",0);
								}
						}
						if(mm_Close(port[i].deviceInfo[MM].devH,NULL)==-1) {
										printandlog(i, MM, NULL, "mm_Close() failed",0);
						}
#ifdef USE_CNF
						if(cnf_CloseParty(port[i].deviceInfo[CNFPTY].devH,NULL)==-1) {
										printandlog(i, CNFPTY, NULL, "cnf_CloseParty() failed",0);
						}
#endif
				}
				if(bM3gStarted==M3G_TRUE) {
								stopM3g(BOARD_DEVICE);
				}
#ifdef USE_CNF
				if(cnf_Close(cnfBrdH,NULL)==-1) {
						printandlog(BOARD_DEVICE, CNF, NULL, "cnf_Close() failed",0);
				} else {
						printandlog(BOARD_DEVICE, CNF, NULL, "cnf_Close() success",0);
				}
#endif
				initiate_exit();
}

static void initiate_exit_video_call()
{
				static int     exit_active = NO;
				if (exit_active == YES) {
								return;
				}
				exit_active=YES;
				printandlog(ALL_DEVICES, MISC, NULL, "********** Exit initiated (VIDEO) ***********\n", 0);
				if(bM3gTraceStarted==M3G_TRUE) {
								stopM3gTrace(BOARD_DEVICE,m3gBoardH);
				}
#ifdef USE_CNF
				enable_disable_cnf_dev_events(false);
				enable_disable_cnf_brd_dev_events(false);
#endif
				reset_linedevs_left_to_complete=-1;
				cleanResourcesBeforeExit();
}

static void printMenu()
{
	char str[MAX_STRING_SIZE];
	sprintf(str,"You can perform followings actions:\n\
		h)	To display help menu\n\
		d)	To Disconnect Call from a list of active calls\n\
		s) 	To toggle send/stop periodic UII digits to remote 3G EP\n\
		v) 	To make outgoing VIDEO call from a list of free outgoing channels\n\
		a) 	To make outgoing AUDIO call from a list of free outgoing channels\n\
		p) 	To Print Channel Status and direction\n\
		q)	To Quit Application\n");
	printf("\n%s",str);
}

static void m3gEnableDisableEvents(int index,int handle)
{
				unsigned int enableEvtBitMask=M3G_H245_UII_EVT_TYP|M3G_MES_EVTS_EVT_TYP|M3G_REMOTE_VENDORID_EVT_TYP|M3G_VERBOSE_MSD_EVT_TYP;
				unsigned int disableEvtBitMask=M3G_H245_FASTUPDATE_EVT_TYP|M3G_H245_TEMP_SPAT_TRDFF_EVT_TYP|M3G_H245_VIDEO_FREEZE_EVT_TYP;

				if(m3g_EnableEvents(handle, enableEvtBitMask)!=M3G_SUCCESS) {
								printandlog(index,M3G,NULL,"m3g_EnableEvents() failed",0);
				}
				else {
								printandlog(index,M3G,NULL,"m3g_EnableEvents() successful",0);
				}
				if(m3g_DisableEvents(handle, disableEvtBitMask)!=M3G_SUCCESS) {
								printandlog(index,M3G,NULL,"m3g_DisableEvents() failed",0);
				}
				else {
								printandlog(index,M3G,NULL,"m3g_DisableEvents() successful",0);
				}

}

static void stopPlayRecord(int index,bool rec)
{
		if( ((port[index].bPlayActive==false) && (rec==false)) ||  
						((port[index].bRecordActive==false) && (rec==true)) ) {
				return;
		}
		if(rec==false) {
				port[index].bPlayActive=false;
		} else {
				port[index].bRecordActive=false;
		}

		char str[MAX_STRING_SIZE];

		MM_STOP mmStopInfo[2];
		MM_STOP_DETAILS mmStopDetails;

 		INIT_MM_STOP_DETAILS(&mmStopDetails);
    INIT_MM_STOP(mmStopInfo);

		mmStopInfo[0].unVersion = 0;
		mmStopInfo[0].ItemType = (rec==false?EMM_STOP_AUDIO_PLAY:EMM_STOP_AUDIO_RECORD);
		mmStopInfo[0].details = mmStopDetails;

		if(port[index].callType==CALL_TYPE_AUDIO) {
				mmStopInfo[0].ItemChain = EMM_ITEM_EOT;
				mmStopInfo[0].next = NULL;
		} else {
				mmStopInfo[0].ItemChain = EMM_ITEM_CONT;
				mmStopInfo[0].next = &mmStopInfo[1];
				mmStopInfo[1].unVersion = 0;
				mmStopInfo[1].ItemChain = EMM_ITEM_EOT;
				mmStopInfo[1].ItemType = (rec==false?EMM_STOP_VIDEO_PLAY:EMM_STOP_VIDEO_RECORD);
				mmStopInfo[1].details = mmStopDetails;
				mmStopInfo[1].next = NULL;
		}

		if (mm_Stop(port[index].deviceInfo[MM].devH, mmStopInfo, &index) < 0) {
				sprintf(str,"mm_Stop() to stop %s failed",(rec==false?"Play":"Record"));
		} else  {
				sprintf(str,"mm_Stop() to stop %s success",(rec==false?"Play":"Record"));
		}
		printandlog(index,MM,NULL,str,0);
}

/****************************************************************
 *   NAME: void mmPlayRecordAudioVideoPrompt(int index,bool rec)
 *   DESCRIPTION: audio or audio+video prompt using mm_Play()
 *   INPUT: index
 *   RETURNS: None
 *   CAUTIONS: NA
 ****************************************************************/
static void mmPlayRecordAudioVideoPrompt(int index,bool rec)
{
	char str[MAX_STRING_SIZE];
  struct tm     *Time;
  time_t        Timet;
	int retVal=-1;
  struct timeb    Timeb;
  char          TimeStr[MAX_STRING_SIZE];     /* will hold the time string when built */
	char audioFileName[512];
	char videoFileName[512];
	char fileExtn[10];


	if(port[index].isCallDisconnectInProgress==true) {
			return;
	}

  TimeStr[0] = '\0';
  time(&Timet);
  Time = localtime(&Timet);
  ftime(&Timeb);

  sprintf(TimeStr, "%02d%02d%02d_%02dh%02dm%02ds%03dms", 
										Time->tm_mon+1, Time->tm_mday, (1900+Time->tm_year), 
										Time->tm_hour, Time->tm_min, Time->tm_sec, Timeb.millitm);

		MM_AUDIO_CODEC audioCodecType;
		MM_PLAY_RECORD_LIST m_playRecordList[2];
		MM_PLAY_INFO        m_playInfo;
		MM_MEDIA_ITEM_LIST m_audioMediaList;
		MM_MEDIA_ITEM_LIST m_videoMediaList;

		if(rec==true) {
			port[index].bRecordActive=true;
		} else {
			port[index].bPlayActive=true;
		}
    INIT_MM_MEDIA_ITEM_LIST(&m_audioMediaList);
    INIT_MM_MEDIA_AUDIO(&m_audioMediaList.item.audio);
		INIT_MM_AUDIO_CODEC(&audioCodecType);

		audioCodecType.unSampleRate        = MM_DRT_8KHZ;
		audioCodecType.unBitsPerSample        = 16;

//		if( (port[index].callType==CALL_TYPE_AUDIO) && (port[index].techtype==SIP) )  {
		if(port[index].techtype==SIP)  {
        audioCodecType.unCoding = MM_DATA_FORMAT_ALAW;
        if(rec==false) {
          strcpy(audioFileName,"clip1.g711a");
        } else {
          sprintf(audioFileName,"%s_%s.g711a",port[index].deviceInfo[MM].device_name,TimeStr);
        }
    } else if((port[index].callType==CALL_TYPE_AUDIO && port[index].techtype==BICC) )  {
				if(port[index].audioCodec==AMR) {
						audioCodecType.unCoding = MM_DATA_FORMAT_AMR_NB_12_20K;
						strcpy(fileExtn,"amr");
				} else if(port[index].audioCodec==G711U_5MS || port[index].audioCodec==G711U_20MS) {
						audioCodecType.unCoding = MM_DATA_FORMAT_MULAW;
						strcpy(fileExtn,"g711u");
				} else if(port[index].audioCodec==G711A_5MS || port[index].audioCodec==G711A_20MS) {
						audioCodecType.unCoding = MM_DATA_FORMAT_ALAW;
						strcpy(fileExtn,"g711a");
				}
				if(rec==false) {
					sprintf(audioFileName,"clip1.%s",fileExtn);
				} else {
					sprintf(audioFileName,"%s_%s.%s",port[index].deviceInfo[MM].device_name,TimeStr,fileExtn);
				}
		} else {
				audioCodecType.unCoding = MM_DATA_FORMAT_AMR_NB_12_20K;
			if(rec==false) {
				strcpy(audioFileName,"clip1.amr");
			} else {
					sprintf(audioFileName,"%s_%s.amr",port[index].deviceInfo[MM].device_name,TimeStr);
			}
		}
    m_audioMediaList.ItemChain             		= EMM_ITEM_EOT;
    m_audioMediaList.item.audio.codec         = audioCodecType;
    m_audioMediaList.item.audio.unMode 				= 0;
    m_audioMediaList.item.audio.unOffset 			= 0;
    m_audioMediaList.item.audio.szFileName    = audioFileName;
    m_audioMediaList.item.audio.unAccessMode 	= MM_MEDIA_ACCESS_MODE_FILE;

	if(port[index].callType==CALL_TYPE_AUDIO) {
		// SKS audio only for BICC + NbUP audio call
			m_audioMediaList.item.audio.eFileFormat   = EMM_AUD_FILEFORMAT_PROPRIETARY;

    	INIT_MM_PLAY_RECORD_LIST(&m_playRecordList[0]);
      m_playRecordList[0].ItemType      = EMM_MEDIA_TYPE_AUDIO;
      m_playRecordList[0].list        	= &m_audioMediaList;
      m_playRecordList[0].ItemChain     = EMM_ITEM_EOT;

      INIT_MM_PLAY_RECORD_INFO(&m_playInfo);
      m_playInfo.eFileFormat          = EMM_FILE_FORMAT_PROPRIETARY;
      m_playInfo.list                 = m_playRecordList;
			sprintf(str,"AudioCall:: Starting %s audioFile=%s",(rec==false?"Play":"Record"),m_audioMediaList.item.audio.szFileName);
	} else {
 				m_audioMediaList.item.audio.eFileFormat      = EMM_AUD_FILEFORMAT_PROPRIETARY;
				
				INIT_MM_MEDIA_ITEM_LIST(&m_videoMediaList);
				INIT_MM_MEDIA_VIDEO(&m_videoMediaList.item.video);
				m_videoMediaList.ItemChain = EMM_ITEM_EOT;
//				m_videoMediaList.item.video.unMode = MM_MODE_VID_BEEPINITIATED;
				m_videoMediaList.item.video.eFileFormat      = EMM_FILE_FORMAT_PROPRIETARY;

#ifdef USE_MPEG4
				m_videoMediaList.item.video.codec.Coding               = VIDEO_CODING_MP4V_ES;
				m_videoMediaList.item.video.codec.Profile              = VIDEO_PROFILE_LEVEL_SP0_MPEG4;
				if(rec==false) {
						strcpy(videoFileName,"clip1_mpeg4.vid");
				} else {
						sprintf(videoFileName,"%s_mpeg4_%s.vid",port[index].deviceInfo[MM].device_name,TimeStr);
				}
				m_videoMediaList.item.video.szFileName       = videoFileName;
#else
				m_videoMediaList.item.video.codec.Coding               = VIDEO_CODING_H263;
				m_videoMediaList.item.video.codec.Profile              = VIDEO_PROFILE_0_H263;
				m_videoMediaList.item.video.codec.Level              	 = VIDEO_LEVEL_10_H263;
				if(rec==false) {
						strcpy(videoFileName,"clip1_h263.vid");
				} else {
						sprintf(videoFileName,"%s_h263_%s.vid",port[index].deviceInfo[MM].device_name,TimeStr);
				}
				m_videoMediaList.item.video.szFileName       = videoFileName;
#endif
				m_videoMediaList.item.video.codec.ImageWidth           = VIDEO_IMAGE_WIDTH_176;
				m_videoMediaList.item.video.codec.ImageHeight          = VIDEO_IMAGE_HEIGHT_144;
				m_videoMediaList.item.video.codec.FramesPerSec         = VIDEO_FRAMESPERSEC_15;
				m_videoMediaList.item.video.codec.BitRate             = VIDEO_BITRATE_40K;
				m_videoMediaList.item.video.codec.SamplingRate        = VIDEO_SAMPLING_RATE_90000;
				m_videoMediaList.item.video.codec.VisualConfigSize    = 0;
				m_videoMediaList.item.video.codec.VisualConfiguration = NULL;

				m_videoMediaList.item.video.unAccessMode = MM_MEDIA_ACCESS_MODE_FILE;

				INIT_MM_PLAY_RECORD_LIST(&m_playRecordList[0]);
				m_playRecordList[0].ItemType        = EMM_MEDIA_TYPE_AUDIO;
				m_playRecordList[0].list            = &m_audioMediaList;
				m_playRecordList[0].ItemChain          = EMM_ITEM_CONT;

				INIT_MM_PLAY_RECORD_LIST(&m_playRecordList[1]);
				m_playRecordList[1].ItemType        = EMM_MEDIA_TYPE_VIDEO;
				m_playRecordList[1].list            = &m_videoMediaList;
				m_playRecordList[1].ItemChain          = EMM_ITEM_EOT;

				INIT_MM_PLAY_RECORD_INFO(&m_playInfo);
				m_playInfo.eFileFormat           = EMM_FILE_FORMAT_PROPRIETARY;
				m_playInfo.list                     = m_playRecordList;

				sprintf(str,"VideoCall::Starting %s on %s vidFile=%s|audFile=%s",(rec==false?"Play":"Record"),
								port[index].deviceInfo[MM].device_name,
									m_videoMediaList.item.video.szFileName,m_audioMediaList.item.audio.szFileName);
	}
	printandlog(index,MM,NULL,str,0);

	if(rec==false) {
		retVal=mm_Play(port[index].deviceInfo[MM].devH, &m_playInfo, NULL, NULL);
	} else {

		MM_RUNTIME_CONTROL mmRtc;
		INIT_MM_RUNTIME_CONTROL(&mmRtc);

		mmRtc.Reason = EMM_TERM_MAXTIME;
		if(port[index].callType==CALL_TYPE_AUDIO) {
    	mmRtc.Action = EMM_TA_AUDIO_STOP;
		} else {
    	mmRtc.Action = EMM_TA_AUDIO_VIDEO_STOP;
		}
    mmRtc.unValue = 60000; // 60 seconds

		retVal=mm_Record(port[index].deviceInfo[MM].devH, &m_playInfo, &mmRtc, NULL);
	}

	if (retVal < 0 ) {
			sprintf(str,"%s() failed on %s", (rec==false?"mm_Play":"mm_Record") ,port[index].deviceInfo[MM].device_name);
	} else {
			sprintf(str,"%s() success on %s",(rec==false?"mm_Play":"mm_Record"), port[index].deviceInfo[MM].device_name);
	}
	printandlog(index,MM,NULL,str,0);
}

static void startIpmMedia(int index)
{
			if( (port[index].isCallDisconnectInProgress==true) || 
						(port[index].bIpmMediaStarted==true)) {
      		return;
    	}

      port[index].bIpmMediaStarted=true;

			char str[MAX_STRING_SIZE];

			if(port[index].techtype==BICC) {
					setupH223CallMediaInformation(index);
					sprintf(str,"local[port=%d|ip=%s] || remote[port=%d|ip=%s] || PT[rx=%d|tx=%d]",
									 port[index].ipmLocalRemoteMediaInfo.MediaData[0].mediaInfo.PortInfo.unPortId,
									 port[index].ipmLocalRemoteMediaInfo.MediaData[0].mediaInfo.PortInfo.cIPAddress,
									 port[index].ipmLocalRemoteMediaInfo.MediaData[1].mediaInfo.PortInfo.unPortId,
									 port[index].ipmLocalRemoteMediaInfo.MediaData[1].mediaInfo.PortInfo.cIPAddress,
									 port[index].ipmLocalRemoteMediaInfo.MediaData[2].mediaInfo.NBUPProfileInfo.ucRxPLT,
									 port[index].ipmLocalRemoteMediaInfo.MediaData[2].mediaInfo.NBUPProfileInfo.ucTxPLT);
    							printandlog(index,MISC, NULL,str,0);
			} else {
					if(port[index].callType==CALL_TYPE_AUDIO) {
								setupAudioCallMediaInformation(index);
					}
					if(port[index].callType==CALL_TYPE_VIDEO) {
								setupVideoCallMediaInformation(index);
					}
			}

			if(ipm_StartMedia(port[index].deviceInfo[IPM].devH, &port[index].ipmLocalRemoteMediaInfo, DATA_IP_TDM_BIDIRECTIONAL, EV_ASYNC) == -1)
					sprintf(str,"ipm_StartMedia failed for device=%s",port[index].deviceInfo[IPM].device_name);
			else
					sprintf(str,"ipm_StartMedia successful for device=%s",port[index].deviceInfo[IPM].device_name);
			printandlog(index,IPM,NULL,str,0);
}

static void processIpmevTelephonyEvent(int index, int dig)
{
#ifdef USE_CNF
											if(dig==0 && port[index].bPartyIntoCnf==false) {
													stopPlayRecord(index,false);
													stopPlayRecord(index,true);
													routeUnrouteDevices(index,IPM,MM,UNROUTE,DMFL_TRANSCODE_ON);
                      }
#endif
}
static void processIpmevOpen(int index)
{
                               getLocalMediaInfo(index);
                                getTxRxPortInfo(index,port[index].deviceInfo[IPM].devH);
                                gc_demo_open_mm_channel(index);
                                enableIpmEvents(index);
                                setDTMFMode(index);

}
static void processIpmevStopped(int index)
{
                     if(port[index].bPartyIntoCnf==true) {
#ifdef USE_CNF
                          cnfRemoveParty(index);
#endif
                      } else {
                    if(port[index].techtype!=BICC) {
                      if (port[index].callType==CALL_TYPE_VIDEO) {
                        routeUnrouteDevices(index,IPM,MM,UNROUTE,DMFL_TRANSCODE_ON);
                      } else {
                          unRouteIpmAndDx(index);
                          disconnect_video_call(index);
                      }
                    } else {
                      if (port[index].callType==CALL_TYPE_VIDEO) {
                        routeUnrouteDevices(index,IPM,M3G,UNROUTE,DMFL_TRANSCODE_NATIVE);
                      } else {
                        routeUnrouteDevices(index,IPM,MM,UNROUTE,DMFL_TRANSCODE_NATIVE);
                      }
                    }
                      }

}

static void processIpmevStartMedia(int index)
{
	    if(port[index].techtype==BICC) {
           if(port[index].callType==CALL_TYPE_VIDEO) {
                              routeUnrouteDevices(index,M3G,MM,ROUTE,DMFL_TRANSCODE_NATIVE);
            }
            if(port[index].direction==DIR_OUT) {
                            initSendIpm(index);
            }
      } else {    // SIP call
           if(port[index].callType==CALL_TYPE_VIDEO) {
                          if(port[index].direction==DIR_OUT) {
                              mmPlayRecordAudioVideoPrompt(index,false);
                          } else {
                              mmPlayRecordAudioVideoPrompt(index,true);
                          }
//                        sendIFrameRequest(index);
           } else {
                          dxPlayAudioPrompt(index);
           }
     }

}
static void stopIpmMedia (int index)
{
				char str[MAX_STRING_SIZE];
				
				if(port[index].bIpmMediaStarted==false) {
                return;
        }
				port[index].bIpmMediaStarted=false;
				stopPlayRecord(index,false);
				stopPlayRecord(index,true);
				if (ipm_Stop(port[index].deviceInfo[IPM].devH, STOP_ALL, EV_ASYNC) == -1)
								sprintf(str,"ipm_Stop for failed device=%s",port[index].deviceInfo[IPM].device_name);
				else {
								sprintf(str,"ipm_Stop successful for device=%s",port[index].deviceInfo[IPM].device_name);
				}
				printandlog(index,IPM,NULL,str,0);
}

static void setupVideoCallMediaInformation(int index)
{
				int nMediaCnt=0;
				memset(&port[index].ipmLocalRemoteMediaInfo, 0, sizeof(IPM_MEDIA_INFO));

				// G.711a
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].eMediaType = MEDIATYPE_AUDIO_REMOTE_CODER_INFO;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.AudioCoderInfo.eCoderType = CODER_TYPE_G711ALAW64K;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.AudioCoderInfo.eFrameSize = CODER_FRAMESIZE_20;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.AudioCoderInfo.unFramesPerPkt= 1;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.AudioCoderInfo.eVadEnable  = CODER_VAD_DISABLE;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.AudioCoderInfo.unCoderPayloadType    = port[index].remoteEPAudioPayloadID;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.AudioCoderInfo.unRedPayloadType  = port[index].remoteEPAudioPayloadID;

				++nMediaCnt;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].eMediaType = MEDIATYPE_AUDIO_LOCAL_CODER_INFO;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.AudioCoderInfo.eCoderType = CODER_TYPE_G711ALAW64K;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.AudioCoderInfo.eFrameSize = CODER_FRAMESIZE_20;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.AudioCoderInfo.unFramesPerPkt = 1;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.AudioCoderInfo.eVadEnable = CODER_VAD_DISABLE;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.AudioCoderInfo.unCoderPayloadType  = port[index].remoteEPAudioPayloadID;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.AudioCoderInfo.unRedPayloadType = port[index].remoteEPAudioPayloadID;


				++nMediaCnt;
				INIT_IPM_VIDEO_CODER_INFO(&port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.VideoCoderInfo);
				INIT_IPM_VIDEO_CODER_INFO_EX(&port[index].localVideoInfoEx);
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].eMediaType = MEDIATYPE_VIDEO_REMOTE_CODER_INFO;
#ifndef USE_MPEG4
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.VideoCoderInfo.eCoderType = CODER_TYPE_H263;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.VideoCoderInfo.unCoderPayloadType    = port[index].remoteEPVideoPayloadID;
				port[index].localVideoInfoEx.eProfile=VIDEO_PROFILE_0_H263;
#else 
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.VideoCoderInfo.eCoderType = CODER_TYPE_MP4V_ES;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.VideoCoderInfo.unCoderPayloadType    = port[index].remoteEPVideoPayloadID;
				port[index].localVideoInfoEx.eProfile=VIDEO_PROFILE_LEVEL_SP1_MPEG4;
#endif
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.VideoCoderInfo.pExtraCoderInfo=&port[index].localVideoInfoEx;
				port[index].localVideoInfoEx.unBitRate = VIDEO_BITRATE_128K;
				port[index].localVideoInfoEx.eFramesPerSec = VIDEO_FRAMESPERSEC_15;
				port[index].localVideoInfoEx.eSamplingRate = VIDEO_SAMPLING_RATE_90000;
				port[index].localVideoInfoEx.eImageWidth=VIDEO_IMAGE_WIDTH_176;
				port[index].localVideoInfoEx.eImageHeight = VIDEO_IMAGE_HEIGHT_144;
				port[index].localVideoInfoEx.unVisualConfigSize = 0;
				port[index].localVideoInfoEx.szVisualConfiguration = NULL;

				++nMediaCnt;

				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].eMediaType = MEDIATYPE_VIDEO_LOCAL_CODER_INFO;
				INIT_IPM_VIDEO_CODER_INFO(&port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.VideoCoderInfo);
				INIT_IPM_VIDEO_CODER_INFO_EX(&port[index].localVideoInfoEx);

#ifndef USE_MPEG4
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.VideoCoderInfo.eCoderType = CODER_TYPE_H263;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.VideoCoderInfo.unCoderPayloadType  = port[index].remoteEPVideoPayloadID;
				port[index].localVideoInfoEx.eProfile=VIDEO_PROFILE_0_H263;
#else
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.VideoCoderInfo.eCoderType = CODER_TYPE_MP4V_ES;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.VideoCoderInfo.unCoderPayloadType  = port[index].remoteEPVideoPayloadID;
				port[index].localVideoInfoEx.eProfile=VIDEO_PROFILE_LEVEL_SP1_MPEG4;
#endif
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.VideoCoderInfo.pExtraCoderInfo=&port[index].localVideoInfoEx;
				port[index].localVideoInfoEx.unBitRate = VIDEO_BITRATE_128K;
				port[index].localVideoInfoEx.eSamplingRate = VIDEO_SAMPLING_RATE_90000;
				port[index].localVideoInfoEx.eFramesPerSec = VIDEO_FRAMESPERSEC_15;
				port[index].localVideoInfoEx.eImageWidth=VIDEO_IMAGE_WIDTH_176;
				port[index].localVideoInfoEx.eImageHeight = VIDEO_IMAGE_HEIGHT_144;
				port[index].localVideoInfoEx.unVisualConfigSize = 0;
				port[index].localVideoInfoEx.szVisualConfiguration = NULL;

				// remote audio port
				++nMediaCnt;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].eMediaType = MEDIATYPE_AUDIO_REMOTE_RTP_INFO;
				strcpy(port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.PortInfo.cIPAddress, port[index].remoteEPIpAddress);
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.PortInfo.unPortId = port[index].remoteEPAudioPortNum;

				++nMediaCnt;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].eMediaType = MEDIATYPE_AUDIO_REMOTE_RTCP_INFO;
				strcpy(port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.PortInfo.cIPAddress, port[index].remoteEPIpAddress);
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.PortInfo.unPortId = port[index].remoteEPAudioPortNum+1;


				// remote video port
				++nMediaCnt;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].eMediaType = MEDIATYPE_VIDEO_REMOTE_RTP_INFO;
				strcpy(port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.PortInfo.cIPAddress, port[index].remoteEPIpAddress);
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.PortInfo.unPortId = port[index].remoteEPVideoPortNum;

				++nMediaCnt;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].eMediaType = MEDIATYPE_VIDEO_REMOTE_RTCP_INFO;
				strcpy(port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.PortInfo.cIPAddress, port[index].remoteEPIpAddress);
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.PortInfo.unPortId = port[index].remoteEPVideoPortNum+1;


				port[index].ipmLocalRemoteMediaInfo.unCount = nMediaCnt + 1;
}

static void setupAudioCallMediaInformation(int index)
{
				int nMediaCnt=0;
				// G.711
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].eMediaType = MEDIATYPE_AUDIO_REMOTE_CODER_INFO;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.CoderInfo.eCoderType = CODER_TYPE_G711ALAW64K;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.CoderInfo.eFrameSize = CODER_FRAMESIZE_20;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.CoderInfo.unFramesPerPkt= 1;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.CoderInfo.eVadEnable  = CODER_VAD_DISABLE;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.CoderInfo.unCoderPayloadType    = port[index].remoteEPAudioPayloadID;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.CoderInfo.unRedPayloadType  = port[index].remoteEPAudioPayloadID;
				++nMediaCnt;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].eMediaType = MEDIATYPE_AUDIO_LOCAL_CODER_INFO;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.CoderInfo.eCoderType = CODER_TYPE_G711ALAW64K;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.CoderInfo.eFrameSize = CODER_FRAMESIZE_20;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.CoderInfo.unFramesPerPkt = 1;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.CoderInfo.eVadEnable = CODER_VAD_DISABLE;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.CoderInfo.unCoderPayloadType  = port[index].remoteEPAudioPayloadID;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.CoderInfo.unRedPayloadType = port[index].remoteEPAudioPayloadID;
				// remote audio port
				++nMediaCnt;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].eMediaType = MEDIATYPE_AUDIO_REMOTE_RTP_INFO;
				strcpy(port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.PortInfo.cIPAddress, port[index].remoteEPIpAddress);
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.PortInfo.unPortId = port[index].remoteEPAudioPortNum;

				++nMediaCnt;
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].eMediaType = MEDIATYPE_AUDIO_REMOTE_RTCP_INFO;
				strcpy(port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.PortInfo.cIPAddress, port[index].remoteEPIpAddress);
				port[index].ipmLocalRemoteMediaInfo.MediaData[nMediaCnt].mediaInfo.PortInfo.unPortId = port[index].remoteEPAudioPortNum+1;

				port[index].ipmLocalRemoteMediaInfo.unCount = nMediaCnt + 1;
}

static void setupH223CallMediaInformation(int index)
{
		char str[MAX_STRING_SIZE];
    memset(&port[index].ipmLocalRemoteMediaInfo, 0, sizeof(IPM_MEDIA_INFO));

    port[index].ipmLocalRemoteMediaInfo.unCount = 3;

    port[index].ipmLocalRemoteMediaInfo.MediaData[0].eMediaType = MEDIATYPE_NBUP_LOCAL_RTP_INFO;
    port[index].ipmLocalRemoteMediaInfo.MediaData[0].mediaInfo.PortInfo.unPortId = port[index].localEPNbUPPortNum;
    strcpy (port[index].ipmLocalRemoteMediaInfo.MediaData[0].mediaInfo.PortInfo.cIPAddress, port[index].localEPIpAddress);

    port[index].ipmLocalRemoteMediaInfo.MediaData[1].eMediaType = MEDIATYPE_NBUP_REMOTE_RTP_INFO;
    port[index].ipmLocalRemoteMediaInfo.MediaData[1].mediaInfo.PortInfo.unPortId = port[index].remoteEPNbUPPortNum;
    strcpy (port[index].ipmLocalRemoteMediaInfo.MediaData[1].mediaInfo.PortInfo.cIPAddress, port[index].remoteEPIpAddress);

    port[index].ipmLocalRemoteMediaInfo.MediaData[2].eMediaType = MEDIATYPE_NBUP_PROFILE_INFO;
    INIT_IPM_NBUP_PROFILE_INFO (&(port[index].ipmLocalRemoteMediaInfo.MediaData[2].mediaInfo.NBUPProfileInfo));

		if(port[index].callType==CALL_TYPE_VIDEO) {
			printf("using NBUP_PROFILE_3G324M\n");
	    port[index].ipmLocalRemoteMediaInfo.MediaData[2].mediaInfo.NBUPProfileInfo.eProfileType = NBUP_PROFILE_3G324M;
		} else if(port[index].callType==CALL_TYPE_AUDIO) {
			switch(port[index].audioCodec) 
			{
				case AMR:
				{
					port[index].ipmLocalRemoteMediaInfo.MediaData[2].mediaInfo.NBUPProfileInfo.eProfileType =NBUP_PROFILE_AMR_NB;
					break;
				}
				case G711U_5MS:
				{
					port[index].ipmLocalRemoteMediaInfo.MediaData[2].mediaInfo.NBUPProfileInfo.eProfileType =NBUP_PROFILE_G711ULAW64K_5MS;
					break;
				}
				case G711U_20MS:
				{
					port[index].ipmLocalRemoteMediaInfo.MediaData[2].mediaInfo.NBUPProfileInfo.eProfileType =NBUP_PROFILE_G711ULAW64K_20MS;
					break;
				}
				case G711A_5MS:
				{
						port[index].ipmLocalRemoteMediaInfo.MediaData[2].mediaInfo.NBUPProfileInfo.eProfileType =NBUP_PROFILE_G711ALAW64K_5MS;
					break;
				}
				case G711A_20MS:
        {
						port[index].ipmLocalRemoteMediaInfo.MediaData[2].mediaInfo.NBUPProfileInfo.eProfileType =NBUP_PROFILE_G711ALAW64K_20MS;
          break;
        }
				default:
				{
					printf("Invalid NbUP audio codec %d\n",port[index].audioCodec);
					break;
				}
			}
		}

    port[index].ipmLocalRemoteMediaInfo.MediaData[2].mediaInfo.NBUPProfileInfo.ucTxPLT = port[index].remoteEPNbUPPayloadID;
    port[index].ipmLocalRemoteMediaInfo.MediaData[2].mediaInfo.NBUPProfileInfo.ucRxPLT = port[index].remoteEPNbUPPayloadID;

		sprintf(str,"NbUpProfile=%d | remote[port=%d|ip=%s|pt=%d] && local[port=%d|ip=%s|pt=%d]",
								port[index].ipmLocalRemoteMediaInfo.MediaData[2].mediaInfo.NBUPProfileInfo.eProfileType,
								port[index].remoteEPNbUPPortNum,port[index].remoteEPIpAddress,port[index].remoteEPNbUPPayloadID,
								port[index].localEPNbUPPortNum,port[index].localEPIpAddress,port[index].remoteEPNbUPPayloadID);
		printandlog(index,MISC, NULL,str,0);
}

static int parseApmMessage(METAEVENT *metaevent,struct channel *pline)
{
    int         nCtr;
		int retVal;
    S7_IE_BLK      ie_blk;
    int         sdpLen;
    int         nAllIELen=0;
		char sdpValue[MAX_STRING_SIZE];

		if(pline->techtype==BICC) {
       if (gc_GetSigInfo(pline->ldev,(char *)&ie_blk, 0, metaevent) != GC_SUCCESS)  {
          // process error
          printandlog(pline->index, MISC, NULL, "gc_GetSigInfo() call failed", 0);
       }
    }

		nAllIELen=(int)ie_blk.length;
    for (nCtr=0,sdpLen=0;nCtr<(nAllIELen-1);nCtr++)  {
				if( (ie_blk.data[nCtr]>=32 && ie_blk.data[nCtr]<=122) || ie_blk.data[nCtr]=='\n' || ie_blk.data[nCtr]=='\r' ) {
						sdpValue[sdpLen]=(char)ie_blk.data[nCtr];
						sdpLen++;
				}
		}
		sdpValue[sdpLen]='\0';
		if(sdpLen!=0) { 
					printandlog(pline->index, MISC, NULL,sdpValue, 0);
					retVal=storeSdpValues(sdpValue,pline->index);
		}
		return retVal;
}
static void initSendIpm(int index)
{
  char str[MAX_STRING_SIZE];
  IPM_INIT_SEND initSend;
  IPM_NBUP_INIT_SEND nbupSend;

  INIT_IPM_INIT_SEND (&initSend);
  initSend.eProtocol = RTP_PROTOCOL_NBUP;
  initSend.data.pNBUP = &nbupSend;


	if(port[index].callType==CALL_TYPE_VIDEO) {
		printf("initSendIpm(handle=%ld | ucID=0x44 && eFlowSize=NBUP_FLOW_SIZE_320_BITS)\n",port[index].deviceInfo[IPM].devH);
  	INIT_IPM_NBUP_INIT_SEND (&nbupSend, 1, 1);
  	nbupSend.pRFCIs[0].ucID = (unsigned char) 0x44;
  	nbupSend.pRFCIs[0].pSubFlows[0].eFlowSize = NBUP_FLOW_SIZE_320_BITS;
	} else {
	switch(port[index].audioCodec)
	{
		case AMR:
		{
  		INIT_IPM_NBUP_INIT_SEND (&nbupSend, 3, 2);
			switch(port[index].remoteEPAmrModeSet) 
			{
				case 0: 
				{
        	nbupSend.pRFCIs[0].ucID = 2;
        	nbupSend.pRFCIs[0].pSubFlows[0].eFlowSize = (eIPM_NBUP_FLOW_SIZE)42;
        	nbupSend.pRFCIs[0].pSubFlows[1].eFlowSize = (eIPM_NBUP_FLOW_SIZE)53;
        	nbupSend.pRFCIs[0].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
        	nbupSend.pRFCIs[1].ucID = 16;
        	nbupSend.pRFCIs[1].pSubFlows[0].eFlowSize = NBUP_FLOW_SIZE_39_BITS;
        	nbupSend.pRFCIs[1].pSubFlows[1].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
        	nbupSend.pRFCIs[1].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
					break;
				}
      	case 1:
				{
         	nbupSend.pRFCIs[0].ucID = 3;
         	nbupSend.pRFCIs[0].pSubFlows[0].eFlowSize = (eIPM_NBUP_FLOW_SIZE)49;
         	nbupSend.pRFCIs[0].pSubFlows[1].eFlowSize = (eIPM_NBUP_FLOW_SIZE)54;
         	nbupSend.pRFCIs[0].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
         	nbupSend.pRFCIs[1].ucID = 15;
         	nbupSend.pRFCIs[1].pSubFlows[0].eFlowSize = NBUP_FLOW_SIZE_39_BITS;
         	nbupSend.pRFCIs[1].pSubFlows[1].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
         	nbupSend.pRFCIs[1].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
					break;
				}
				case 2:
				{
         	nbupSend.pRFCIs[0].ucID = 4;
         	nbupSend.pRFCIs[0].pSubFlows[0].eFlowSize = (eIPM_NBUP_FLOW_SIZE)55;
         	nbupSend.pRFCIs[0].pSubFlows[1].eFlowSize = (eIPM_NBUP_FLOW_SIZE)63;
         	nbupSend.pRFCIs[0].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
         	nbupSend.pRFCIs[1].ucID = 14;
         	nbupSend.pRFCIs[1].pSubFlows[0].eFlowSize = NBUP_FLOW_SIZE_39_BITS;
         	nbupSend.pRFCIs[1].pSubFlows[1].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
         	nbupSend.pRFCIs[1].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
					break;
				}
				case 3:
				{
         	nbupSend.pRFCIs[0].ucID = 5;
         	nbupSend.pRFCIs[0].pSubFlows[0].eFlowSize = (eIPM_NBUP_FLOW_SIZE)58;
         	nbupSend.pRFCIs[0].pSubFlows[1].eFlowSize = (eIPM_NBUP_FLOW_SIZE)76;
         	nbupSend.pRFCIs[0].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
         	nbupSend.pRFCIs[1].ucID = 13;
         	nbupSend.pRFCIs[1].pSubFlows[0].eFlowSize = NBUP_FLOW_SIZE_39_BITS;
         	nbupSend.pRFCIs[1].pSubFlows[1].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
         	nbupSend.pRFCIs[1].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
					break;
				}
				case 4:
				{
         	nbupSend.pRFCIs[0].ucID = 6;
         	nbupSend.pRFCIs[0].pSubFlows[0].eFlowSize = (eIPM_NBUP_FLOW_SIZE)61;
         	nbupSend.pRFCIs[0].pSubFlows[1].eFlowSize = (eIPM_NBUP_FLOW_SIZE)87;
         	nbupSend.pRFCIs[0].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
         	nbupSend.pRFCIs[1].ucID = 12;
         	nbupSend.pRFCIs[1].pSubFlows[0].eFlowSize = NBUP_FLOW_SIZE_39_BITS;
         	nbupSend.pRFCIs[1].pSubFlows[1].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
         	nbupSend.pRFCIs[1].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
					break;
				}
				case 5:
				{
         	nbupSend.pRFCIs[0].ucID = 7;
         	nbupSend.pRFCIs[0].pSubFlows[0].eFlowSize = (eIPM_NBUP_FLOW_SIZE)75;
         	nbupSend.pRFCIs[0].pSubFlows[1].eFlowSize = (eIPM_NBUP_FLOW_SIZE)84;
         	nbupSend.pRFCIs[0].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
         	nbupSend.pRFCIs[1].ucID = 11;
         	nbupSend.pRFCIs[1].pSubFlows[0].eFlowSize = NBUP_FLOW_SIZE_39_BITS;
         	nbupSend.pRFCIs[1].pSubFlows[1].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
         	nbupSend.pRFCIs[1].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
					break;
				}
				case 6:
				{
         	nbupSend.pRFCIs[0].ucID = 8;
         	nbupSend.pRFCIs[0].pSubFlows[0].eFlowSize = (eIPM_NBUP_FLOW_SIZE)65;
         	nbupSend.pRFCIs[0].pSubFlows[1].eFlowSize = (eIPM_NBUP_FLOW_SIZE)99;
         	nbupSend.pRFCIs[0].pSubFlows[2].eFlowSize = (eIPM_NBUP_FLOW_SIZE)40;
         	nbupSend.pRFCIs[1].ucID = 10;
         	nbupSend.pRFCIs[1].pSubFlows[0].eFlowSize = NBUP_FLOW_SIZE_39_BITS;
         	nbupSend.pRFCIs[1].pSubFlows[1].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
         	nbupSend.pRFCIs[1].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
					break;
				}
				case 7:
				{
         	nbupSend.pRFCIs[0].ucID = 9;
         	nbupSend.pRFCIs[0].pSubFlows[0].eFlowSize = NBUP_FLOW_SIZE_81_BITS;
         	nbupSend.pRFCIs[0].pSubFlows[1].eFlowSize = NBUP_FLOW_SIZE_103_BITS;
         	nbupSend.pRFCIs[0].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_60_BITS;
         	nbupSend.pRFCIs[1].ucID = 1;
         	nbupSend.pRFCIs[1].pSubFlows[0].eFlowSize = NBUP_FLOW_SIZE_39_BITS;
         	nbupSend.pRFCIs[1].pSubFlows[1].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
         	nbupSend.pRFCIs[1].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
					break;
      	}
				default:
					break;
			} // AMR case ends here
			break;
		}
		case G711A_5MS:
		{
  	  INIT_IPM_NBUP_INIT_SEND (&nbupSend, 3, 1);
      nbupSend.pRFCIs[0].ucID = 21;
      nbupSend.pRFCIs[0].pSubFlows[0].eFlowSize = NBUP_FLOW_SIZE_320_BITS;
      nbupSend.pRFCIs[0].pSubFlows[1].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
      nbupSend.pRFCIs[0].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
			break;
		}
		case G711A_20MS:
		{
  	  INIT_IPM_NBUP_INIT_SEND (&nbupSend, 3, 1);
      nbupSend.pRFCIs[0].ucID = 22;
      nbupSend.pRFCIs[0].pSubFlows[0].eFlowSize = NBUP_FLOW_SIZE_1280_BITS;
      nbupSend.pRFCIs[0].pSubFlows[1].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
      nbupSend.pRFCIs[0].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
			break;
		}
		case G711U_5MS:
		{
  	  INIT_IPM_NBUP_INIT_SEND (&nbupSend, 3, 1);
      nbupSend.pRFCIs[0].ucID = 23;
      nbupSend.pRFCIs[0].pSubFlows[0].eFlowSize = NBUP_FLOW_SIZE_320_BITS;
      nbupSend.pRFCIs[0].pSubFlows[1].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
      nbupSend.pRFCIs[0].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
			break;
		}
		case G711U_20MS:
    {
  	  INIT_IPM_NBUP_INIT_SEND (&nbupSend, 3, 1);
      nbupSend.pRFCIs[0].ucID = 24;
      nbupSend.pRFCIs[0].pSubFlows[0].eFlowSize = NBUP_FLOW_SIZE_1280_BITS;
      nbupSend.pRFCIs[0].pSubFlows[1].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
      nbupSend.pRFCIs[0].pSubFlows[2].eFlowSize = NBUP_FLOW_SIZE_0_BITS;
			break;
   }
	 default:
			printandlog(index,IPM,NULL,"Invalid Audio Codec Type",0);
			break;
	} // switch audioCodec ends here
	}


  if (ipm_InitSend (port[index].deviceInfo[IPM].devH, &initSend) == -1)
      sprintf(str,"ipm_InitSend for Nbup failed for device=%s",port[index].deviceInfo[IPM].device_name);
  else
      sprintf(str,"ipm_InitSend successful for Nbup for device=%s",port[index].deviceInfo[IPM].device_name);
  printandlog(index,IPM,NULL,str,0);

  FREE_IPM_NBUP_INIT_SEND (&nbupSend);
}
static void initResponseSendIpm(int index)
{
  char str[MAX_STRING_SIZE];
  IPM_INIT_RESPONSE initResp;
  IPM_NBUP_INIT_RESPONSE nbupResp;

  INIT_IPM_INIT_RESPONSE (&initResp);
  initResp.eProtocol = RTP_PROTOCOL_NBUP;
  initResp.data.pNBUP = &nbupResp;

  INIT_IPM_NBUP_INIT_RESPONSE (&nbupResp);
  nbupResp.eResponse = NBUP_INIT_RESPONSE_ACK;

 if (ipm_InitResponseSend (port[index].deviceInfo[IPM].devH, &initResp) == -1)
     sprintf(str,"ipm_InitResponseSend for Nbup failed for device=%s",port[index].deviceInfo[IPM].device_name);
  else
      sprintf(str,"ipm_InitResponseSend successful for Nbup for device=%s",port[index].deviceInfo[IPM].device_name);
    printandlog(index,DX,NULL,str,0);

}

static void processInitSendComplete(int index)
{
	char str[MAX_STRING_SIZE];
  sprintf(str,"processInitSendComplete()");
  printandlog(index,IPMEV,NULL,str,0);

	if(port[index].callType==CALL_TYPE_VIDEO) {
    	startH245(index);
	} else {
			mmPlayRecordAudioVideoPrompt(index,((port[index].direction==DIR_OUT)?false:true));
	}
}

static void modifyMedia(int index,M3G_E_DIRECTION direction)
{
		if(port[index].isCallDisconnectInProgress==true) {
      return;
  	}
		char str[MAX_STRING_SIZE];

		if(port[index].bMediaStarted==false) {
    		port[index].bMediaStarted=true;
  	} else if(direction!=M3G_E_IDLE) {
    		direction=M3G_E_TXRX;
  	}

		if (m3g_ModifyMedia(port[index].deviceInfo[M3GAUD].devH, direction)==M3G_SUCCESS) {
				sprintf(str,"m3g_ModifyMedia(%ld) for audio success dir=%s",port[index].deviceInfo[M3GAUD].devH,dirStr[direction]);
		} else {
				sprintf(str,"m3g_ModifyMedia(%ld) audio failure dir=%s",port[index].deviceInfo[M3GAUD].devH,dirStr[direction]);
		}
		printandlog(index,M3G,NULL,str,0);

		if (m3g_ModifyMedia(port[index].deviceInfo[M3GVID].devH, direction)==M3G_SUCCESS) {
				sprintf(str,"m3g_ModifyMedia(%ld) video success dir=%s",port[index].deviceInfo[M3GVID].devH,dirStr[direction]);
		} else {
				sprintf(str,"m3g_ModifyMedia(%ld) video failure dir=%s",port[index].deviceInfo[M3GVID].devH,dirStr[direction]);
		}
		printandlog(index,M3G,NULL,str,0);
}

static void sendOLCAck(int index,M3G_REMOTE_OLC_REQ* remoteOlcReq)
{
	char str[MAX_STRING_SIZE];
	if(remoteOlcReq->capabilityType==M3G_E_AUDIO_CAPABILITY) {
		port[index].remoteAudioLogicalChannel=remoteOlcReq->logicalChannelNumber;
	} else {
		port[index].remoteVideoLogicalChannel=remoteOlcReq->logicalChannelNumber;
	}
	if (m3g_RespondToOLC(port[index].deviceInfo[M3G].devH, remoteOlcReq->logicalChannelNumber, M3G_E_OLCACK)==M3G_SUCCESS ) {
			sprintf(str,"m3g_RespondToLC(%ld) success for type %d lcn=%d",
						port[index].deviceInfo[M3G].devH,remoteOlcReq->capabilityType,remoteOlcReq->logicalChannelNumber);
	} else {
			sprintf(str,"m3g_RespondToLC(%ld) failure for type %d lcn=%d",
						port[index].deviceInfo[M3G].devH,remoteOlcReq->capabilityType,remoteOlcReq->logicalChannelNumber);
	}
	printandlog(index,MISC,NULL,str,0);
}

static void sendOpenLC(int index)
{
				char str[MAX_STRING_SIZE];
				int i=0;

#ifdef USE_MPEG4
				int length=0;
				static const unsigned char MPEG4_DCI_HELIX_QCIF[]=
				{
								0x00,0x00,0x01,0xb0,0x08,0x00,0x00,0x01,0xb5,0x09,
								0x00,0x00,0x01,0x00,0x00,0x00,0x01,0x20,0x00,0x84,
								0x5d,0x4c,0x28,0x2c,0x20,0x90,0xa2,0x8f
				};
#if 0
// Helix DCI
								0x00,0x00,0x01,0xb0,0x08,0x00,0x00,0x01,0xb5,0x0e,
								0xa0,0x20,0x20,0x2f,0x00,0x00,0x01,0x00,0x00,0x00,
								0x01,0x20,0x00,0xc7,0x88,0xba,0x98,0x50,0x58,0x41,
								0x21,0x46,0x3f
#endif
#endif

				// since we have only one audio/video into local capability list
				M3G_H223_LC_PARAMS h223OLCParams;

				int handle=port[index].deviceInfo[M3G].devH;

				M3G_AUDIO_CAPABILITY* pAudioCap = NULL;
				M3G_VIDEO_CAPABILITY* pVideoCap = NULL;

				for(i=0;i<port[index].videoCapsList.numCaps;i++) {
								sprintf(str,"Video:: tableEntryNo=%u|coderType=%s|direction=%s",port[index].videoCapsList.capability[i].videoCapability.tableEntryNumber,
																(port[index].videoCapsList.capability[i].videoCapability.coderType==M3G_E_H263?"M3G_E_H263":"M3G_E_MPEG4"),
																(port[index].videoCapsList.capability[i].videoCapability.direction==M3G_E_RX?"M3G_E_RX":"M3G_E_TX") );
								printandlog(index,M3G,NULL,str,0);
								if(port[index].videoCapsList.capability[i].videoCapability.direction==M3G_E_TX) {
												pVideoCap= &port[index].videoCapsList.capability[i].videoCapability;
								}
				}
				for(i=0;i<port[index].audioCapsList.numCaps;i++) {
								sprintf(str,"Audio:: tableEntryNo=%u|coderType=%s|direction=%s",port[index].audioCapsList.capability[i].audioCapability.tableEntryNumber,
																(port[index].audioCapsList.capability[i].audioCapability.coderType==M3G_E_GSM_AMR_NB?"M3G_E_GSM_AMR_NB":"M3G_E_G7231"),
																(port[index].audioCapsList.capability[i].audioCapability.direction==M3G_E_RX?"M3G_E_RX":"M3G_E_TX") );
								printandlog(index,M3G,NULL,str,0);
								if(port[index].audioCapsList.capability[i].audioCapability.direction==M3G_E_TX) {
												pAudioCap= &port[index].audioCapsList.capability[i].audioCapability;
								}
				}
				// SKS JACK ERROR handling if no matching capability
				if(pVideoCap==NULL || pAudioCap==NULL) {
								stopH245(index);
								return;
				}

				// send audio OLC

				INIT_M3G_H223_LC_PARAMS(&h223OLCParams);
				h223OLCParams.adaptationLayerType=M3G_E_AL2_WITHOUT_SEQ_NUMS;	//M3G_E_AL2_WITHOUT_SEQ_NUMS
				h223OLCParams.segmentable = M3G_FALSE;									// audio is always not segmented

				//SKS rest default values will be initialzed by INIT_M3G_H223_LC_PARAMS() function
				if(m3g_OpenLC(handle, &h223OLCParams, M3G_E_AUDIO_CAPABILITY,(M3G_CAPABILITY*)pAudioCap) == M3G_SUCCESS) {
								sprintf(str,"m3g_OpenLC(%d) success for audio",handle);
				} else {
								sprintf(str,"m3g_OpenLC(%d) failed for audio",handle);
				}
				printandlog(index,M3G,NULL,str,0);

				memset(&h223OLCParams, 0, sizeof(M3G_H223_LC_PARAMS));

				// send video OLC

				INIT_M3G_H223_LC_PARAMS(&h223OLCParams);

				h223OLCParams.adaptationLayerType=M3G_E_AL2_WITH_SEQ_NUMS;	//M3G_E_AL2_WITH_SEQ_NUMS
				h223OLCParams.segmentable = M3G_TRUE;									// video is always segmented

#ifndef USE_MPEG4
				pVideoCap->options.h263.qcifMPI=2;
				pVideoCap->options.h263.sqcifMPI=0;
#else 
				pVideoCap->options.mpeg4.profileAndLevel = 0x8;
				pVideoCap->options.mpeg4.object = 1;
				length=sizeof(MPEG4_DCI_HELIX_QCIF);
				pVideoCap->options.mpeg4.decoderConfigLength=length;
				memcpy(pVideoCap->options.mpeg4.decoderConfigInfo,MPEG4_DCI_HELIX_QCIF,length);
				pVideoCap->options.mpeg4.visualBackChannel=M3G_FALSE;
#endif

				//SKS rest default values will be initialzed by INIT_M3G_H223_LC_PARAMS() function
				if(m3g_OpenLC(handle, &h223OLCParams, M3G_E_VIDEO_CAPABILITY,(M3G_CAPABILITY*)pVideoCap) == M3G_SUCCESS) {
								sprintf(str,"m3g_OpenLC(%d) success for video",handle);
				} else {
								sprintf(str,"m3g_OpenLC(%d) failed for video",handle);
				}
				printandlog(index,M3G,NULL,str,0);
}

void getMatchedCapability(int index,int handle, M3G_CAPS_LIST *matchedCapsList)
{
				char str[MAX_STRING_SIZE];
				if (m3g_GetMatchedCaps(handle, matchedCapsList)==M3G_SUCCESS) {
								sprintf(str,"m3g_GetMatchedCaps() success for handle=%d",handle);
				} else {
								sprintf(str,"m3g_GetMatchedCaps() failed for handle=%d errMsg=%s",handle,ATDV_ERRMSGP(handle));
				}
				printandlog(index,M3G,NULL,str,0);
}
static void getMatchedCapabilityAndOpenLC(int index)
{
				//	M3G_AUDIO_CAPABILITY
				//	M3G_VIDEO_CAPABILITY

				//	getMatchedCapability(port[index].deviceInfo[M3G].devH,&port[index].h223CapsList);
				getMatchedCapability(index,port[index].deviceInfo[M3GVID].devH,&port[index].videoCapsList);
				getMatchedCapability(index,port[index].deviceInfo[M3GAUD].devH,&port[index].audioCapsList);
				/*
					 port[index].h223CapsList=getMatchedCapability(port[index].deviceInfo[M3G].devH);
					 port[index].videoCapsList=getMatchedCapability(port[index].deviceInfo[M3GVID].devH);
					 port[index].audioCapsList=getMatchedCapability(port[index].deviceInfo[M3GAUD].devH);
				 */

				sendOpenLC(index);
}
static void resetM3gDevice(int index,int handle)
{
				char str[MAX_STRING_SIZE];
				if (m3g_Reset(handle) !=M3G_SUCCESS) {
								sprintf(str,"m3g_Reset(%d) failed",handle);
				}
				else {
								sprintf(str,"m3g_Reset(%d) successfull",handle);
				}
				printandlog(index,M3G,NULL,str,0);
}
static void openM3gMmChannel()
{
				printf("openM3gMmChannel() num_devices=%d\n",num_devices);
				int index;
				for(index=0;index<num_devices;index++) {
								if( (port[index].techtype==ISDN) || (port[index].techtype==SS7) ) {
												gc_demo_open_m3g_channel(index);
												gc_demo_open_mm_channel(index);
								} else if(port[index].techtype==BICC) {
												gc_demo_open_m3g_channel(index);
								}
				}
}
static void stopM3gTrace(int index,int handle)
{
				if (m3g_StopTrace(handle)!=M3G_SUCCESS) {
						printandlog(index,M3G,NULL,"m3g_StopTrace failure",0);
				} else {
						printandlog(index,M3G,NULL,"m3g_StopTrace successfull",0);
				}
}
static void startM3gTrace(int index,int handle)
{
				M3G_TRACE_INFO trcInfo;
				INIT_M3G_TRACE_INFO(&trcInfo);
				
				char traceFileName[256];
				
			if(handle==m3gBoardH) {
				strcpy(traceFileName,"../log/gc_basic_video_m3g_trc.txt");
			} else { 
				sprintf(traceFileName,"../log/gc_basic_video_m3g_trc_%s.txt",port[index].deviceInfo[M3G].device_name);
			}

//				trcInfo.logfile = "../log/gc_basic_video_m3g_trc.txt";
				trcInfo.logfile = traceFileName;
				trcInfo.bitmask =M3G_TRACE_H245|M3G_TRACE_H223|M3G_TRACE_AUDIO|M3G_TRACE_VIDEO|M3G_TRACE_INTERNALS;
				if (m3g_StartTrace(handle, &trcInfo)!=M3G_SUCCESS) {
								printandlog(index,M3G,NULL,"m3g_StartTrace failed",0);
				}
				else {
								printandlog(index,M3G,NULL,"m3g_StartTrace successfull",0);
				}
}
static void processDmevPortConnect(int index)
{
	                        if(port[index].dmevPortConnectCtr==3 && port[index].bM3gAndCnfPartyConnected==true) {
                          modifyMedia(index,M3G_E_TX);
                          modifyMedia(index,M3G_E_TXRX);
                        }
                        if( (port[index].techtype==BICC || port[index].techtype==SIP) && port[index].dmevPortConnectCtr==2) {
                              startIpmMedia(index);
                        }

}
static void processDmevPortDisconnect(int index)
{
                        if(port[index].isCallDisconnectInProgress==true &&
                            ( (port[index].dmevPortDisconnectCtr==2 && port[index].techtype==SIP) ||
                              (port[index].dmevPortDisconnectCtr==5 && port[index].techtype==BICC) ||
                              (port[index].dmevPortDisconnectCtr==2 && port[index].callType==CALL_TYPE_AUDIO && port[index].techtype==BICC) )) {
                              disconnect_video_call(index);
                        }
                        if(port[index].isCallDisconnectInProgress==true &&
                            port[index].dmevPortDisconnectCtr==2 &&
                            port[index].techtype!=SIP &&
                            port[index].techtype!=BICC) {
                              disconnectM3gAndDti(index);
                        }
                        if(port[index].isCallDisconnectInProgress==false &&
                            port[index].dmevPortDisconnectCtr==3 &&
                              port[index].techtype!=SIP) {
                            port[index].dmevPortDisconnectCtr=0;
                            port[index].dmevPortConnectCtr=0;
#ifdef USE_CNF
                            routeUnrouteDevices(index,M3G,CNFPTY,ROUTE,DMFL_TRANSCODE_ON);
                            cnfAddParty(index);
#endif
                        }
                        if(port[index].isCallDisconnectInProgress==false && port[index].dmevPortDisconnectCtr==2 && port[index].techtype==SIP) {
                            port[index].dmevPortDisconnectCtr=0;
                            port[index].dmevPortConnectCtr=0;
#ifdef USE_CNF
                            routeUnrouteDevices(index,IPM,CNFPTY,ROUTE,DMFL_TRANSCODE_ON);
                            cnfAddParty(index);
#endif
//                            break;
                        }

}
static void stopMedia(int index)
{
				char str[MAX_STRING_SIZE];

				if(port[index].bMediaStarted==false) {
								return;
				}
		
				stopPlayRecord(index,true);
				stopPlayRecord(index,false);

				port[index].bMediaStarted=false;

				if (m3g_StopMedia(port[index].deviceInfo[M3GVID].devH)==M3G_SUCCESS) {
								sprintf(str,"m3g_StopMedia() success for m3g Video");
				} else {
								sprintf(str,"m3g_StopMedia() failed for m3g Video");
				}
				printandlog(index,M3G,NULL,str,0);

				if (m3g_StopMedia(port[index].deviceInfo[M3GAUD].devH)==M3G_SUCCESS) {
								sprintf(str,"m3g_StopMedia() success for m3g Audio");
				} else {
								sprintf(str,"m3g_StopMedia() failed for m3g Audio");
				}
				printandlog(index,M3G,NULL,str,0);
}
static void processModifyMediaCmplt(int index)
{
                        if(port[index].modifyMediaCtr==START_PLAY && port[index].bM3gAndMmConnected==true) {
                            sendH245VFU(index);
                            if(port[index].direction==DIR_IN) {
                              mmPlayRecordAudioVideoPrompt(index,true);
                            } else {
                              mmPlayRecordAudioVideoPrompt(index,false);
                            }
                        } else if(port[index].modifyMediaIdleCtr==2) {
                            port[index].bMediaStarted=false;
                            routeUnrouteDevices(index,M3G,MM,UNROUTE,DMFL_TRANSCODE_ON);
                            port[index].modifyMediaIdleCtr=0;
                        }

}
static void processStopMediaCmplt(int index, SRL_DEVICE_HANDLE devH)
{
#ifdef USE_CNF
                        if(devH==port[index].deviceInfo[M3GVID].devH && port[index].bPartyIntoCnf==true) {
                          cnfRemoveParty(index);
                        } else {
                          stopH245(index);
                        }
#else
                        stopH245(index);
#endif

}
static void stopH245(int index)
{
				char str[MAX_STRING_SIZE];

				if(port[index].bH245Started==false) {
								return;
				}
				port[index].bH245Started=false;

				int handle = port[index].deviceInfo[M3G].devH;
				if (m3g_StopH245(handle)==M3G_SUCCESS ) {
								sprintf(str,"m3g_StopH245(%d) successfull",handle);
				} else {
								sprintf(str,"m3g_StopH245(%d) failure",handle);
				}
				printandlog(index,M3G,NULL,str,0);
}
static void processStopH245Complete(int index)
{
	if(port[index].techtype==BICC) {
 				stopIpmMedia(index);
//				routeUnrouteDevices(index,IPM,M3G,UNROUTE,DMFL_TRANSCODE_NATIVE);
	}
	if(port[index].bPartyIntoCnf==true) {
				routeUnrouteDevices(index,M3G,CNFPTY,UNROUTE,DMFL_TRANSCODE_ON);
	} else {
				routeUnrouteDevices(index,M3G,MM,UNROUTE,DMFL_TRANSCODE_NATIVE);
	}
}
static void startH245(int index)
{
				char str[MAX_STRING_SIZE];


				if(port[index].bH245Started==true) {
								return;
				}
				port[index].bH245Started=true;


				M3G_H223_SESSION h223Session;
				INIT_M3G_H223_SESSION(&h223Session);

				int handle=port[index].deviceInfo[M3G].devH;

				h223Session.defaultH223MuxLevel=M3G_E_H223_MUX_LEVEL2;
				h223Session.maxALSDUSize=(unsigned int)1024;
				h223Session.isWNSRPEnabled=M3G_TRUE;
				h223Session.isMONAEnabled=M3G_FALSE;
				h223Session.isMultipleMsgsPerPdu=M3G_TRUE;

				if (m3g_StartH245(handle,&h223Session)!=M3G_SUCCESS ) {
								sprintf(str,"m3g_StartH245() failure handle=%d",handle);
				} else {
								sprintf(str,"m3g_StartH245() successfull handle=%d",handle);
				}
				printandlog(index,M3G,NULL,str,0);
}
static void printPortInfoList(int index,DM_PORT_INFO_LIST* pPortInfoList)
{
				char temp[512];
				char tmpBuff[128] = {0};
				char tokenBuff[10] = {0};
				int listCtr;
				int ctr;
				sprintf(temp,"DM_PORT_INFO_LIST(%d):----> ",pPortInfoList->unCount);
				printandlog(index,DM,NULL,temp,0);

				for (listCtr=0; listCtr< pPortInfoList->unCount; listCtr++ ) {
								for (ctr=0; ctr<16; ctr++ ) {
												sprintf(tokenBuff, "%.2x", (int)pPortInfoList->port_info[listCtr].device_ID[ctr]);
												strcat(tmpBuff,tokenBuff);
								}
//								sprintf(temp,"DM_PORT_INFO#%d devID=%s|portID=%s|mediaType=%s",listCtr,tmpBuff,(char *)(pPortInfoList->port_info[listCtr].port_ID),
//																portMediaTypeStr[pPortInfoList->port_info[listCtr].port_media_type]);
								sprintf(temp,"DM_PORT_INFO#%d portID=%s|mediaType=%s",listCtr,(char *)(pPortInfoList->port_info[listCtr].port_ID),
																portMediaTypeStr[pPortInfoList->port_info[listCtr].port_media_type]);
								printandlog(index,DM,NULL,temp,0);
				}
}
static void printPortInfo(int index,int devType)
{
		char temp[512];
		char tmpBuff[128] = {0};
		char tokenBuff[10] = {0};
		int ctr;

		DM_PORT_INFO* pPortInfo;


		if(port[index].techtype==BICC && (devType==IPM || devType==M3G)) {
			pPortInfo=&port[index].deviceInfo[devType].h223RxPortInfo;
			for (ctr=0; ctr<16; ctr++ ) {
                sprintf(tokenBuff, "%.2x",(int)pPortInfo->device_ID[ctr]);
                strcat(tmpBuff,tokenBuff);
    	}
    	sprintf(temp,"[%s]: portID=%s|mediaType=%s",port[index].deviceInfo[devType].device_name,
                              (char *)(pPortInfo->port_ID),portMediaTypeStr[pPortInfo->port_media_type]);
    	printandlog(index,DM,NULL,temp,0);

    	memset(&pPortInfo, 0, sizeof(DM_PORT_INFO));
    	pPortInfo=&port[index].deviceInfo[devType].h223TxPortInfo;

    	for (ctr=0; ctr<16; ctr++ ) {
                sprintf(tokenBuff, "%.2x",(int)pPortInfo->device_ID[ctr]);
                strcat(tmpBuff,tokenBuff);
    	}
    	sprintf(temp,"[%s]: portID=%s|mediaType=%s",port[index].deviceInfo[devType].device_name,
                              (char *)(pPortInfo->port_ID),portMediaTypeStr[pPortInfo->port_media_type]);
    	printandlog(index,DM,NULL,temp,0);
//			return;
		}

		pPortInfo=&port[index].deviceInfo[devType].videoRxPortInfo;

		for (ctr=0; ctr<16; ctr++ ) {
								sprintf(tokenBuff, "%.2x",(int)pPortInfo->device_ID[ctr]);
								strcat(tmpBuff,tokenBuff);
		}
//		sprintf(temp,"[%s]: devID=%s|portID=%s|mediaType=%s",port[index].deviceInfo[devType].device_name,tmpBuff,
//															(char *)(pPortInfo->port_ID),portMediaTypeStr[pPortInfo->port_media_type]);
		sprintf(temp,"[%s]: portID=%s|mediaType=%s",port[index].deviceInfo[devType].device_name,
															(char *)(pPortInfo->port_ID),portMediaTypeStr[pPortInfo->port_media_type]);
		printandlog(index,DM,NULL,temp,0);

		memset(&pPortInfo, 0, sizeof(DM_PORT_INFO));
		pPortInfo=&port[index].deviceInfo[devType].videoTxPortInfo;

		for (ctr=0; ctr<16; ctr++ ) {
								sprintf(tokenBuff, "%.2x",(int)pPortInfo->device_ID[ctr]);
								strcat(tmpBuff,tokenBuff);
		}
//		sprintf(temp,"[%s]: devID=%s|portID=%s|mediaType=%s",port[index].deviceInfo[devType].device_name,tmpBuff,
//															(char *)(pPortInfo->port_ID),portMediaTypeStr[pPortInfo->port_media_type]);
		sprintf(temp,"[%s]: portID=%s|mediaType=%s",port[index].deviceInfo[devType].device_name,
															(char *)(pPortInfo->port_ID),portMediaTypeStr[pPortInfo->port_media_type]);
		printandlog(index,DM,NULL,temp,0);
			
  	pPortInfo=&port[index].deviceInfo[devType].audioRxPortInfo;

    for (ctr=0; ctr<16; ctr++ ) {
                sprintf(tokenBuff, "%.2x",(int)pPortInfo->device_ID[ctr]);
                strcat(tmpBuff,tokenBuff);
    }
//    sprintf(temp,"[%s]: devID=%s|portID=%s|mediaType=%s",port[index].deviceInfo[devType].device_name,tmpBuff,
//                              (char *)(pPortInfo->port_ID),portMediaTypeStr[pPortInfo->port_media_type]);
    sprintf(temp,"[%s]: portID=%s|mediaType=%s",port[index].deviceInfo[devType].device_name,
                              (char *)(pPortInfo->port_ID),portMediaTypeStr[pPortInfo->port_media_type]);
    printandlog(index,DM,NULL,temp,0);

    memset(&pPortInfo, 0, sizeof(DM_PORT_INFO));
    pPortInfo=&port[index].deviceInfo[devType].audioTxPortInfo;

    for (ctr=0; ctr<16; ctr++ ) {
                sprintf(tokenBuff, "%.2x",(int)pPortInfo->device_ID[ctr]);
                strcat(tmpBuff,tokenBuff);
    }
//    sprintf(temp,"[%s]: devID=%s|portID=%s|mediaType=%s",port[index].deviceInfo[devType].device_name,tmpBuff,
//                              (char *)(pPortInfo->port_ID),portMediaTypeStr[pPortInfo->port_media_type]);
    sprintf(temp,"[%s]: portID=%s|mediaType=%s",port[index].deviceInfo[devType].device_name,
                              (char *)(pPortInfo->port_ID),portMediaTypeStr[pPortInfo->port_media_type]);
    printandlog(index,DM,NULL,temp,0);

    memset(&pPortInfo, 0, sizeof(DM_PORT_INFO));
}
static bool checkDevicesIsConnected(int index, int devType1, int devType2)
{
	char str[MAX_STRING_SIZE];
	bool retVal=false;
	sprintf(str,"checkDevicesIsConnected(%d,%s,%s)",index,devEvtTypeStr[devType1],devEvtTypeStr[devType2]);
	if ( ((devType1==M3G && devType2==MM) || (devType2==M3G && devType1==MM)) && port[index].bM3gAndMmConnected==true) {
			sprintf(str,"%s bM3gAndMmConnected=%d ",str,port[index].bM3gAndMmConnected);
			retVal=true;
	}
	if ( ((devType1==M3G && devType2==CNFPTY) || (devType2==M3G && devType1==CNFPTY)) && port[index].bM3gAndCnfPartyConnected==true) {
			sprintf(str,"%s bM3gAndCnfPartyConnected=%d ",str,port[index].bM3gAndCnfPartyConnected);
      retVal=true;
  }
	
	if ( ((devType1==IPM && devType2==M3G) || (devType2==IPM && devType1==M3G)) && port[index].bIpmAndM3gConnected==true) {
      sprintf(str,"%s bIpmAndM3gConnected=%d ",str,port[index].bIpmAndM3gConnected);
      retVal=true;
  }

  if ( ((devType1==IPM && devType2==MM) || (devType2==IPM && devType1==MM)) && port[index].bIpmAndMmConnected==true) {
			sprintf(str,"%s bIpmAndMmConnected=%d ",str,port[index].bIpmAndMmConnected);
      retVal=true;
  }
  if ( ((devType1==IPM && devType2==CNFPTY) || (devType2==IPM && devType1==CNFPTY)) && (port[index].bIpmAndCnfPartyConnected==true) ) {
			sprintf(str,"%s bIpmAndCnfPartyConnected=%d ",str,port[index].bIpmAndCnfPartyConnected);
      retVal=true;
  }

	printandlog(index,DM,NULL,str,0);
	return retVal;
}
static void setDeviceConnectionFlagOnOff(int index,int devType1, int devType2, int flag)
{
	char str[MAX_STRING_SIZE];
	sprintf(str,"setDeviceConnectionFlagOnOff(%d,%s,%s,%s)",index,devEvtTypeStr[devType1],devEvtTypeStr[devType2],((flag==ROUTE)?"ROUTE":"UNROUTE"));
// M3G and MM
		if ( (devType1==M3G && devType2==MM) || (devType2==M3G && devType1==MM)) {
				if(flag==ROUTE) {
       			port[index].bM3gAndMmConnected=true;
						sprintf(str,"%s bM3gAndMmConnected=%d ",str,port[index].bM3gAndMmConnected);
				} else {
       			port[index].bM3gAndMmConnected=false;
					sprintf(str,"%s bM3gAndMmConnected=%d ",str,port[index].bM3gAndMmConnected);
				}
		}
// IPM and MM
    if ( (devType1==IPM && devType2==MM) || (devType2==IPM && devType1==MM)) {
				if(flag==ROUTE) {
        		port[index].bIpmAndMmConnected=true;
						sprintf(str,"%s bIpmAndMmConnected=%d ",str,port[index].bIpmAndMmConnected);
    		} else {
        		port[index].bIpmAndMmConnected=false;
						sprintf(str,"%s bIpmAndMmConnected=%d ",str,port[index].bIpmAndMmConnected);
    		}
		}
// M3G and CNFPTY
    if ( (devType1==M3G && devType2==CNFPTY) || (devType2==M3G && devType1==CNFPTY)) {
				if(flag==ROUTE) {
        	port[index].bM3gAndCnfPartyConnected=true;
					sprintf(str,"%s bM3gAndCnfPartyConnected=%d ",str,port[index].bM3gAndCnfPartyConnected);
    		} else {
        	port[index].bM3gAndCnfPartyConnected=false;
					sprintf(str,"%s bM3gAndCnfPartyConnected=%d ",str,port[index].bM3gAndCnfPartyConnected);
    	}
		}
// IPM and CNFPTY
    if ((devType1==IPM && devType2==CNFPTY) || (devType2==IPM && devType1==CNFPTY)) {
				if(flag==ROUTE) {
        	port[index].bIpmAndCnfPartyConnected=true;
					sprintf(str,"%s bIpmAndCnfPartyConnected=%d ",str,port[index].bIpmAndCnfPartyConnected);
    		} else {
        	port[index].bIpmAndCnfPartyConnected=false;
					sprintf(str,"%s bIpmAndCnfPartyConnected=%d ",str,port[index].bIpmAndCnfPartyConnected);
    	}
		}
// IPM and M3G
    if ( (devType1==IPM && devType2==M3G) || (devType2==IPM && devType1==M3G)) {
			if(flag==ROUTE) {
        	port[index].bIpmAndM3gConnected=true;
					sprintf(str,"%s bIpmAndM3gConnected=%d ",str,port[index].bIpmAndM3gConnected);
    	} else {
        	port[index].bIpmAndM3gConnected=false;
					sprintf(str,"%s bIpmAndM3gConnected=%d ",str,port[index].bIpmAndM3gConnected);
    	}
		}
	printandlog(index,DM,NULL,str,0);
}

static void connectDevices(int index, int txDevType, int rxDevType, DM_PORT_CONNECT_INFO_LIST *portConnectInfoList1)
{
//	printandlog(index,DM,NULL,"connectDevices()",0);
	char str[MAX_STRING_SIZE];
	char audXcodeFlag[27];
	char vidXcodeFlag[27];

	if(txDevType==M3GAUD) {
		strcpy(audXcodeFlag,(portConnectInfoList1->port_connect_info[0].unFlags==DMFL_TRANSCODE_ON)?"TRANSCODE":"NATIVE");
		strcpy(vidXcodeFlag,"NA");
	} else if (txDevType==M3GVID) {
		strcpy(vidXcodeFlag,(portConnectInfoList1->port_connect_info[0].unFlags==DMFL_TRANSCODE_ON)?"TRANSCODE":"NATIVE");
		strcpy(audXcodeFlag,"NA");
	} else {
		strcpy(audXcodeFlag,(portConnectInfoList1->port_connect_info[0].unFlags==DMFL_TRANSCODE_ON)?"TRANSCODE":"NATIVE");
		strcpy(vidXcodeFlag,(portConnectInfoList1->port_connect_info[1].unFlags==DMFL_TRANSCODE_ON)?"TRANSCODE":"NATIVE");
	}
	if( dev_PortConnect(port[index].deviceInfo[txDevType].devH, portConnectInfoList1, &index)==DEV_SUCCESS ) {
         sprintf(str,"dev_PortConnect() success Tx(%s) - Rx(%s) AudXcodeFlag=%s|VidXcodeFlag=%s",
						port[index].deviceInfo[txDevType].device_name ,port[index].deviceInfo[rxDevType].device_name,
						audXcodeFlag,vidXcodeFlag);
  } else {
         sprintf(str,"dev_PortConnect() failed Tx(%s) - Rx(%s) AudXcodeFlag=%s|VidXcodeFlag=%s",
						port[index].deviceInfo[txDevType].device_name ,port[index].deviceInfo[rxDevType].device_name,
						audXcodeFlag,vidXcodeFlag);
  }
  printandlog(index,DM,NULL,str,0);
}
static void disconnectDevices(int index, int txDevType, int rxDevType, DM_PORT_CONNECT_INFO_LIST *portConnectInfoList1)
{
//	printandlog(index,DM,NULL,"disconnectDevices()",0);
	char str[MAX_STRING_SIZE];
	 char audXcodeFlag[27];
  char vidXcodeFlag[27];

  if(txDevType==M3GAUD) {
    strcpy(audXcodeFlag,(portConnectInfoList1->port_connect_info[0].unFlags==DMFL_TRANSCODE_ON)?"TRANSCODE":"NATIVE");
    strcpy(vidXcodeFlag,"NA");
  } else if (txDevType==M3GAUD) {
    strcpy(vidXcodeFlag,(portConnectInfoList1->port_connect_info[0].unFlags==DMFL_TRANSCODE_ON)?"TRANSCODE":"NATIVE");
    strcpy(audXcodeFlag,"NA");
  } else {
    strcpy(audXcodeFlag,(portConnectInfoList1->port_connect_info[0].unFlags==DMFL_TRANSCODE_ON)?"TRANSCODE":"NATIVE");
    strcpy(vidXcodeFlag,(portConnectInfoList1->port_connect_info[1].unFlags==DMFL_TRANSCODE_ON)?"TRANSCODE":"NATIVE");
  }


	if( dev_PortDisconnect(port[index].deviceInfo[txDevType].devH, portConnectInfoList1, &index)==DEV_SUCCESS ) {
         sprintf(str,"dev_PortDisconnect() success Tx(%s) - Rx(%s) AudXcodeFlag=%s|VidXcodeFlag=%s",
						port[index].deviceInfo[txDevType].device_name ,port[index].deviceInfo[rxDevType].device_name,
						audXcodeFlag,vidXcodeFlag);
  } else {
         sprintf(str,"dev_PortDisconnect() failed Tx(%s) - Rx(%s) AudXcodeFlag=%s|VidXcodeFlag=%s",
						port[index].deviceInfo[txDevType].device_name ,port[index].deviceInfo[rxDevType].device_name,
						audXcodeFlag,vidXcodeFlag);
  }
  printandlog(index,DM,NULL,str,0);
}

static void routeUnrouteDevices(int index, int devType1, int devType2, int routeFlag,int xcodeFlag)
{
		bool isDevConnected=false;
		
//		printandlog(index,DM,NULL,"routeUnrouteDevices()",0);

		isDevConnected=checkDevicesIsConnected(index,devType1,devType2);
		if( (isDevConnected==false && routeFlag==UNROUTE) || (routeFlag==ROUTE && isDevConnected==true) ){
			return;
		}
		setDeviceConnectionFlagOnOff(index,devType1,devType2,routeFlag);		

    DM_PORT_CONNECT_INFO_LIST portConnectInfoList={0};

    INIT_DM_PORT_CONNECT_INFO_LIST(&portConnectInfoList);
    portConnectInfoList.unCount = 2;

    INIT_DM_PORT_CONNECT_INFO(&portConnectInfoList.port_connect_info[0]);
    INIT_DM_PORT_CONNECT_INFO(&portConnectInfoList.port_connect_info[1]);
    portConnectInfoList.port_connect_info[0].unFlags = xcodeFlag;
    portConnectInfoList.port_connect_info[1].unFlags = xcodeFlag;

    INIT_DM_PORT_INFO(&portConnectInfoList.port_connect_info[0].port_info_tx);
    INIT_DM_PORT_INFO(&portConnectInfoList.port_connect_info[0].port_info_rx);
    INIT_DM_PORT_INFO(&portConnectInfoList.port_connect_info[1].port_info_tx);
    INIT_DM_PORT_INFO(&portConnectInfoList.port_connect_info[1].port_info_rx);

    if(routeFlag==ROUTE) {
				printPortInfo(index,devType1);
				printPortInfo(index,devType2);
    }

		// SKS first NbUP connection

		if(port[index].callType==CALL_TYPE_AUDIO &&  port[index].techtype==BICC) {
    		portConnectInfoList.port_connect_info[0].unFlags = DMFL_TRANSCODE_NATIVE;
		}

		if( ((devType1==M3G && devType2==IPM) || (devType2==M3G && devType1==IPM)) && port[index].techtype==BICC) {
			  portConnectInfoList.unCount = 1;
        memcpy(&(portConnectInfoList.port_connect_info[0].port_info_tx), &port[index].deviceInfo[devType1].h223TxPortInfo, sizeof(DM_PORT_INFO));
        memcpy(&(portConnectInfoList.port_connect_info[0].port_info_rx), &port[index].deviceInfo[devType2].h223RxPortInfo, sizeof(DM_PORT_INFO));
        if(routeFlag==ROUTE) {
          connectDevices(index, devType1, devType2, &portConnectInfoList);
        } else {
          disconnectDevices(index, devType1, devType2, &portConnectInfoList);
        }
        portConnectInfoList.unCount = 1;
        memcpy(&(portConnectInfoList.port_connect_info[0].port_info_tx), &port[index].deviceInfo[devType2].h223TxPortInfo, sizeof(DM_PORT_INFO));
        memcpy(&(portConnectInfoList.port_connect_info[0].port_info_rx), &port[index].deviceInfo[devType1].h223RxPortInfo, sizeof(DM_PORT_INFO));
        if(routeFlag==ROUTE) {
          connectDevices(index, devType2, devType1, &portConnectInfoList);
        } else {
          disconnectDevices(index, devType2, devType1, &portConnectInfoList);
        }
				return;
		}


		if(devType1==M3G) {
    		portConnectInfoList.unCount = 1;
    		memcpy(&(portConnectInfoList.port_connect_info[0].port_info_tx), &port[index].deviceInfo[devType1].audioTxPortInfo, sizeof(DM_PORT_INFO));
    		memcpy(&(portConnectInfoList.port_connect_info[0].port_info_rx), &port[index].deviceInfo[devType2].audioRxPortInfo, sizeof(DM_PORT_INFO));
    		if(routeFlag==ROUTE) {
					connectDevices(index, M3GAUD, devType2, &portConnectInfoList);
				} else {
					disconnectDevices(index, M3GAUD, devType2, &portConnectInfoList);
				}
    		portConnectInfoList.unCount = 1;
    		memcpy(&(portConnectInfoList.port_connect_info[0].port_info_tx), &port[index].deviceInfo[devType1].videoTxPortInfo, sizeof(DM_PORT_INFO));
    		memcpy(&(portConnectInfoList.port_connect_info[0].port_info_rx), &port[index].deviceInfo[devType2].videoRxPortInfo, sizeof(DM_PORT_INFO));
				if(routeFlag==ROUTE) {
					connectDevices(index, M3GVID, devType2, &portConnectInfoList);
				} else {
					disconnectDevices(index, M3GVID, devType2, &portConnectInfoList);
				}
		} else {
#if 0
				if(port[index].callType==CALL_TYPE_AUDIO && port[index].techtype==BICC) {
						portConnectInfoList.unCount = 1;
				} else {
    				portConnectInfoList.unCount = 2;
				}
#endif
    		memcpy(&(portConnectInfoList.port_connect_info[0].port_info_tx), &port[index].deviceInfo[devType1].audioTxPortInfo, sizeof(DM_PORT_INFO));
    		memcpy(&(portConnectInfoList.port_connect_info[0].port_info_rx), &port[index].deviceInfo[devType2].audioRxPortInfo, sizeof(DM_PORT_INFO));
    		memcpy(&(portConnectInfoList.port_connect_info[1].port_info_tx), &port[index].deviceInfo[devType1].videoTxPortInfo, sizeof(DM_PORT_INFO));
    		memcpy(&(portConnectInfoList.port_connect_info[1].port_info_rx), &port[index].deviceInfo[devType2].videoRxPortInfo, sizeof(DM_PORT_INFO));
    		if(routeFlag==ROUTE) {
					connectDevices(index, devType1, devType2, &portConnectInfoList);
				} else {
					disconnectDevices(index, devType1, devType2, &portConnectInfoList);
				}
		}

        // now ipm tx to mm rx
        INIT_DM_PORT_CONNECT_INFO_LIST(&portConnectInfoList);

        portConnectInfoList.unCount = 2;
        INIT_DM_PORT_CONNECT_INFO(&portConnectInfoList.port_connect_info[0]);
        INIT_DM_PORT_CONNECT_INFO(&portConnectInfoList.port_connect_info[1]);
        portConnectInfoList.port_connect_info[0].unFlags = xcodeFlag;
        portConnectInfoList.port_connect_info[1].unFlags = xcodeFlag;

		if(port[index].callType==CALL_TYPE_AUDIO &&  port[index].techtype==BICC) {
    		portConnectInfoList.port_connect_info[0].unFlags = DMFL_TRANSCODE_NATIVE;
		}

		if(devType2==M3G) {
        portConnectInfoList.unCount = 1;
        memcpy(&(portConnectInfoList.port_connect_info[0].port_info_tx), &port[index].deviceInfo[devType2].audioTxPortInfo, sizeof(DM_PORT_INFO));
        memcpy(&(portConnectInfoList.port_connect_info[0].port_info_rx), &port[index].deviceInfo[devType1].audioRxPortInfo, sizeof(DM_PORT_INFO));
        if(routeFlag==ROUTE) {
          connectDevices(index, M3GAUD, devType1, &portConnectInfoList);
        } else {
          disconnectDevices(index, M3GAUD, devType1, &portConnectInfoList);
        }
        portConnectInfoList.unCount = 1;
        memcpy(&(portConnectInfoList.port_connect_info[0].port_info_tx), &port[index].deviceInfo[devType2].videoTxPortInfo, sizeof(DM_PORT_INFO));
        memcpy(&(portConnectInfoList.port_connect_info[0].port_info_rx), &port[index].deviceInfo[devType1].videoRxPortInfo, sizeof(DM_PORT_INFO));
        if(routeFlag==ROUTE) {
          connectDevices(index, M3GVID, devType1, &portConnectInfoList);
        } else {
          disconnectDevices(index, M3GVID, devType1, &portConnectInfoList);
        }
    } else {
#if 0
					if(port[index].callType==CALL_TYPE_AUDIO && port[index].techtype==BICC) {
            	portConnectInfoList.unCount = 1;
        	} else {
        			portConnectInfoList.unCount = 2;
					}
#endif
        memcpy(&(portConnectInfoList.port_connect_info[0].port_info_tx), &port[index].deviceInfo[devType2].audioTxPortInfo, sizeof(DM_PORT_INFO));
        memcpy(&(portConnectInfoList.port_connect_info[0].port_info_rx), &port[index].deviceInfo[devType1].audioRxPortInfo, sizeof(DM_PORT_INFO));
        memcpy(&(portConnectInfoList.port_connect_info[1].port_info_tx), &port[index].deviceInfo[devType2].videoTxPortInfo, sizeof(DM_PORT_INFO));
        memcpy(&(portConnectInfoList.port_connect_info[1].port_info_rx), &port[index].deviceInfo[devType1].videoRxPortInfo, sizeof(DM_PORT_INFO));
        if(routeFlag==ROUTE) {
          connectDevices(index, devType2, devType1, &portConnectInfoList);
        } else {
          disconnectDevices(index, devType2, devType1, &portConnectInfoList);
        }
    }
}

static void sendIFrameRequest(int index)
{
   GC_PARM_BLKP l_gcParmBlk_mime = 0;
   GC_PARM_BLKP l_gcParmBlk_mime1 = 0;
   GC_PARM_BLKP l_gcParmBlk_info = 0;
	 const char* iFrame = "<?xml version=\"1.0\" encoding=\"utf-8\" ?><media_control><vc_primitive><to_encoder><picture_fast_update></picture_fast_update></to_encoder></vc_primitive></media_control>";
   char* l_bodyType = "Content-Type:application/media_control+xml"; // specify the body type

   printandlog(index,GC_APICALL,NULL,"sendIFrameRequest()",0);
   if ( gc_util_insert_parm_ref(&l_gcParmBlk_mime, IPSET_MIME, IPPARM_MIME_PART_TYPE, (unsigned char)(strlen(l_bodyType) + 1), l_bodyType) < 0 )
   {
      printandlog(index,GC_APIERR,NULL, "gc_util_insert_parm_ref(IPSET_MIME,IPPARM_MIME_PART_TYPE)",0);
   }

   // insert the body size
   if ( gc_util_insert_parm_val(&l_gcParmBlk_mime, IPSET_MIME, IPPARM_MIME_PART_BODY_SIZE, sizeof(unsigned long), strlen(iFrame)) < 0 )
   {
      printandlog(index,GC_APIERR,NULL, "gc_util_insert_parm_val(IPSET_MIME,IPPARM_MIME_PART_BODY_SIZE)",0);
   }

   // insert the body
   if ( gc_util_insert_parm_val(&l_gcParmBlk_mime, IPSET_MIME, IPPARM_MIME_PART_BODY, sizeof(unsigned long), (unsigned long)(iFrame)) < 0 )
   {
      printandlog(index,GC_APIERR,NULL, "gc_util_insert_parm_val(IPSET_MIME,IPPARM_MIME_PART_BODY)",0);
   }
  // insert the list of parmBlks into the top level parmBlk
   if ( gc_util_insert_parm_val(&l_gcParmBlk_mime1, IPSET_MIME, IPPARM_MIME_PART, sizeof(unsigned long), (unsigned long)l_gcParmBlk_mime) < 0 )
   {
      printandlog(index,GC_APIERR,NULL, "gc_util_insert_parm_val(IPSET_MIME,IPPARM_MIME_PART)",0);
   }

   // now set it on the device
   if ( gc_SetUserInfo(GCTGT_GCLIB_CRN, port[index].call[0].crn, l_gcParmBlk_mime1, GC_SINGLECALL) < 0 ) // for this call only
   {
      printandlog(index,GC_APIERR,NULL, "gc_SetUserInfo()",0);
   }

   // insert the message type
   if ( gc_util_insert_parm_val(&l_gcParmBlk_info, IPSET_MSG_SIP, IPPARM_MSGTYPE, sizeof(int), IP_MSGTYPE_SIP_INFO) < 0 )
   {
      printandlog(index,GC_APIERR,NULL, "gc_util_insert_parm_val(IPSET_MSG_SIP,IPPARM_MSGTYPE)",0);
   }

// SKS IP
	// assuming only crn is active otherwise user has to handle it
   if ( gc_Extension(GCTGT_GCLIB_CRN, port[index].call[0].crn, IPEXTID_SENDMSG, l_gcParmBlk_info, NULL, EV_ASYNC) < 0 )
   {
      printandlog(index,GC_APIERR,NULL, "gc_Extension()",0);
   }
   gc_util_delete_parm_blk(l_gcParmBlk_info);
   gc_util_delete_parm_blk(l_gcParmBlk_mime);
}

static void sendH245VFU(int index)
{
    M3G_H245_MISC_CMD h245MiscCmd;
    INIT_M3G_H245_MISC_CMD(&h245MiscCmd);

    h245MiscCmd.logicalChannelNumber = port[index].remoteVideoLogicalChannel;
    h245MiscCmd.h245MiscCmdType = M3G_E_FAST_UPDATE_PICTURE;
    h245MiscCmd.h245MiscCmdParams.noParams = NULL;

    if (m3g_SendH245MiscCmd(port[index].deviceInfo[M3G].devH, &h245MiscCmd)==M3G_SUCCESS) {
      printandlog(index,M3G,NULL,"m3g_SendH245MiscCmd(VFU) Video Fast Update success",0);
    } else {
      printandlog(index,M3G,NULL,"m3g_SendH245MiscCmd(VFU) Video Fast Update failed",0);
    }
}


static void storeRxPortInfo(int index,METAEVENT *metaevent)
{
  char str[MAX_STRING_SIZE];
	DM_PORT_INFO_LIST* pRxPortInfoList = (DM_PORT_INFO_LIST*)metaevent->evtdatap;
	unsigned int i;
  if(metaevent->evtdev==port[index].deviceInfo[M3GAUD].devH) {
      sprintf(str,"Storing Rx Audio port info for m3g device %s",port[index].deviceInfo[M3GAUD].device_name);
      memcpy(&port[index].deviceInfo[M3G].audioRxPortInfo, &pRxPortInfoList->port_info[0], sizeof(DM_PORT_INFO));
			printandlog(index,DM,NULL,str,0);
  } else if(metaevent->evtdev==port[index].deviceInfo[M3GVID].devH) {
       sprintf(str,"Storing Rx Video port info for m3g device %s",port[index].deviceInfo[M3GVID].device_name);
       memcpy(&port[index].deviceInfo[M3G].videoRxPortInfo, &pRxPortInfoList->port_info[0], sizeof(DM_PORT_INFO));
			printandlog(index,DM,NULL,str,0);
	} else if(metaevent->evtdev==port[index].deviceInfo[M3G].devH) {
       sprintf(str,"Storing Rx H223 port info for m3g device %s",port[index].deviceInfo[M3G].device_name);
       memcpy(&port[index].deviceInfo[M3G].h223RxPortInfo, &pRxPortInfoList->port_info[0], sizeof(DM_PORT_INFO));
      printandlog(index,DM,NULL,str,0);

  } else if(metaevent->evtdev==port[index].deviceInfo[MM].devH) {		// mm
			for ( i=0; i< pRxPortInfoList->unCount; i++ ) {
				switch ( pRxPortInfoList->port_info[i].port_media_type ) {
         		case DM_PORT_MEDIA_TYPE_AUDIO:
               sprintf(str,"Storing Rx Audio port info for mm device %s",port[index].deviceInfo[MM].device_name);
               memcpy(&port[index].deviceInfo[MM].audioRxPortInfo, &pRxPortInfoList->port_info[i], sizeof(DM_PORT_INFO));
								printandlog(index,DM,NULL,str,0);
            	break;
         		case DM_PORT_MEDIA_TYPE_VIDEO:
               sprintf(str,"Storing Rx Video port info for mm device %s",port[index].deviceInfo[MM].device_name);
               memcpy(&port[index].deviceInfo[MM].videoRxPortInfo, &pRxPortInfoList->port_info[i], sizeof(DM_PORT_INFO));
							printandlog(index,DM,NULL,str,0);
            	break;
					case DM_PORT_MEDIA_TYPE_NONE:
					case DM_PORT_MEDIA_TYPE_H223:
					case DM_PORT_MEDIA_TYPE_AUDIO_NATIVE: 
					case DM_PORT_MEDIA_TYPE_VIDEO_NATIVE: 
					case DM_PORT_MEDIA_TYPE_AUDIO_LINEAR: 
					case DM_PORT_MEDIA_TYPE_VIDEO_LINEAR:
					break;
			}
		}	// for loop for mm media type
  } else if(metaevent->evtdev==port[index].deviceInfo[IPM].devH) {   // ipm
      for ( i=0; i< pRxPortInfoList->unCount; i++ ) {
        switch ( pRxPortInfoList->port_info[i].port_media_type ) {
            case DM_PORT_MEDIA_TYPE_AUDIO:
               sprintf(str,"Storing Rx Audio port info for ipm device %s",port[index].deviceInfo[IPM].device_name);
               memcpy(&port[index].deviceInfo[IPM].audioRxPortInfo, &pRxPortInfoList->port_info[i], sizeof(DM_PORT_INFO));
                printandlog(index,DM,NULL,str,0);
              break;
            case DM_PORT_MEDIA_TYPE_VIDEO:
               sprintf(str,"Storing Rx Video port info for ipm device %s",port[index].deviceInfo[IPM].device_name);
               memcpy(&port[index].deviceInfo[IPM].videoRxPortInfo, &pRxPortInfoList->port_info[i], sizeof(DM_PORT_INFO));
              printandlog(index,DM,NULL,str,0);
              break;
          	case DM_PORT_MEDIA_TYPE_H223:
          	{
                sprintf(str,"Storing Rx H223 port info for ipm device %s",port[index].deviceInfo[IPM].device_name);
                memcpy(&port[index].deviceInfo[IPM].h223RxPortInfo, &pRxPortInfoList->port_info[i], sizeof(DM_PORT_INFO));
                printandlog(index,DM,NULL,str,0);
                break;
          	}
          case DM_PORT_MEDIA_TYPE_NONE:
          case DM_PORT_MEDIA_TYPE_AUDIO_NATIVE:
          case DM_PORT_MEDIA_TYPE_VIDEO_NATIVE:
          case DM_PORT_MEDIA_TYPE_AUDIO_LINEAR:
          case DM_PORT_MEDIA_TYPE_VIDEO_LINEAR:
          break;
      }
    } // for loop for cnf media type
  } // ipm
  else if(metaevent->evtdev==port[index].deviceInfo[CNFPTY].devH) {   // cnf party
      for ( i=0; i< pRxPortInfoList->unCount; i++ ) {
        switch ( pRxPortInfoList->port_info[i].port_media_type ) {
            case DM_PORT_MEDIA_TYPE_AUDIO:
						{
               sprintf(str,"Storing Rx Audio port info for cnfParty device %s",port[index].deviceInfo[CNFPTY].device_name);
               memcpy(&port[index].deviceInfo[CNFPTY].audioRxPortInfo, &pRxPortInfoList->port_info[i], sizeof(DM_PORT_INFO));
                printandlog(index,DM,NULL,str,0);
              break;
						}
            case DM_PORT_MEDIA_TYPE_VIDEO:
						{
               sprintf(str,"Storing Rx Video port info for cnfParty device %s",port[index].deviceInfo[CNFPTY].device_name);
               memcpy(&port[index].deviceInfo[CNFPTY].videoRxPortInfo, &pRxPortInfoList->port_info[i], sizeof(DM_PORT_INFO));
              printandlog(index,DM,NULL,str,0);
              break;
						}
					default:
              printandlog(index,DM,NULL,"UNKNOWN MEDIA TYPE",0);
          	break;
      }
    } // for loop for ipm media type
  } // cnf party
}
static void storeTxPortInfo(int index,METAEVENT *metaevent)
{
  char str[MAX_STRING_SIZE];
	DM_PORT_INFO_LIST	*pTxPortInfoList = (DM_PORT_INFO_LIST*)metaevent->evtdatap;
	unsigned int i;
  	if(metaevent->evtdev==port[index].deviceInfo[M3GAUD].devH) {
       sprintf(str,"Storing Tx Audio port info for m3g device %s",port[index].deviceInfo[M3GAUD].device_name);
       memcpy(&port[index].deviceInfo[M3G].audioTxPortInfo, &pTxPortInfoList->port_info[0], sizeof(DM_PORT_INFO));
								printandlog(index,DM,NULL,str,0);
  	} else if(metaevent->evtdev==port[index].deviceInfo[M3GVID].devH) {
       	sprintf(str,"Storing Tx Video port info for m3g device %s",port[index].deviceInfo[M3GVID].device_name);
       	memcpy(&port[index].deviceInfo[M3G].videoTxPortInfo, &pTxPortInfoList->port_info[0], sizeof(DM_PORT_INFO));
				printandlog(index,DM,NULL,str,0);
  	} else if(metaevent->evtdev==port[index].deviceInfo[M3G].devH) {
      	sprintf(str,"Storing Tx H223 port info for m3g device %s",port[index].deviceInfo[M3G].device_name);
       	memcpy(&port[index].deviceInfo[M3G].h223TxPortInfo, &pTxPortInfoList->port_info[0], sizeof(DM_PORT_INFO));
      	printandlog(index,DM,NULL,str,0);
  	} else if(metaevent->evtdev==port[index].deviceInfo[MM].devH) {
			for ( i=0; i< pTxPortInfoList->unCount; i++) {
				switch ( pTxPortInfoList->port_info[i].port_media_type ) {
         	case DM_PORT_MEDIA_TYPE_AUDIO:
               sprintf(str,"Storing Tx Audio port info for mm device %s",port[index].deviceInfo[MM].device_name);
               memcpy(&port[index].deviceInfo[MM].audioTxPortInfo, &pTxPortInfoList->port_info[i], sizeof(DM_PORT_INFO));
								printandlog(index,DM,NULL,str,0);
            	break;
         	case DM_PORT_MEDIA_TYPE_VIDEO:
               sprintf(str,"Storing Tx Video port info for mm device %s",port[index].deviceInfo[MM].device_name);
               memcpy(&port[index].deviceInfo[MM].videoTxPortInfo, &pTxPortInfoList->port_info[i], sizeof(DM_PORT_INFO));
								printandlog(index,DM,NULL,str,0);
            	break;
					case DM_PORT_MEDIA_TYPE_H223:
					case DM_PORT_MEDIA_TYPE_NONE:
					case DM_PORT_MEDIA_TYPE_AUDIO_NATIVE: 
					case DM_PORT_MEDIA_TYPE_VIDEO_NATIVE: 
					case DM_PORT_MEDIA_TYPE_AUDIO_LINEAR: 
					case DM_PORT_MEDIA_TYPE_VIDEO_LINEAR:
						break;
				}
			}
  	} else if(metaevent->evtdev==port[index].deviceInfo[IPM].devH) {   // ipm
      for ( i=0; i< pTxPortInfoList->unCount; i++ ) {
        switch ( pTxPortInfoList->port_info[i].port_media_type ) {
            case DM_PORT_MEDIA_TYPE_AUDIO:
               sprintf(str,"Storing Tx Audio port info for ipm device %s",port[index].deviceInfo[IPM].device_name);
               memcpy(&port[index].deviceInfo[IPM].audioTxPortInfo, &pTxPortInfoList->port_info[i], sizeof(DM_PORT_INFO));
                printandlog(index,DM,NULL,str,0);
              break;
            case DM_PORT_MEDIA_TYPE_VIDEO:
               sprintf(str,"Storing Tx Video port info for ipm device %s",port[index].deviceInfo[IPM].device_name);
               memcpy(&port[index].deviceInfo[IPM].videoTxPortInfo, &pTxPortInfoList->port_info[i], sizeof(DM_PORT_INFO));
              printandlog(index,DM,NULL,str,0);
              break;
					case DM_PORT_MEDIA_TYPE_H223:
					{
               	sprintf(str,"Storing Tx H223 port info for ipm device %s",port[index].deviceInfo[IPM].device_name);
               	memcpy(&port[index].deviceInfo[IPM].h223TxPortInfo, &pTxPortInfoList->port_info[i], sizeof(DM_PORT_INFO));
              	printandlog(index,DM,NULL,str,0);
								break;
					}
          case DM_PORT_MEDIA_TYPE_NONE:
          case DM_PORT_MEDIA_TYPE_AUDIO_NATIVE:
          case DM_PORT_MEDIA_TYPE_VIDEO_NATIVE:
          case DM_PORT_MEDIA_TYPE_AUDIO_LINEAR:
          case DM_PORT_MEDIA_TYPE_VIDEO_LINEAR:
          break;
      }
    } // for loop for ipm media type
  } // ipm
else if(metaevent->evtdev==port[index].deviceInfo[CNFPTY].devH) {   // cnf
      for ( i=0; i< pTxPortInfoList->unCount; i++ ) {
        switch ( pTxPortInfoList->port_info[i].port_media_type ) {
          case DM_PORT_MEDIA_TYPE_AUDIO:
					{
               	sprintf(str,"Storing Tx Audio port info for cnf device %s",port[index].deviceInfo[CNFPTY].device_name);
               	memcpy(&port[index].deviceInfo[CNFPTY].audioTxPortInfo, &pTxPortInfoList->port_info[i], sizeof(DM_PORT_INFO));
                printandlog(index,DM,NULL,str,0);
              	break;
					}
          case DM_PORT_MEDIA_TYPE_VIDEO:
					{
            		sprintf(str,"Storing Tx Video port info for cnf device %s",port[index].deviceInfo[CNFPTY].device_name);
              	memcpy(&port[index].deviceInfo[CNFPTY].videoTxPortInfo, &pTxPortInfoList->port_info[i], sizeof(DM_PORT_INFO));
              	printandlog(index,DM,NULL,str,0);
              	break;
					}
					case DM_PORT_MEDIA_TYPE_H223:
          case DM_PORT_MEDIA_TYPE_NONE:
          case DM_PORT_MEDIA_TYPE_AUDIO_NATIVE:
          case DM_PORT_MEDIA_TYPE_VIDEO_NATIVE:
          case DM_PORT_MEDIA_TYPE_AUDIO_LINEAR:
          case DM_PORT_MEDIA_TYPE_VIDEO_LINEAR:
          break;
      }
    } // for loop for cnf media type
  } // cnf
}

static void getTxRxPortInfo(int index,int handle)
{
  char str[MAX_STRING_SIZE];
   // retrieve Tx and Rx PStream ports:
   if (dev_GetTransmitPortInfo(handle, &index)!=DEV_SUCCESS ) {
		sprintf(str,"dev_GetTransmitPortInfo(%d) failure",handle);
   } else {
		sprintf(str,"dev_GetTransmitPortInfo(%d) successful",handle);
   }
	printandlog(index,DM,NULL,str,0);

   if (dev_GetReceivePortInfo(handle, &index)!=DEV_SUCCESS ) {
		sprintf(str,"dev_GetReceivePortInfo(%d) failure",handle);
   } else {
		sprintf(str,"dev_GetReceivePortInfo(%d) successful",handle);
   }
	printandlog(index,DM,NULL,str,0);
}
static void connectM3gAndDti(int index)
{
  char str[MAX_STRING_SIZE];
   
	if(port[index].bM3gAndDtiConnected==true) {
		return;
	}

  port[index].bM3gAndDtiConnected=true;

  if ( gc_GetResourceH(port[index].ldev, &port[index].networkH,GC_NETWORKDEVICE) != GC_SUCCESS ) {
     sprintf(str,"gc_GetResourceH(%s) failed",port[index].netdevice);
		// Handle Error
  } else {
     sprintf(str,"gc_GetResourceH(%s) successful",port[index].netdevice);
  }
  printandlog(index,GC_APICALL,NULL,str,0);
  sr_setparm(port[index].networkH, SR_USERCONTEXT, (void *)&port[index].index);
  if (dev_Connect(port[index].deviceInfo[M3G].devH, port[index].networkH, DM_FULLDUP, EV_ASYNC)!=DEV_SUCCESS ) {
  		sprintf(str,"dev_Connect() failed for 3G-DTI");
  } else {
     	sprintf(str,"dev_Connect() successful for 3G-DTI");
  }
	printandlog(index,DM,NULL,str,0);
}
static void disconnectM3gAndDti(int index)
{
  char str[MAX_STRING_SIZE];
  if (dev_Disconnect(port[index].deviceInfo[M3G].devH, EV_ASYNC)!=DEV_SUCCESS ) {
      sprintf(str,"dev_Disconnect() failed for 3G");
  } else {
      sprintf(str,"dev_Disconnect() successful for 3G");
  }
  printandlog(index,DM,NULL,str,0);

  if (dev_Disconnect(port[index].networkH, EV_ASYNC)!=DEV_SUCCESS ) {
      sprintf(str,"dev_Disconnect() failed for DTI");
  } else {
      sprintf(str,"dev_Disconnect() successful for DTI");
  }
  printandlog(index,DM,NULL,str,0);
}

static void storeLocalMediaInfo(int index,IPM_MEDIA_INFO *pMediaInfo)
{
  char str[MAX_STRING_SIZE];
  int ctr;
  for (ctr=0;ctr<pMediaInfo->unCount;ctr++)
  {
    switch (pMediaInfo->MediaData[ctr].eMediaType)
    {
      case MEDIATYPE_AUDIO_LOCAL_RTP_INFO:
			{
          port[index].localEPAudioPortNum=pMediaInfo->MediaData[ctr].mediaInfo.PortInfo.unPortId;
          strcpy(port[index].localEPIpAddress,pMediaInfo->MediaData[ctr].mediaInfo.PortInfo.cIPAddress);
          break;
			}
			case MEDIATYPE_VIDEO_LOCAL_RTP_INFO:
			{
          port[index].localEPVideoPortNum=pMediaInfo->MediaData[ctr].mediaInfo.PortInfo.unPortId;
          strcpy(port[index].localEPIpAddress,pMediaInfo->MediaData[ctr].mediaInfo.PortInfo.cIPAddress);
          break;
			}
			case MEDIATYPE_NBUP_LOCAL_RTP_INFO:
			{
          port[index].localEPNbUPPortNum=pMediaInfo->MediaData[ctr].mediaInfo.PortInfo.unPortId;
          strcpy(port[index].localEPIpAddress,pMediaInfo->MediaData[ctr].mediaInfo.PortInfo.cIPAddress);
          break;
			}
      default:
        break;
    }
  }
  sprintf(str,"localIP=%s|localAudioPort=%d|localVideoPort=%d|localNbUPPort=%d",
						port[index].localEPIpAddress,port[index].localEPAudioPortNum,port[index].localEPVideoPortNum,port[index].localEPNbUPPortNum);
  printandlog(index,MISC,NULL,str,0);
}
static void getLocalMediaInfo(int index)
{
  char str[MAX_STRING_SIZE];
  IPM_MEDIA_INFO mediaInfo;
  mediaInfo.unCount = 2;
  mediaInfo.MediaData[0].eMediaType = MEDIATYPE_AUDIO_LOCAL_RTP_INFO;
  mediaInfo.MediaData[1].eMediaType = MEDIATYPE_VIDEO_LOCAL_RTP_INFO;

  if(ipm_GetLocalMediaInfo(port[index].deviceInfo[IPM].devH, &mediaInfo, EV_ASYNC) == -1)
  {
    sprintf(str,"ipm_GetLocalMediaInfo failed for device name %s", port[index].deviceInfo[IPM].device_name);
  }
  else {
    sprintf(str,"ipm_GetLocalMediaInfo successful for device name %s", port[index].deviceInfo[IPM].device_name);
  }
  printandlog(index,MISC,NULL,str,0);

// now NbUP

  IPM_MEDIA_INFO mediaInfoNbUP;
  mediaInfoNbUP.unCount = 1;
  mediaInfoNbUP.MediaData[0].eMediaType = MEDIATYPE_NBUP_LOCAL_RTP_INFO;

  if(ipm_GetLocalMediaInfo(port[index].deviceInfo[IPM].devH, &mediaInfoNbUP, EV_ASYNC) == -1)
  {
    sprintf(str,"ipm_GetLocalMediaInfo NbUP failed for device name %s", port[index].deviceInfo[IPM].device_name);
  }
  else {
    sprintf(str,"ipm_GetLocalMediaInfo NbUP successful for device name %s", port[index].deviceInfo[IPM].device_name);
  }
  printandlog(index,MISC,NULL,str,0);


}

static int parseOfferedSdp(METAEVENT *metaeventp, struct channel *pline)
{
	if(pline->bSdpReceived==true) {
			return SDP_ALREADY_OFFERED;
	}
  printandlog(pline->index, MISC, NULL,"parseOfferedSdp()",0);
  char str[MAX_STRING_SIZE];
	int retVal=SDP_NOT_OFFERED;
//  printandlog(pline->index, MISC, NULL,"Processing incoming SDP",0);
  GC_PARM_BLK *pParmBlock = (GC_PARM_BLK*)(metaeventp->extevtdatap);
  GC_PARM_DATA_EXT parm_data_ext;
  INIT_GC_PARM_DATA_EXT(&parm_data_ext);

  while (gc_util_next_parm_ex(pParmBlock, &parm_data_ext) == 0)
  {
      switch (parm_data_ext.set_ID)
      {
        case IPSET_SDP:
        {
            switch (parm_data_ext.parm_ID)
            {
							case IPPARM_SDP_ANSWER:
              case IPPARM_SDP_OFFER:
              {
                  printandlog(pline->index, MISC, NULL,"SDP is available",0);
									retVal=SDP_OFFERED;
                  char *sdpData;
                  sdpData=calloc(parm_data_ext.data_size + 1,sizeof(char));
                  strncpy(sdpData,(char*)parm_data_ext.pData,parm_data_ext.data_size);
                  sdpData[parm_data_ext.data_size]='\0';
                  sprintf(str,"Offered SDP is \n%s",sdpData);
                  printandlog(pline->index, MISC, NULL, str,0);
                  retVal=storeSdpValues(sdpData,pline->index);
                  break;
              }
							
            }
            break;
        }
     }
  }
	return retVal;
}

static int buildApmMessage(int index)
{
	int length;
	int sdpLen;
	int secondByteFirst4Bit;
	int firstByteValue;
	int ctr;
	static char header[] = {
		0x85,0x81,0xc0,0x00,
		0x00,0x01,0x82,0x83,
		0x03,0x08,0xf4,0x83,			// 0xf4=116 bytes for sdp
		0x20,0x20
	};
	
	sdpLen=strlen(port[index].sdpText);

	port[index].biccApmMessage[0]=0x78;
	port[index].biccApmMessage[1]=(sdpLen+sizeof(header));

	for(length=2,ctr=0;ctr<sizeof(header);ctr++,length++)
	{	
		port[index].biccApmMessage[length]=header[ctr];
		if(ctr==10) {
			if((sdpLen+3)<128) {
				port[index].biccApmMessage[length]=(sdpLen+3+128);
				printf("sdpLen=%u|totalLength=%u|sdpLenInsideApm=%u\n",sdpLen,(sdpLen+sizeof(header)),(sdpLen+3+128));
			} else {
				secondByteFirst4Bit=((sdpLen+3)/128);
				firstByteValue=((sdpLen+3)-(secondByteFirst4Bit*128));
				printf("length first and second byte without adding extension bit firstByteValue=%u|secondByteFirst4Bit=%u\n",firstByteValue,secondByteFirst4Bit);
				
				port[index].biccApmMessage[length]=firstByteValue;
				length++;
				port[index].biccApmMessage[length]=(secondByteFirst4Bit+128);
				port[index].biccApmMessage[1]+=1;
				printf("sdpLen=%u|totalLength=%u|1stByteSdpLenInsideApm=0x%02x(%u)|2ndByteSdpLenInsideApm=0x%02x(%u)\n",
									sdpLen,port[index].biccApmMessage[1],firstByteValue,firstByteValue,
									port[index].biccApmMessage[length],port[index].biccApmMessage[length]);
			}
		}
	}
	
	for(ctr=0;ctr<strlen(port[index].sdpText);ctr++,length++)
  {
    port[index].biccApmMessage[length]=port[index].sdpText[ctr];
  }
	return length;
}

static void buildOfferAnswerSdp(int index)
{
	int vidPT,audPT,NbUpPT;
	vidPT=port[index].remoteEPVideoPayloadID;
	audPT=port[index].remoteEPAudioPayloadID;
	NbUpPT=port[index].remoteEPNbUPPayloadID;
	if(port[index].callType==CALL_TYPE_VIDEO) {
		if(port[index].techtype!=BICC) {
			#ifndef USE_MPEG4
					sprintf(port[index].sdpText,"v=0\no=gc_basic_call_model_video 0 0 IN IP4 %s\ns=Video Demonstration\nc=IN IP4 %s\nt=0 0\nm=video %d RTP/AVP 34\nb=AS:128\na=rtpmap:34 h263/90000\na=fmtp:34 QCIF=2\nm=audio %d RTP/AVP 8 101\na=rtpmap:8 pcma/8000\na=rtpmap:101 telephone-event/8000",
								port[index].localEPIpAddress,port[index].localEPIpAddress,port[index].localEPVideoPortNum,port[index].localEPAudioPortNum);
			#else 
					sprintf(port[index].sdpText,"v=0\no=gc_basic_call_model_video 0 0 IN IP4 %s\ns=Video Demonstration\nc=IN IP4 %s\nt=0 0\nm=video %d RTP/AVP %d\nb=AS:128\na=rtpmap:%d mp4v-es/90000\na=fmtp:%d profile-level-id=1; config=000001b008000001b50ea020202f000001000000012000c788ba9850584121463f\nm=audio %d RTP/AVP %d 101\na=rtpmap:%d pcma/8000\na=rtpmap:101 telephone-event/8000",
				port[index].localEPIpAddress,port[index].localEPIpAddress,port[index].localEPVideoPortNum,vidPT,vidPT,vidPT,port[index].localEPAudioPortNum,audPT,audPT);
			#endif
		} else if(port[index].techtype==BICC) {
				if(port[index].direction==DIR_IN) {
 					sprintf(port[index].sdpText,"v=0\r\no=- 0 0 IN IP4 %s\r\ns=bicc_nbup_gc_demo\r\nc=IN IP4 %s\r\nt=0 0\r\na=ipbcp:1 Accepted\r\nm=audio %d RTP/AVP %d\r\na=rtpmap:%d VND.3GPP.IUFP/16000\r\na=ptime:5\r\n",
							port[index].localEPIpAddress,port[index].localEPIpAddress,port[index].localEPNbUPPortNum,NbUpPT,NbUpPT);
				}else {
          sprintf(port[index].sdpText,"v=0\r\no=- 0 0 IN IP4 %s\r\ns=bicc_nbup_gc_demo\r\nc=IN IP4 %s\r\nt=0 0\r\na=ipbcp:1 Request\r\nm=audio %d RTP/AVP %d\r\na=rtpmap:%d VND.3GPP.IUFP/16000\r\na=ptime:5\r\n",
							port[index].localEPIpAddress,port[index].localEPIpAddress,port[index].localEPNbUPPortNum,NbUpPT,NbUpPT);
      }
		}
	} else {		// SKS AUDIO CALL
		char mAttrStr[MAX_STRING_SIZE];
		if(port[index].direction==DIR_IN) {
			switch(port[index].audioCodec) {
					case AMR:
//							sprintf(mAttrStr,"a=rtpmap:%d AMR/8000\r\na=fmtp:%d mode-set=%d;octet-align=1",
							sprintf(mAttrStr,"a=rtpmap:%d AMR/8000\r\na=fmtp:%d mode-set=%d",
									port[index].remoteEPAudioPayloadID,port[index].remoteEPAudioPayloadID,port[index].remoteEPAmrModeSet);
							break;
					case G711U_5MS:
							sprintf(mAttrStr,"a=rtpmap:0 pcmu/8000\r\na=ptime:5");
							break;
					case G711U_20MS:
							sprintf(mAttrStr,"a=rtpmap:0 pcmu/8000\r\na=ptime:20");
							break;
          case G711A_5MS:
							sprintf(mAttrStr,"a=rtpmap:8 pcma/8000\r\na=ptime:5");
              break;
          case G711A_20MS:
							sprintf(mAttrStr,"a=rtpmap:8 pcma/8000\r\na=ptime:20");
              break;
			}
		}
		if(port[index].techtype==SIP) {
				if(port[index].direction==DIR_IN) {
/*					// SKS TODO  uncomment when play using mm for SIP call, currently using dx to play prompt
 							sprintf(port[index].sdpText,"v=0\no=gc 0 0 IN IP4 %s\ns=Vid\nc=IN IP4 %s\nt=0 0\nm=audio %d RTP/AVP %d 101\n%s\na=rtpmap:101 telephone-event/8000\n",
														port[index].localEPIpAddress,port[index].localEPIpAddress,port[index].localEPAudioPortNum,port[index].remoteEPAudioPayloadID,mAttrStr);
*/
 					sprintf(port[index].sdpText,"v=0\no=gc 0 0 IN IP4 %s\ns=Vid\nc=IN IP4 %s\nt=0 0\nm=audio %d RTP/AVP 8 101\na=rtpmap:8 pcma/8000\na=rtpmap:101 telephone-event/8000\n",
														port[index].localEPIpAddress,port[index].localEPIpAddress,port[index].localEPAudioPortNum);
				} else {
 							sprintf(port[index].sdpText,"v=0\no=gc 0 0 IN IP4 %s\ns=Vid\nc=IN IP4 %s\nt=0 0\nm=audio %d RTP/AVP 8 0 97\na=rtpmap:8 pcma/8000\na=rtpmap:0 pcmu/8000\na=rtpmap:97 amr/8000\na=fmtp:97 mode-set=7;octet-align=1\na=rtpmap:101 telephone-event/8000",port[index].localEPIpAddress,port[index].localEPIpAddress,port[index].localEPAudioPortNum);
				}
		} else if(port[index].techtype==BICC) {
			if(port[index].direction==DIR_IN) {
 					sprintf(port[index].sdpText,"v=0\r\no=- 0 0 IN IP4 %s\r\ns=bicc_nbup_gc_demo\r\nc=IN IP4 %s\r\nt=0 0\r\na=ipbcp:1 Accepted\r\nm=audio %d RTP/AVP %d\r\n%s\r\n",
									port[index].localEPIpAddress,port[index].localEPIpAddress,port[index].localEPNbUPPortNum,port[index].remoteEPNbUPPayloadID,mAttrStr);
			} else {
// PCM a or u working
// 					sprintf(port[index].sdpText,"v=0\r\no=- 0 0 IN IP4 %s\r\ns=bicc_nbup_gc_demo\r\nc=IN IP4 %s\r\nt=0 0\r\na=ipbcp:1 Request\r\nm=audio %d RTP/AVP 8\r\na=ptime:5\r\na=rtpmap:8 pcma/8000\r\n",port[index].localEPIpAddress,port[index].localEPIpAddress,port[index].localEPNbUPPortNum);
// AMR only working
// 					sprintf(port[index].sdpText,"v=0\r\no=- 0 0 IN IP4 %s\r\ns=bicc_nbup_gc_demo\r\nc=IN IP4 %s\r\nt=0 0\r\na=ipbcp:1 Request\r\nm=audio %d RTP/AVP 97\r\na=ptime:5\r\na=rtpmap:97 AMR/8000\r\na=fmtp:97 mode-set=7;octet-align=1\r\n",port[index].localEPIpAddress,port[index].localEPIpAddress,port[index].localEPNbUPPortNum);


// working amr and g711u looks like hit maximum length of the SDP so it fails with additional amr infomration
// 					sprintf(port[index].sdpText,"v=0\r\no=- 0 0 IN IP4 %s\r\ns=bicc_nbup_gc_demo\r\nc=IN IP4 %s\r\nt=0 0\r\na=ipbcp:1 Request\r\nm=audio %d RTP/AVP 0 97\r\na=ptime:5\r\na=rtpmap:0 pcmu/8000\r\na=mid:1\r\na=rtpmap:97 AMR/8000\r\na=mid:2\r\n",port[index].localEPIpAddress,port[index].localEPIpAddress,port[index].localEPNbUPPortNum);
 					sprintf(port[index].sdpText,"v=0\r\no=- 0 0 IN IP4 %s\r\ns=bicc_nbup_gc_demo\r\nc=IN IP4 %s\r\nt=0 0\r\na=ipbcp:1 Request\r\nm=audio %d RTP/AVP 8 97\r\na=ptime:5\r\na=rtpmap:97 AMR/8000\r\na=mid:2\r\na=rtpmap:8 pcma/8000\r\na=mid:1\r\n",port[index].localEPIpAddress,port[index].localEPIpAddress,port[index].localEPNbUPPortNum);
			}
		}
	}
//	printf("SKS buildOfferAnswerSdp() SDP Value=%s\n",port[index].sdpText);
}

static int storeSdpValues(char *sdpVal,int index)
{
  char    str[MAX_STRING_SIZE];
	char 	codec[20];
  char *tempStr=NULL;
  char *mediaStr=NULL;
	char codec_name[35];
	int audioPT;
	int ptime;

	int callType=CALL_TYPE_AUDIO;

	strcpy(codec,"NULL");

  tempStr=strstr(sdpVal,"c=IN");
  if(tempStr!=NULL) {
      sscanf(tempStr,"c=IN IP4 %[0-9.]",port[index].remoteEPIpAddress);
						sprintf(str,"Connection Information found c=%s",port[index].remoteEPIpAddress);
     	printandlog(index, MISC, NULL,str, 0);
  } else {
     printandlog(index, MISC, NULL, "Connection Information not found", 0);
	}

	tempStr=NULL;
 tempStr=strstr(sdpVal,"a=ipbcp:");
	if(tempStr!=NULL) {
				sprintf(str,"IPBCP Information found");
    printandlog(index, MISC, NULL,str, 0);
	}

  tempStr=NULL;
  tempStr=strstr(sdpVal,"m=audio");
  if(tempStr!=NULL) {
      sscanf(tempStr,"m=audio %d",&port[index].remoteEPAudioPortNum);
			port[index].remoteEPNbUPPortNum=port[index].remoteEPAudioPortNum;
			sprintf(str,"Audio Information found port=%d",port[index].remoteEPAudioPortNum);
     	printandlog(index, MISC, NULL,str, 0);
  }  else {
     printandlog(index, MISC, NULL, "Audio Information not found", 0);
  }

  	tempStr=NULL;
		if(port[index].techtype==BICC) {
  		tempStr=strcasestr(sdpVal,"vnd.3gpp.iufp");
				if(tempStr!=NULL) {
						tempStr=NULL;
					 	tempStr=strstr(sdpVal,"a=rtpmap:");
     				sscanf(tempStr,"a=rtpmap:%d",&port[index].remoteEPNbUPPayloadID);
     				callType=CALL_TYPE_VIDEO;
    		} else if ( (port[index].direction==DIR_OUT) && (port[index].callType==CALL_TYPE_VIDEO) ) {
            printandlog(index, MISC, NULL, "Video media VND.3GPP.IUFP not found in Answer SDP, check answer SDP", 0);
            return NO_MATCHING_VIDEO;
        }
		} else {
  		tempStr=strstr(sdpVal,"m=video");
  		if(tempStr!=NULL) {
    			sscanf(tempStr,"m=video %d",&port[index].remoteEPVideoPortNum);
					callType=CALL_TYPE_VIDEO;
  		} else if ( (port[index].direction==DIR_OUT) && (port[index].callType==CALL_TYPE_VIDEO) ) {
     				printandlog(index, MISC, NULL, "Video media not found in Answer SDP, check answer SDP", 0);
						return NO_MATCHING_VIDEO;
				}
		}
	port[index].callType=callType;

	if((port[index].callType!=CALL_TYPE_VIDEO) && (port[index].techtype==BICC)) {
			int mediaPriority;
			bool bAudioFound=false;
			tempStr=NULL;
    	tempStr=strcasestr(sdpVal,"amr");
    	if(tempStr!=NULL) {
				printf("Found AMR codec\n");
      	mediaStr=strcasestr(tempStr,"a=mid:");
				if(mediaStr!=NULL) {
					sscanf(mediaStr,"a=mid:%d",&mediaPriority);
				} else {
					mediaPriority=1;
				}
				if(mediaPriority==1) {
					port[index].audioCodec=AMR;
					bAudioFound=true;
				}
			}		// if AMR
			if(bAudioFound==false) {
				tempStr=strcasestr(sdpVal,"pcma");
				if(tempStr!=NULL) {
					printf("Found PMCA codec\n");
        	mediaStr=strcasestr(tempStr,"a=mid:");
        	if(mediaStr!=NULL) {
          	sscanf(mediaStr,"a=mid:%d",&mediaPriority);
        	} else {
						mediaPriority=1;
					}
        	if(mediaPriority==1) {
          	port[index].audioCodec=PCMA;
						bAudioFound=true;
        	}
				}
			}
			if(bAudioFound==false) {
				tempStr=strcasestr(sdpVal,"pcmu");
        if(tempStr!=NULL) {
						printf("Found PMCU codec\n");
          	mediaStr=strcasestr(tempStr,"a=mid:");
          	if(mediaStr!=NULL) {
            		sscanf(mediaStr,"a=mid:%d",&mediaPriority);
          	} else {
							mediaPriority=1;
						}
          	if(mediaPriority==1) {
            		port[index].audioCodec=PCMU;
								bAudioFound=true;
          	}
					}
			} 
			if(bAudioFound==false) {
							tempStr=strstr(sdpVal,"rtpmap:");
          		if(tempStr==NULL) {
            			printandlog(index, MISC, NULL,"SDP Not Found",0);
          		} else {
            		sscanf(tempStr,"rtpmap:%d %[0-9A-Za-z-.]", &audioPT, codec_name);
            		sprintf(str,"pt=%d|codec_name=%s",audioPT,codec_name);
            		printandlog(index, MISC, NULL,str, 0);
            		printandlog(index, MISC, NULL, "Supported Audio Codec(aLaw) not offered in SDP", 0);
          		}
          		return NO_MATCHING_AUDIO;
			}// if not AMR, PCMA and PCMU

		tempStr=NULL;
		switch(port[index].audioCodec) 
		{
			case AMR:
			{
  			tempStr=strcasestr(sdpVal,"amr");
  			if(tempStr!=NULL) {
      		tempStr-=13;
      		if(sscanf(tempStr,"a=rtpmap:%d",&audioPT)<1) {
        		 tempStr++;
         		sscanf(tempStr,"a=rtpmap:%d",&audioPT);
      		}
					tempStr=NULL;
    			tempStr=strcasestr(sdpVal,"mode-set=");
					if(tempStr!=NULL) {
      			if(sscanf(tempStr,"mode-set=%d",&port[index].remoteEPAmrModeSet)<1) {
         			port[index].remoteEPAmrModeSet=7;
      			}
						if(port[index].remoteEPAmrModeSet<0 || port[index].remoteEPAmrModeSet>7) {
         			port[index].remoteEPAmrModeSet=7;
						}
					}
					port[index].remoteEPNbUPPayloadID=port[index].remoteEPAudioPayloadID=audioPT;
      		printandlog(index, MISC, NULL, "Found Audio Codec (AMR)", 0);
				} 
				break;
			}	
			case PCMA:
			{
  			tempStr=strcasestr(sdpVal,"a=rtpmap:8");
  			if(tempStr!=NULL) {
        	port[index].remoteEPNbUPPayloadID=port[index].remoteEPAudioPayloadID=8;
					tempStr=NULL;
					tempStr=strcasestr(sdpVal,"a=ptime:");
					if(tempStr!=NULL) {
						sscanf(sdpVal,"a=ptime:%d",&ptime);
					}
					if(ptime==20) {
						port[index].audioCodec=G711A_20MS;
					} else {
						port[index].audioCodec=G711A_5MS;
					}
    			printandlog(index, MISC, NULL, "Found Audio Codec (aLaw)", 0);
    		}
				break; 
			}
			case PCMU:
			{
    			tempStr=strcasestr(sdpVal,"a=rtpmap:0");
    			if(tempStr!=NULL) {
        		port[index].remoteEPNbUPPayloadID=port[index].remoteEPAudioPayloadID=0;
						tempStr=NULL;
						tempStr=strcasestr(sdpVal,"a=ptime:");
						if(tempStr!=NULL) {
							sscanf(sdpVal,"a=ptime:%d",&ptime);
						}
						if(ptime==20) {
							port[index].audioCodec=G711U_20MS;
						} else {
							port[index].audioCodec=G711U_5MS;
						}
        		printandlog(index, MISC, NULL, "Found Audio Codec (uLaw)", 0);
    			}
					break;
			}
			default:
        	printandlog(index, MISC, NULL, "Matching CODEC not found", 0);
					break;
		}
	}	// if BICC and audio call
//	if(callType==CALL_TYPE_AUDIO && port[index].techtype==SIP) {
	if(port[index].techtype==SIP) {
			tempStr=strcasestr(sdpVal,"a=rtpmap:8");
      if(tempStr!=NULL) {
        port[index].remoteEPNbUPPayloadID=port[index].remoteEPAudioPayloadID=8;
        tempStr=NULL;
        tempStr=strcasestr(sdpVal,"a=ptime:");
        if(tempStr!=NULL) {
          sscanf(sdpVal,"a=ptime:%d",&ptime);
        }
        if(ptime==20) {
            port[index].audioCodec=G711A_20MS;
        } else {
            port[index].audioCodec=G711A_5MS;
        }
        printandlog(index, MISC, NULL, "Found Audio Codec (aLaw)", 0);
			} else {
          tempStr=strstr(sdpVal,"rtpmap:");
          if(tempStr==NULL) {
            printandlog(index, MISC, NULL,"SDP Not Found",0);
          } else {
          	sscanf(tempStr,"rtpmap:%d %[0-9A-Za-z-.]", &audioPT, codec_name);
          	sprintf(str,"pt=%d|codec_name=%s",audioPT,codec_name);
          	printandlog(index, MISC, NULL,str, 0);
          	printandlog(index, MISC, NULL, "Supported Audio Codec(aLaw) not offered in SDP", 0);
          }
          return NO_MATCHING_AUDIO;
			}
	}	// if using SIP
	if(callType==CALL_TYPE_VIDEO && port[index].techtype==SIP) {
		tempStr=NULL;
#ifndef USE_MPEG4
  	tempStr=strstr(sdpVal,"a=rtpmap:34");
  	if(tempStr!=NULL) {
						port[index].remoteEPVideoPayloadID=34;
     	printandlog(index, MISC, NULL, "Found Video Codec found (H263)", 0);
		 	strcpy(codec,"H263");
  	} else {
			printandlog(index, MISC, NULL, "Supported Video Codec(H263) not offered in SDP", 0);
			return NO_MATCHING_VIDEO;
  	}
#else
  	tempStr=strcasestr(sdpVal,"mp4v-es");
  	if(tempStr!=NULL) {
						tempStr-=13;
     	if(sscanf(tempStr,"a=rtpmap:%d",&port[index].remoteEPVideoPayloadID)<1) {
							tempStr++;
							sscanf(tempStr,"a=rtpmap:%d",&port[index].remoteEPVideoPayloadID);
						}
     	printandlog(index, MISC, NULL, "Found Video Codec found (MPEG4)", 0);
		 			strcpy(codec,"MPEG4");
  	} else {
						printandlog(index, MISC, NULL, "Supported Video Codec(MPEG4) not offered in SDP", 0);
						return NO_MATCHING_VIDEO;
  	}
#endif
	}
	if(port[index].techtype==BICC) {
  				sprintf(str,"Remote Info: IP=%s|NbUP RTP PayloadID=%d|NbUP RTP Port=%d", port[index].remoteEPIpAddress,
								port[index].remoteEPNbUPPayloadID,port[index].remoteEPNbUPPortNum);
	} else {
		if(port[index].callType==CALL_TYPE_VIDEO) {
  			sprintf(str,"Remote Info: IP=%s|aud=aLaw(8)|audPort=%d|vid=%s(%d)|vidPort=%d", port[index].remoteEPIpAddress,
								port[index].remoteEPAudioPortNum,codec,port[index].remoteEPVideoPayloadID,port[index].remoteEPVideoPortNum);
		} else {
  			sprintf(str,"Remote Info: IP=%s|aud=aLaw(8)|audPort=%d", port[index].remoteEPIpAddress, port[index].remoteEPAudioPortNum);
		}
	}
  printandlog(index, MISC, NULL, str, 0);
	
	return SDP_PARSE_SUCCESS;
}

static void sendH245UII(int index)
{
  M3G_H245_UII h245UIIBuf;
	static int digit=1;
	char str[MAX_STRING_SIZE];
  char digBuf[2];
  sprintf(digBuf,"%d",digit);
	INIT_M3G_H245_UII(&h245UIIBuf);
  h245UIIBuf.numDigits = 1;
  h245UIIBuf.digitBuffer[0]=digBuf[0];

  if (m3g_SendH245UII(port[index].deviceInfo[M3G].devH, &h245UIIBuf)==M3G_SUCCESS) {
			sprintf(str,"m3g_SendH245UII(%d) successful",digBuf[0]);
      printandlog(index,M3G,NULL,str,0);
  }
  else {
      printandlog(index,M3G,NULL,"m3g_SendH245UII() successful",0);
  }
	digit++;
	if(digit>9) {
  	g_bDtmfSend=false;
	}
}

static void getLocalCapability(int index,int handle)
{
	char str[MAX_STRING_SIZE];

   if (m3g_GetLocalCaps(handle)!=M3G_SUCCESS) {
      sprintf(str,"m3g_GetLocalCaps() failure handle=%d",handle);
   } else {
      sprintf(str,"m3g_GetLocalCaps() successful Handle=%d",handle);
   }
	printandlog(index,M3G,NULL,str,0);
}
static void printLocalCapability(int index)
{
	char str[MAX_STRING_SIZE];
	int ctr;

	sprintf(str,"Channel Index# %d",index);
	printandlog(index,M3G,NULL,str,0);
	
	sprintf(str,"H223 Cap [ AL=%u | AL2Size=%u ]",
							port[index].simultaneousCapabilityList.pH223Capabilities->capability[0].h223Capability.adaptationLayerMedia,
							port[index].simultaneousCapabilityList.pH223Capabilities->capability[0].h223Capability.maxAL2SDUSize);
	printandlog(index,M3G,NULL,str,0);

	for(ctr=0;ctr<port[index].simultaneousCapabilityList.pAudioCapabilities->numCaps;ctr++) {
		sprintf(str,"Audio Cap#%u [coderType=%s|maxFramesPerSDU=%d|direction=%d]",ctr,
							strCoderType[port[index].simultaneousCapabilityList.pAudioCapabilities->capability[ctr].audioCapability.coderType],
							port[index].simultaneousCapabilityList.pAudioCapabilities->capability[ctr].audioCapability.maxFramesPerSDU,
							port[index].simultaneousCapabilityList.pAudioCapabilities->capability[ctr].audioCapability.direction);
		printandlog(index,M3G,NULL,str,0);
	}
	for(ctr=0;ctr<port[index].simultaneousCapabilityList.pVideoCapabilities->numCaps;ctr++) {
		sprintf(str,"Video Cap#%u [coderType=%s|maxBitRate=%d|direction=%d]",ctr,
							strCoderType[port[index].simultaneousCapabilityList.pVideoCapabilities->capability[ctr].videoCapability.coderType+2],
							port[index].simultaneousCapabilityList.pVideoCapabilities->capability[ctr].videoCapability.maxBitRate,
							port[index].simultaneousCapabilityList.pVideoCapabilities->capability[ctr].videoCapability.direction);	
		printandlog(index,M3G,NULL,str,0);
	}
}
static void setLocalCapability(int index)
{
	char str[MAX_STRING_SIZE];
	printLocalCapability(index);		// SKS print before setting it
   if(m3g_SetTCS(port[index].deviceInfo[M3G].devH, 1, &port[index].simultaneousCapabilityList)==M3G_SUCCESS) {
      	sprintf(str,"Successful m3g_SetTCS() completion - %ld",port[index].deviceInfo[M3G].devH);
   } else {
      	sprintf(str,"m3g_SetTCS() failure - %ld",port[index].deviceInfo[M3G].devH);
   }
	printandlog(index,M3G,NULL,str,0);
}
static void storeLocalCapability(int index,int handle,M3G_CAPS_LIST *pDefaultCapsList)
{
	char str[MAX_STRING_SIZE];
	int ctr;

	if(port[index].deviceInfo[M3G].devH==handle) {

//		M3G_CAPS_LIST  port[index].preferredH223Caps = {0};

		port[index].preferredH223Caps=*pDefaultCapsList;

		port[index].preferredH223Caps.version=M3G_LIBRARY_VERSION;
		port[index].preferredH223Caps.numCaps=1;
		port[index].preferredH223Caps.capabilityType=M3G_E_H223_CAPABILITY;
				
		port[index].preferredH223Caps.capability[0].h223Capability.version=M3G_LIBRARY_VERSION;
   	port[index].preferredH223Caps.capability[0].h223Capability.adaptationLayerMedia = M3G_AUDIO_AL2|M3G_VIDEO_AL2;
   	port[index].preferredH223Caps.capability[0].h223Capability.maxAL2SDUSize = 1024;
   	port[index].preferredH223Caps.capability[0].h223Capability.maxAL3SDUSize = 1024; 
   	port[index].preferredH223Caps.capability[0].h223Capability.frameH223AnnexA = M3G_TRUE; 
   	port[index].preferredH223Caps.capability[0].h223Capability.frameH223DoubleFlag = M3G_TRUE;
   	port[index].preferredH223Caps.capability[0].h223Capability.frameAnnexB = M3G_TRUE;
   	port[index].preferredH223Caps.capability[0].h223Capability.frameAnnexBWithHead = M3G_TRUE; 
   	port[index].preferredH223Caps.capability[0].h223Capability.rsCodeCapability = M3G_TRUE;
   	port[index].preferredH223Caps.capability[0].h223Capability.mobileOpXmitCap = M3G_TRUE; 
   	port[index].preferredH223Caps.capability[0].h223Capability.bitRate = 450;

   	port[index].simultaneousCapabilityList.pH223Capabilities = &port[index].preferredH223Caps;

		sprintf(str,"H223 Cap [ AL=%u | AL2Size=%u ]",
							port[index].simultaneousCapabilityList.pH223Capabilities->capability[0].h223Capability.adaptationLayerMedia,
							port[index].simultaneousCapabilityList.pH223Capabilities->capability[0].h223Capability.maxAL2SDUSize);
//		printandlog(index,M3G,NULL,str,0);
	}
	else if(port[index].deviceInfo[M3GAUD].devH==handle) {

  // 	M3G_CAPS_LIST    port[index].preferredAudioCaps = {0};
   	port[index].preferredAudioCaps.version = M3G_LIBRARY_VERSION;
   	port[index].preferredAudioCaps.numCaps = 0;
   	port[index].preferredAudioCaps.capabilityType = M3G_E_AUDIO_CAPABILITY;

   	for (ctr=0; ctr < pDefaultCapsList->numCaps; ctr++ ) {
      	if (pDefaultCapsList->capability[ctr].audioCapability.coderType==M3G_E_GSM_AMR_NB) {
				sprintf(str,"coderType=%s",strCoderType[pDefaultCapsList->capability[ctr].audioCapability.coderType]);
				printandlog(index,M3G,NULL,str,0);
            port[index].preferredAudioCaps.capability[port[index].preferredAudioCaps.numCaps].audioCapability=pDefaultCapsList->capability[ctr].audioCapability;
				port[index].preferredAudioCaps.capability[port[index].preferredAudioCaps.numCaps].audioCapability.maxFramesPerSDU=1;
            port[index].preferredAudioCaps.numCaps++;
      	}
   	}
   	port[index].simultaneousCapabilityList.pAudioCapabilities = &port[index].preferredAudioCaps;
	} 
	else if(port[index].deviceInfo[M3GVID].devH==handle) {
   	// configure the local video caps per the configured preferences:
//   	M3G_CAPS_LIST    port[index].preferredVideoCaps = {0};
   	port[index].preferredVideoCaps.version = M3G_LIBRARY_VERSION;
   	port[index].preferredVideoCaps.numCaps = 0;
   	port[index].preferredVideoCaps.capabilityType = M3G_E_VIDEO_CAPABILITY;

  		for(ctr=0; ctr < pDefaultCapsList->numCaps; ctr++ ) {
#ifndef USE_MPEG4
         if (pDefaultCapsList->capability[ctr].videoCapability.coderType==M3G_E_H263 )
#else 
         if (pDefaultCapsList->capability[ctr].videoCapability.coderType==M3G_E_MPEG4 )
#endif
				{
				sprintf(str,"coderType=%s",strCoderType[pDefaultCapsList->capability[ctr].videoCapability.coderType+2]);
				printandlog(index,M3G,NULL,str,0);
            port[index].preferredVideoCaps.capability[port[index].preferredVideoCaps.numCaps].videoCapability=pDefaultCapsList->capability[ctr].videoCapability;
				//port[index].preferredVideoCaps.capability[preferredVideoCaps.numCaps].videoCapability.maxBitRate=420;
				port[index].preferredVideoCaps.capability[port[index].preferredVideoCaps.numCaps].videoCapability.maxBitRate=560;
            port[index].preferredVideoCaps.numCaps++;
         }
   	}	// end of for loop
   	port[index].simultaneousCapabilityList.pVideoCapabilities = &port[index].preferredVideoCaps;
	}
	printandlog(index,M3G,NULL,str,0);
}

static void m3gSetParameter(int index,int handle)
{
	char str[MAX_STRING_SIZE];
	M3G_PARM_INFO parmInfo = {0};
	INIT_M3G_PARM_INFO(&parmInfo);

	parmInfo.parameterType = M3G_E_PRM_H245_TERMINAL_TYPE;
	parmInfo.parmValue.h245TerminalType = 127;
	if (m3g_SetParm(handle, &parmInfo) !=M3G_SUCCESS) {
         sprintf(str,"m3g_SetParm() failure on handle=%d",handle);
   } else {
         sprintf(str,"m3g_SetParm() successful on handle=%d",handle);
   }
	printandlog(index,M3G,NULL,str,0);

   parmInfo.parameterType = M3G_E_PRM_RELAY_DIGIT_TO_MEDIA_DEV;
   parmInfo.parmValue.boolParam= M3G_TRUE;
   if (m3g_SetParm(handle, &parmInfo) !=M3G_SUCCESS) {
         sprintf(str,"m3g_SetParm() failure on handle=%d",handle);
   } else {
         sprintf(str,"m3g_SetParm() successful on handle=%d",handle);
   }
	printandlog(index,M3G,NULL,str,0);

   parmInfo.parameterType = M3G_E_PRM_RELAY_FASTUPDATE_TO_H245;
   parmInfo.parmValue.boolParam =M3G_TRUE;
   if (m3g_SetParm(handle, &parmInfo)!=M3G_SUCCESS ) {
         sprintf(str,"m3g_SetParm() failure on handle=%d",handle);
   } else {
         sprintf(str,"m3g_SetParm() successful on handle=%d",handle);
   }
	printandlog(index,M3G,NULL,str,0);

 	parmInfo.parameterType = M3G_E_PRM_EARLY_MES;
   parmInfo.parmValue.boolParam= M3G_TRUE;
   if (m3g_SetParm(handle, &parmInfo) !=M3G_SUCCESS) {
         sprintf(str,"m3g_SetParm() failure on handle=%d",handle);
   } else {
         sprintf(str,"m3g_SetParm() successful on handle=%d",handle);
   }

	printandlog(index,M3G,NULL,str,0);
}
static void stopM3g(int index)
{
	if (m3g_Stop()!=M3G_SUCCESS) {
      printandlog(index,M3G,NULL,"m3g_Stop() failure",0);
   } else {
      printandlog(index,M3G,NULL,"m3g_Stop() Success",0);
	}
}
static int InitM3g(int index,int numOfEndPoints)
{
	char  str[MAX_STRING_SIZE];
  int nSRLMode = SR_POLLMODE;

	if(numOfEndPoints<=0) {
			return 0;
	}

	m3gBoardParmCtr=0;
	//dmevPortDisconnectCtr=0;
	//dmevPortConnectCtr=0;

   if ( sr_setparm(SRL_DEVICE, SR_MODEID, &nSRLMode) == -1 ) {
      printandlog(index,M3G,NULL,"Error sr_setparm(SRL_DEVICE, SR_MODEID, POLL) failure",0);
		exit(-1);
   }

   M3G_START_STRUCT startParams;
   INIT_M3G_START_STRUCT(&startParams);
   startParams.numEndpoints = numOfEndPoints; 

   if (m3g_Start(&startParams)!=M3G_SUCCESS) {
      printandlog(index,M3G,NULL,"m3g_Start() failure",0);
      return -1;
   }

	bM3gStarted=M3G_TRUE;
	sprintf(str,"m3g_Start() successful for %d devices",numOfEndPoints); 
   printandlog(index,M3G,NULL,str,0);

   // start and allocate resources for the M3G library session:
   // Initialize the specified board defaults:
   m3gBoardH = m3g_OpenEx("m3gB1", NULL, NULL, EV_SYNC);
   if ( m3gBoardH < 0 ) {
   	printandlog(index,M3G,NULL,"m3g_OpenEx(m3gB1) failed",0);
		return -1;
   }
   printandlog(index,M3G,NULL,"Successful m3g_OpenEx(m3gB1) completion",0);
//	 resetM3gDevice(BOARD_DEVICE,m3gBoardH);
	startM3gTrace(BOARD_DEVICE,m3gBoardH);
	m3gSetParameter(BOARD_DEVICE,m3gBoardH);
	m3gEnableDisableEvents(BOARD_DEVICE,m3gBoardH);
	openM3gMmChannel();
	return 0;
}

#ifdef USE_CNF
static void enable_disable_cnf_brd_dev_events(bool bFlag)
{
#if 0
	char str[MAX_STRING_SIZE];
	unsigned int BrdEventList[10];
	BrdEventList[0] = ECNF_BRD_EVT_CONF_OPENED;
	BrdEventList[1] = ECNF_BRD_EVT_CONF_CLOSED;
	BrdEventList[2] = ECNF_BRD_EVT_ACTIVE_TALKER;
	BrdEventList[3] = ECNF_BRD_EVT_PARTY_ADDED;
	BrdEventList[4] = ECNF_BRD_EVT_PARTY_REMOVED;
	BrdEventList[5] = ECNF_BRD_EVT_PARTY_OPENED;
	BrdEventList[6] = ECNF_BRD_EVT_PARTY_CLOSED;

	CNF_EVENT_INFO BrdEventInfo;
	BrdEventInfo.unEventCount = 7;
	BrdEventInfo.punEventList = BrdEventList;

	if(bFlag==true) {
		if (cnf_EnableEvents(cnfBrdH, &BrdEventInfo, NULL) == CNF_ERROR) {
   		sprintf(str,"cnf_EnableEvents(mcxB1) failed");
		} else {
   		sprintf(str,"cnf_EnableEvents(mcxB1) Success");
		}
	} else {
		if (cnf_DisableEvents(cnfBrdH, &BrdEventInfo, NULL) == CNF_ERROR) {
      sprintf(str,"cnf_DisableEvents(mcxB1) failed");
    } else {
      sprintf(str,"cnf_DisableEvents(mcxB1) Success");
    }
	}
  printandlog(BOARD_DEVICE,CNF,NULL,str,0);
#endif
}

static void enable_disable_cnf_dev_events(bool bFlag)
{
#if 0
	char str[MAX_STRING_SIZE];
	unsigned int CnfEventList[10];
	CnfEventList[0] = ECNF_CONF_EVT_PARTY_ADDED;
	CnfEventList[1] = ECNF_CONF_EVT_PARTY_REMOVED;
	CnfEventList[2] = ECNF_CONF_EVT_ACTIVE_TALKER;
	CNF_EVENT_INFO CnfEventInfo;
	CnfEventInfo.unEventCount = 3;
	CnfEventInfo.punEventList = CnfEventList;
	  if(bFlag==true) {
    if (cnf_EnableEvents(cnfH, &CnfEventInfo, NULL) == CNF_ERROR) {
      sprintf(str,"cnf_EnableEvents(cnfH) failed");
    } else {
      sprintf(str,"cnf_EnableEvents(cnfH) Success");
    }
  } else {
    if (cnf_DisableEvents(cnfH, &CnfEventInfo, NULL) == CNF_ERROR) {
      sprintf(str,"cnf_DisableEvents(cnfH) failed");
    } else {
      sprintf(str,"cnf_DisableEvents(cnfH) Success");
    }
  }
  printandlog(BOARD_DEVICE,CNF,NULL,str,0);
#endif
}



static void gc_demo_open_conference()
{
	int index=BOARD_DEVICE;
	if ((cnfH = cnf_OpenConference(cnfBrdH, NULL, NULL, NULL)) == CNF_ERROR)
	{
   	printandlog(BOARD_DEVICE,CNF,NULL,"cnf_OpenConference(mcxB1) failed",0);
		// add error handing
	} else {
   	printandlog(BOARD_DEVICE,CNF,NULL,"cnf_OpenConference(mcxB1) Success",0);
	}
	sr_setparm(cnfH, SR_USERCONTEXT, (void *)&index);
}

static void gc_demo_open_cnf_board()
{
	int index=BOARD_DEVICE;
	cnfBrdH=cnf_OpenEx( "mcxB1", NULL,NULL,EV_SYNC);
	if(cnfBrdH==CNF_ERROR) {
   	printandlog(BOARD_DEVICE,CNF,NULL,"cnf_OpenEx(mcxB1) failed",0);
		exit(-1);
	} else {
   	printandlog(BOARD_DEVICE,CNF,NULL,"cnf_OpenEx(mcxB1) Success",0);
	}
	sr_setparm(cnfBrdH, SR_USERCONTEXT, (void *)&index);

#if 0
// commented to allow user to run multiple application
	if(cnf_ResetDevices(cnfBrdH,NULL,NULL)==CNF_ERROR) {
     printandlog(index,CNF,NULL,"cnf_ResetDevices(mcxB1) failed",0);
  } else {
  	 printandlog(index,CNF,NULL,"cnf_ResetDevices(mcxB1) Success",0);
  }
#endif
 enable_disable_cnf_brd_dev_events(true);
 gc_demo_open_conference();
}

static void gc_demo_open_cnf_party()
{
	int index;
  for(index=0;index<num_devices;index++) {
			printandlog(index,CNF,NULL,"gc_demo_open_cnf_party()",0);
			char str[MAX_STRING_SIZE];
			if ((port[index].deviceInfo[CNFPTY].devH = cnf_OpenParty(cnfBrdH, NULL, NULL,(void *)port[index].index)) == CNF_ERROR) {
				sprintf(str,"cnf_OpenParty() - failed");
			} else {
				sprintf(str,"cnf_OpenParty()- Party#%d|PartyH=%ld Success",index,port[index].deviceInfo[CNFPTY].devH);
			}
   		printandlog(index, CNF, NULL, str,0);
   		sr_setparm(port[index].deviceInfo[CNFPTY].devH, SR_USERCONTEXT, (void *)&port[index].index);
	}
}

static void SetCnfLayout(eLB_LAYOUT_TYPE layoutType,eLB_LAYOUT_SIZE layoutSize)
{
		int nCtr;
    LB_FRAME_HANDLE l_regionList[MAX_REGION_COUNT];
    unsigned int l_regionListCount = MAX_REGION_COUNT;
 		
		layoutTemplateH = lb_CreateLayoutTemplate(layoutType);
   	
		if (layoutTemplateH == MTK_ERROR) {
   			printandlog(BOARD_DEVICE,CNF,NULL,"lb_CreateLayoutTemplate() failed",0);
   	} else {
   			printandlog(BOARD_DEVICE,CNF,NULL,"lb_CreateLayoutTemplate() Successful",0);
      	if((lb_GetRegionList(layoutTemplateH, l_regionList, &l_regionListCount))== MTK_ERROR) {
   				printandlog(BOARD_DEVICE,CNF,NULL,"lb_GetRegionList() failed",0);
      	} else {
   				printandlog(BOARD_DEVICE,CNF,NULL,"lb_GetRegionList() successful",0);
// SKS set cnfRegionListCount to applied template region count
					cnfRegionListCount=l_regionListCount;
         	for (nCtr=0;nCtr<l_regionListCount;nCtr++) {
            if((lb_SetSelectionMode(l_regionList[nCtr], eLB_SELECTION_MODE_USER_SELECT)) == MTK_ERROR) {
   							printandlog(BOARD_DEVICE,CNF,NULL,"lb_SetSelectionMode() failed",0);
               	break;
            } else {
   							printandlog(BOARD_DEVICE,CNF,NULL,"lb_SetSelectionMode() successful",0);
            }
            if((lb_SetDisplayMode (l_regionList[nCtr], eLB_DISPLAY_MODE_LIVE))== MTK_ERROR) {
   							printandlog(BOARD_DEVICE,CNF,NULL,"lb_SetDisplayMode() failed",0);
               	break;
            } else {
   							printandlog(BOARD_DEVICE,CNF,NULL,"lb_SetDisplayMode() Successful",0);
            }
         	}	// end of for loop
      	}
      	memset(&cnfLayoutInfo, 0, sizeof(CNF_VIDEO_LAYOUT_INFO));
      	cnfLayoutInfo.eLayoutSize = layoutSize;
      	cnfLayoutInfo.LayoutHandle = layoutTemplateH;
		 		if((cnf_SetVideoLayout(cnfH, &cnfLayoutInfo, 0)) == CNF_ERROR) {
       		printandlog(BOARD_DEVICE,CNF,NULL,"cnf_SetVideoLayout() failed",0);
     		} else {
        	printandlog(BOARD_DEVICE,CNF,NULL,"cnf_SetVideoLayout() Successful",0);
     		}
		}
}
static void GetCnfLayout()
{
	if( (cnf_GetVideoLayout(cnfH, NULL))==CNF_ERROR) {
			printandlog(BOARD_DEVICE,CNF,NULL,"cnf_GetVideoLayout() failed",0);
  } else {
      printandlog(BOARD_DEVICE,CNF,NULL,"cnf_GetVideoLayout() Successful",0);
  }
}
static void SetPartyAttributes(int index, bool bCoach)
{
	CNF_ATTR_INFO PtyAttrInfo;
	CNF_ATTR PtyAttrList[5];
	PtyAttrList[0].unAttribute = ECNF_PARTY_ATTR_TARIFF_TONE;
	PtyAttrList[0].unValue = ECNF_ATTR_STATE_ENABLED;
	if(bCoach==true) {
		PtyAttrList[1].unAttribute = ECNF_PARTY_ATTR_COACH;
		PtyAttrList[1].unValue = ECNF_ATTR_STATE_ENABLED;
	} else {
		PtyAttrList[1].unAttribute = ECNF_PARTY_ATTR_PUPIL;
		PtyAttrList[1].unValue = ECNF_ATTR_STATE_ENABLED;
	}
	PtyAttrInfo.unAttrCount = 2;
	PtyAttrInfo.pAttrList = PtyAttrList;
	if (cnf_SetAttributes(port[index].deviceInfo[CNFPTY].devH, &PtyAttrInfo, NULL) == CNF_ERROR) {
			printandlog(BOARD_DEVICE,CNF,NULL,"cnf_SetAttributes() failed",0);
  } else {
      printandlog(BOARD_DEVICE,CNF,NULL,"cnf_SetAttributes() Successful",0);
  }
}

static void GetCnfVisiblePartyList()
{
	if( (cnf_GetVisiblePartyList(cnfH,NULL))==CNF_ERROR) {
	      printandlog(BOARD_DEVICE,CNF,NULL,"cnf_GetVisiblePartyList() failed",0);
  } else {
      printandlog(BOARD_DEVICE,CNF,NULL,"cnf_GetVisiblePartyList() Successful",0);
  }
}

static void GetCnfPartyList()
{
  if( (cnf_GetPartyList(cnfH,NULL))==CNF_ERROR) {
        printandlog(BOARD_DEVICE,CNF,NULL,"cnf_GetPartyList() failed",0);
  } else {
      printandlog(BOARD_DEVICE,CNF,NULL,"cnf_GetPartyList() Successful",0);
  }
}

static void SetCnfPartyVisible(unsigned int partyCount,SRL_DEVICE_HANDLE *partyList)
{
	int nCtr;
	char str[MAX_STRING_SIZE];
	LB_FRAME_HANDLE RegionHandleList[MAX_REGION_COUNT];
	
	CNF_VISIBLE_PARTY_LIST VisiblePartyList;
	CNF_VISIBLE_PARTY_INFO VisiblePartyInfo[MAX_REGION_COUNT];
	size_t RegionHandleListSize = MAX_REGION_COUNT;

	eLB_DISPLAY_MODE displayMode=eLB_DISPLAY_MODE_BLANK;
	
	if (lb_GetRegionList(cnfLayoutH, RegionHandleList, &RegionHandleListSize) == MTK_SUCCESS) {
		if (RegionHandleListSize != cnfRegionListCount) {
			printandlog(BOARD_DEVICE,CNF,NULL,"Received invalid region handle list size",0);
		} else {
			for (nCtr = 0; nCtr < RegionHandleListSize; nCtr++) {
				if(nCtr<partyCount) {
					VisiblePartyInfo[nCtr].PartyHandle = partyList[nCtr];
					VisiblePartyInfo[nCtr].RegionHandle = RegionHandleList[nCtr];
					displayMode=eLB_DISPLAY_MODE_LIVE;
				} else {
					VisiblePartyInfo[nCtr].PartyHandle = -1;
					VisiblePartyInfo[nCtr].RegionHandle = RegionHandleList[nCtr];
					displayMode=eLB_DISPLAY_MODE_BLANK;
				}
				if((lb_SetDisplayMode (RegionHandleList[nCtr], displayMode))== MTK_ERROR) {
                printandlog(BOARD_DEVICE,CNF,NULL,"lb_SetDisplayMode() failed",0);
                break;
        } else {
                printandlog(BOARD_DEVICE,CNF,NULL,"lb_SetDisplayMode() Successful",0);
        }
				sprintf(str,"VisiblePartyListH=%ld|RegionHandle=%ld",VisiblePartyInfo[nCtr].PartyHandle,VisiblePartyInfo[nCtr].RegionHandle);
        printandlog(BOARD_DEVICE,CNF,NULL,str,0);
			}
			VisiblePartyList.unCount = nCtr;
			VisiblePartyList.pPartyList = VisiblePartyInfo;

			for(nCtr=0;nCtr<VisiblePartyList.unCount;nCtr++) {
				sprintf(str,"VisiblePartyList.pPartyList[%d].PartyHandle=%ld|VisiblePartyList.pPartyList[%d].RegionHandle=%ld",
										nCtr,VisiblePartyList.pPartyList[nCtr].PartyHandle,nCtr,VisiblePartyList.pPartyList[nCtr].RegionHandle);
        printandlog(BOARD_DEVICE,CNF,NULL,str,0);
			}

			if (cnf_SetVisiblePartyList(cnfH, &VisiblePartyList, NULL) == CNF_ERROR) {
        printandlog(BOARD_DEVICE,CNF,NULL,"cnf_SetVisiblePartyList() failed",0);
 	 		} else {
      	printandlog(BOARD_DEVICE,CNF,NULL,"cnf_SetVisiblePartyList() Successful",0);
  		}
		}
	}
}

static void cnfAddParty(int index)
{
   CNF_PARTY_INFO l_partyInfo;

   l_partyInfo.unVersion = CNF_PARTY_INFO_VERSION_0;
   l_partyInfo.unPartyCount = 1;
   l_partyInfo.pPartyList = &port[index].deviceInfo[CNFPTY].devH;

   if (cnf_AddParty(cnfH, &l_partyInfo, (void *)port[index].index) == CNF_ERROR) {
    printandlog(index, MISC, NULL, "cnf_AddParty() failure", 0);
   } else {
    printandlog(index, MISC, NULL, "cnf_AddParty() successful", 0);
		port[index].bPartyIntoCnf=true;
  }
}

static void cnfRemoveParty(int index)
{
   CNF_PARTY_INFO l_partyInfo;

   l_partyInfo.unVersion = CNF_PARTY_INFO_VERSION_0;
   l_partyInfo.unPartyCount = 1;
   l_partyInfo.pPartyList = &port[index].deviceInfo[CNFPTY].devH;

   if (cnf_RemoveParty(cnfH, &l_partyInfo, (void *)port[index].index) == CNF_ERROR) {
    printandlog(index, MISC, NULL, "cnf_RemoveParty() failure", 0);
   } else {
    printandlog(index, MISC, NULL, "cnf_RemoveParty() successful", 0);
  }
}
// SKS cnf related function defination ends here
#endif

static void gc_demo_open_mm_channel(int index)
{
	char str[MAX_STRING_SIZE];
	int ch=index+1;

	sprintf(port[index].deviceInfo[MM].device_name,"mmB1C%d",ch);
	//if ((port[index].deviceInfo[MM].devH = mm_Open(port[index].deviceInfo[MM].device_name, NULL,(void *) &port[index])) < 0) {
	if ((port[index].deviceInfo[MM].devH = mm_Open(port[index].deviceInfo[MM].device_name, NULL,NULL)) < 0) {
    	sprintf(str,"mm_Open(devicename=%s) Failed", port[index].deviceInfo[MM].device_name);
  	} else {
    	sprintf(str,"mm_Open(devicename=%s) successful", port[index].deviceInfo[MM].device_name);
  	}
   printandlog(index, MM, NULL, str,0);
   sr_setparm(port[index].deviceInfo[MM].devH, SR_USERCONTEXT, (void *)&port[index].index);
}
static void gc_demo_open_m3g_channel(int index)
{
	char	str[MAX_STRING_SIZE];
	int controlH,audioH,videoH;
	int ts=index+1;
	sprintf( port[index].deviceInfo[M3G].device_name,"m3gB1T%d",ts);

	
// SKS_SIMUL
	memset(&port[index].simultaneousCapabilityList, 0, sizeof(M3G_SIMULTANEOUS_CAP_SET));

   port[index].simultaneousCapabilityList.version = M3G_LIBRARY_VERSION;

	controlH = m3g_OpenEx(port[index].deviceInfo[M3G].device_name, NULL, (void *)&port[index],EV_ASYNC);
   if (controlH<0) {
		sprintf(str,"m3g_OpenEx(%s) failure", port[index].deviceInfo[M3G].device_name);
      printandlog(index, M3G, NULL, str,0);
   } else {
		sprintf(str,"Successful m3g_OpenEx(%s) completion Handle=%d", port[index].deviceInfo[M3G].device_name,controlH);
      printandlog(index,M3G,NULL,str,0);
		port[index].deviceInfo[M3G].devH=controlH;
   }
   sr_setparm(port[index].deviceInfo[M3G].devH, SR_USERCONTEXT, (void *)&port[index].index);
   sprintf(port[index].deviceInfo[M3GAUD].device_name, "%s:AUDIO1", port[index].deviceInfo[M3G].device_name);
   audioH = m3g_OpenEx(port[index].deviceInfo[M3GAUD].device_name, NULL,(void *) &port[index],EV_ASYNC);
   if ( audioH<0 ) {
		sprintf(str,"m3g_OpenEx(%s) failure",port[index].deviceInfo[M3GAUD].device_name);
      printandlog(index, M3G, NULL, str,0);
   } else {
		sprintf(str,"Successful m3g_OpenEx(%s) completion Handle=%d",port[index].deviceInfo[M3GAUD].device_name,audioH);
      printandlog(index,M3G,NULL,str,0);
		port[index].deviceInfo[M3GAUD].devH=audioH;
   }
   sr_setparm(port[index].deviceInfo[M3GAUD].devH, SR_USERCONTEXT, (void *)&port[index].index);
   sprintf(port[index].deviceInfo[M3GVID].device_name, "%s:VIDEO1", port[index].deviceInfo[M3G].device_name);
   videoH = m3g_OpenEx(port[index].deviceInfo[M3GVID].device_name, NULL,(void *) &port[index],EV_ASYNC);
   if ( videoH<0 ) {
			sprintf(str,"m3g_OpenEx(%s) failure",port[index].deviceInfo[M3GVID].device_name);
    	printandlog(index, M3G, NULL, str,0);
   } else {
			sprintf(str,"Successful m3g_OpenEx(%s) completion Handle=%d",port[index].deviceInfo[M3GVID].device_name,videoH);
      printandlog(index,M3G,NULL,str,0);
			port[index].deviceInfo[M3GVID].devH=videoH;
   }
   sr_setparm(port[index].deviceInfo[M3GVID].devH, SR_USERCONTEXT, (void *)&port[index].index);
}
#endif

static void resetVariables(int index)
{
				printandlog(index,M3G,NULL,"resetVariables()",0);
				
#ifdef USE_VIDEO
				port[index].remoteAudioLogicalChannel=0;
				port[index].remoteVideoLogicalChannel=0;

#ifdef USE_MPEG4
				port[index].remoteEPVideoPayloadID=102;
#else
				port[index].remoteEPVideoPayloadID=34;
#endif
			port[index].remoteEPAudioPayloadID=8;
			port[index].remoteEPAmrModeSet=7;
			port[index].remoteEPNbUPPayloadID=127;


				port[index].bPlayActive=false;
        port[index].bRecordActive=false;
				port[index].bMediaStarted=false;
				port[index].bIpmMediaStarted=false;
				port[index].bPartyIntoCnf=false;
				port[index].bH245Started=false;


				port[index].bM3gAndCnfPartyConnected=false;
				port[index].bIpmAndCnfPartyConnected=false;
				port[index].bM3gAndDtiConnected=false;
				port[index].bM3gAndMmConnected=false;
				port[index].bDxAndDtiConnected=false;

				port[index].txPortCtr=0;
				port[index].rxPortCtr=0;
				port[index].localCapCtr=0;
				port[index].olcEvtMask=0;
				port[index].modifyMediaCtr=0;
				port[index].modifyMediaIdleCtr=0;

				port[index].dmevPortDisconnectCtr=0;
				port[index].dmevPortConnectCtr=0;
				port[index].dmevConnectCtr=0;
				port[index].dmevDisconnectCtr=0;

				port[index].isCallDisconnectInProgress=false;

				port[index].openAudVidOlcCmplt=0;
				port[index].respondAudVidOlcCmplt=0;

				memset(&port[index].h223CapsList, 0, sizeof(M3G_CAPS_LIST));
				memset(&port[index].audioCapsList, 0, sizeof(M3G_CAPS_LIST));
				memset(&port[index].videoCapsList, 0, sizeof(M3G_CAPS_LIST));
#endif
}
/****************************************************************
*			NAME: gc_demo_open_isdn_channel(index)
* DESCRIPTION: initialize a port used by an ISDN protocol
*		  INPUT: index into port[] for this channel
*		RETURNS: exits with exitdemo() if not successful
*	  CAUTIONS: 1. This will need to be modified for some configurations/environments
*					2. Many parameters are "commented" out. The purpose of this is to 
*					   give you a template of how to change default parameters if you need
*						to do so
*					3. Don't be intimidated by the size of this procedure.
*						As previously indicated, much of this procedure has
*						commented out code which is intended to give examples
*						for those who need it. Many developers will not.
****************************************************************/
static void gc_demo_open_isdn_channel(int index)
{
	char						str[MAX_STRING_SIZE];
	GCLIB_MAKECALL_BLK	*gclib_makecallp;
	GC_PARM_BLKP			parmblkp = NULL;


	printandlog(index, MISC, NULL, "ISDN device being opened", 0);

	/* Initial channel state is blocked */
	port[index].blocked = YES; 
	port[index].techtype = ISDN;
	port[index].makecall_timeout = 0;							/* timeout parameter is ignored for ISDN */

	/* initializes the GC_PARM_BLK with the respective set_id, parm_id and the values */

	/*******
	-- To set the bearer channel transfer mode information
	-- Commented out as this is the default. Change if you wish to use
	-- other than the default value
	--	gc_util_insert_parm_val(&parmblkp, GCIS_SET_BEARERCHNL, GCIS_PARM_TRANSFERMODE, 
	--		sizeof(int), ISDN_ITM_CIRCUIT);
	*******/

	/*******
	-- To set the bearer channel transfer rate information
	-- Commented out as this is the default. Change if you wish to use 
	-- other than the default value
	--	gc_util_insert_parm_val(&parmblkp, GCIS_SET_BEARERCHNL, GCIS_PARM_TRANSFERRATE, 
	--		sizeof(int), BEAR_RATE_64KBPS);
	*******/

	/*******
	-- To set Layer 1 protocol to use on bearer channel
	-- Commented out as this is the default. Change if you wish to use
	-- other than the default value
	--	gc_util_insert_parm_val(&parmblkp, GCSET_CHAN_CAPABILITY, GCPARM_TYPE, 
	--		sizeof(int), GCCAPTYPE_AUDIO);
	*******/

	/*******
	-- To set Layer 1 protocol to use on bearer channel
	-- The default is not good for all our E1 boards, for example Springware
	-- WARNING: this will need to be changed for T1. Commenting out this
	-- function call will work in many T1 cases
	*******/
	gc_util_insert_parm_val(&parmblkp, GCSET_CHAN_CAPABILITY, GCPARM_CAPABILITY, 
		sizeof(int), GCCAP_AUDIO_g711Alaw64k);

	/*******
	-- To set User rate to use on bearer channel (layer 1 rate)
	--  Commented out as this is the default. Change if you wish to use
	-- other than the default value
	--	gc_util_insert_parm_val(&parmblkp, GCSET_CHAN_CAPABILITY, GCPARM_RATE, 
	--		sizeof(int), GCCAPRATE_DEFAULT);
	*******/

	/*******
	-- To support the specific feature or service requested from the network
	-- Commented out as this is not supported by E1 and DM3.
	--	gc_util_insert_parm_val(&parmblkp, GCIS_SET_FACILITY, GCIS_PARM_FACILITY_CODINGVALUE, 
	--		sizeof(int), 0x03);
	*******/

	/*******
	-- To support the specific feature or service requested from the network
	-- Commented out as this is not supported by E1 and DM3.
	--	gc_util_insert_parm_val(&parmblkp, GCIS_SET_FACILITY, GCIS_PARM_FACILITY_FEATURESERVICE, 
	--		sizeof(int), ISDN_SERVICE);
	*******/

	/*******
	-- Note: gclib_makecallp is not freed since a "permanent" data
	-- structure. Assume program exit will free the resource
	*******/
	gclib_makecallp
		= (GCLIB_MAKECALL_BLKP) malloc(sizeof(GCLIB_MAKECALL_BLK) + parmblkp->parm_data_size);
	if (gclib_makecallp == NULL) {
		printandlog(ALL_DEVICES, MISC, NULL, "Could not allocate memory\n", 0);
		exitdemo(1);
	}
	port[index].makecallblk.cclib = NULL;
	port[index].makecallblk.gclib = gclib_makecallp;

	/* Set the extra data in the makecall block */
	gclib_makecallp->ext_datap = parmblkp;

	/*******
	-- Set the called number plan to unknown 
	-- Modified, as this causes issues on some switches
	-- - UNKNOWN works on 99% of E1 switches
	--	gclib_makecallp->destination.address_plan = GCADDRPLAN_ISDN;
	*******/
	gclib_makecallp->destination.address_plan = GCADDRPLAN_UNKNOWN;

	/*******
	-- Set the address type to transparent
	-- Modified, as this causes issues on some switches 
	-- - TRANSPARENT works on 99% of E1 switches
	--	gclib_makecallp->destination.address_type = GCADDRTYPE_NAT;
	*******/
	//gclib_makecallp->destination.address_type = GCADDRTYPE_TRANSPARENT;
	
	// setting to international
	gclib_makecallp->destination.address_type = GCADDRTYPE_TRANSPARENT;

	/*******
	-- Removed sub-addresses, as this causes issues on some switches
	--	strcpy(gclib_makecallp->destination.sub_address, "456");
	--	gclib_makecallp->destination.sub_address_plan = 0;
	--	gclib_makecallp->destination.sub_address_type = GCSUBADDR_OSI;
	*******/
	gclib_makecallp->destination.sub_address[0] = '\x0';
	gclib_makecallp->destination.sub_address_plan = 0;
	gclib_makecallp->destination.sub_address_type = 0;

	/*******
	-- Note to customers: make sure that the Calling Address is set correctly, as some switches
	-- could reject the call if the calling address is not correct.
	*******/
	strcpy(gclib_makecallp->origination.address, "7654321");

	/*******
	-- Set the calling number plan to ISDN type
	-- Modified, as this causes issues on some switches
	-- gclib_makecallp->origination.address_plan = GCADDRPLAN_ISDN;
	*******/
	gclib_makecallp->origination.address_plan = GCADDRPLAN_UNKNOWN;

	/*******
   -- Set the address type
	-- Modified, as this causes issues on some switches
	-- gclib_makecallp->origination.address_type = GCADDRTYPE_NAT;
	*******/
	//gclib_makecallp->origination.address_type = GCADDRTYPE_TRANSPARENT;

	// Set to International 
	gclib_makecallp->origination.address_type = GCADDRTYPE_TRANSPARENT;

	/*******
	-- Removed sub-addresses, as this causes issues on some switches
	--	strcpy(gclib_makecallp->origination.sub_address, "456");
	-- gclib_makecallp->origination.sub_address_plan = 0;
	-- gclib_makecallp->origination.sub_address_type = GCSUBADDR_OSI;
	*******/
	gclib_makecallp->origination.sub_address[0] = '\x0';
	gclib_makecallp->origination.sub_address_plan = 0;
	gclib_makecallp->origination.sub_address_type = 0;

	gclib_makecallp->call_info.address_info = GCADDRINFO_ENBLOC;
	
	/* Open the line device */
	if (gc_OpenEx(&port[index].ldev, port[index].devname, EV_SYNC, (void *)&port[index]) != GC_SUCCESS) {
		sprintf(str, "gc_OpenEx(devicename=%s, mode=EV_SYNC) Failed", port[index].devname);
		printandlog(index, GC_APIERR, NULL, str, 0);
		exitdemo(1);
	}
	sprintf(str, "gc_OpenEx(devicename=%s, mode=EV_SYNC) Success", port[index].devname);
	printandlog(index, GC_APICALL, NULL, str, 0);

	/* Due to implementation, must wait until get GCEV_UNBLOCKED before issueing the reset linedevice */
	port[index].resetlinedev_required_after_unblocked = YES;

	/* Enable alarm notification */
	enable_alarm_notification(&port[index]);
} /* End of Function */

/****************************************************************
*			NAME: gc_demo_open_analog_channel(index)
* DESCRIPTION: initialize a port used by an Analog protocol
*		  INPUT: index into port[] for this channel
*		RETURNS: exits with exitdemo() if not successful
*	  CAUTIONS: none
****************************************************************/
static void gc_demo_open_analog_channel(int index)
{
	char		str[MAX_STRING_SIZE], msgbuffer[MAX_STRING_SIZE];


	printandlog(index, MISC, NULL, "Analog device being opened", 0);

	/* Initial channel state is blocked */
	port[index].blocked = YES; 
	port[index].techtype = ANALOG;
	port[index].makecall_timeout = 60;
	
	if (gc_OpenEx(&port[index].ldev, port[index].devname, EV_SYNC, (void *)&port[index]) != GC_SUCCESS) {
		sprintf(str, "gc_OpenEx(devicename=%s, mode=EV_SYNC) Failed", port[index].devname);
		printandlog(index, GC_APIERR, NULL, str, 0);
		exitdemo(1);
	}
	
	sprintf(str, "gc_OpenEx(devicename=%s, mode=EV_SYNC) Success", port[index].devname);
	printandlog(index, GC_APICALL, NULL, str, 0);

	if ((gc_LoadDxParm(port[index].ldev,"dxchan.vcp", msgbuffer, MAX_STRING_SIZE)) !=0) {
		sprintf(str, "gc_LoadDxParm(linedev=%ld) Failed)%s)", port[index].ldev, msgbuffer);
		printandlog(index, GC_APIERR, NULL, str, 0);
		exitdemo(1);
	}
	sprintf(str, "gc_LoadDxParm(linedev=%ld) Success)", port[index].ldev);
	printandlog(index, MISC, NULL, str, 0);

	/* Some implementations do not reset (i.e. go on-hook) if line is off-hook at open time */
	if (gc_ResetLineDev(port[index].ldev, EV_ASYNC) != GC_SUCCESS) {
		sprintf(str, "gc_ResetLineDev(linedev=%ld, EV_ASYNC) Failed", port[index].ldev);
		printandlog(index, GC_APIERR, NULL, str, 0);
		exitdemo(1);
	}

	port[index].resetlinedev_active = YES;
	sprintf(str, "gc_ResetLineDev(linedev=%ld, EV_ASYNC) Success", port[index].ldev);
	printandlog(index, GC_APICALL, NULL, str, 0);
} /* End of Function */

/****************************************************************
*			NAME: gc_demo_open_E1orT1_channel(index)
* DESCRIPTION: initialize a port used by an E1 or T1 protocol
*		  INPUT: index into port[] for this channel
*		RETURNS: exits with exitdemo if not successful
*	  CAUTIONS: none
****************************************************************/
static void gc_demo_open_E1orT1_channel(int index) 
{
	char		str[MAX_STRING_SIZE]; 

	printandlog(index, MISC, NULL, "E1 or T1 device being opened", 0);

	/* Initial channel state is blocked */
	port[index].blocked = YES; 
	port[index].techtype = E1ORT1_CAS;
	port[index].makecall_timeout = 60;							/* Use 1 minute */
	
	if (gc_OpenEx(&port[index].ldev, port[index].devname, EV_SYNC, (void *)&port[index]) != GC_SUCCESS) {
		sprintf(str, "gc_OpenEx(devicename=%s, mode=EV_SYNC) Failed", port[index].devname);
		printandlog(index, GC_APIERR, NULL, str, 0);
		exitdemo(1);
	}
	
	sprintf(str, "gc_OpenEx(devicename=%s, mode=EV_SYNC) Success", port[index].devname);
	printandlog(index, GC_APICALL, NULL, str, 0);

	set_calling_num(&port[index], "7654321");

	/* Enable alarm notification */
	enable_alarm_notification(&port[index]);
} /* End of function */

/****************************************************************
*			NAME: gc_demo_open_H323_channel(index)
* DESCRIPTION: initialize a port used by H323 IP protocol
*		  INPUT: index into port[] for this channel
*		RETURNS: exits with exitdemo if not successful
*	  CAUTIONS: none
****************************************************************/
static void gc_demo_open_H323_channel(int index)
{
	char				str[MAX_STRING_SIZE]; 

	printandlog(index, MISC, NULL, "H.323 device being opened", 0);

	/* Initial channel state is blocked */
	port[index].blocked = YES; 
	port[index].techtype = H323;
	port[index].makecall_timeout = 60;						/* Use 1 minute */

	if (gc_OpenEx(&port[index].ldev, port[index].devname, EV_ASYNC, (void *)&port[index]) != GC_SUCCESS) {
		sprintf(str, "gc_OpenEx(devicename=%s, mode=EV_ASYNC) Failed", port[index].devname);
		printandlog(index, GC_APIERR, NULL, str, 0);
		exitdemo(1);
	}

	sprintf(str, "gc_OpenEx(devicename=%s, mode=EV_ASYNC) Success", port[index].devname);
	printandlog(index, GC_APICALL, NULL, str, 0);
	
	/* Specify TX and RX coders for device */
	gc_set_channel_codecs(index);
	
} /* End of function */

/****************************************************************
*			NAME: gc_demo_open_SIP_channel(index)
* DESCRIPTION: initialize a port used by SIP protocol
*		  INPUT: index into port[] for this channel
*		RETURNS: exits with exitdemo if not successful
*	  CAUTIONS: none
****************************************************************/
static void gc_demo_open_SIP_channel(int index)
{
	char						 str[MAX_STRING_SIZE]; 
	GCLIB_MAKECALL_BLK	*gclib_makecallp;
	struct hostent			*phostent = NULL;
	struct in_addr			 addr;
	
	printandlog(index, MISC, NULL, "SIP device being opened", 0);

	/* Initial channel state is blocked */
	port[index].blocked = YES; 
	port[index].techtype = SIP;
	port[index].makecall_timeout = 60;						/* Use 1 minute */

	/* Note: gclib_makecallp is not freed since a "permanent" data */
	/* structure. Assume program exit will free the resource			*/
	gclib_makecallp = (GCLIB_MAKECALL_BLKP)
	malloc(sizeof(GCLIB_MAKECALL_BLK) );
	if (gclib_makecallp == NULL) {
		printandlog(ALL_DEVICES, MISC, NULL, "Could not allocate memory\n", 0);
		exitdemo(1);
	}

	port[index].makecallblk.cclib = NULL;
	port[index].makecallblk.gclib = gclib_makecallp;
	gclib_makecallp->ext_datap = NULL;

	/* set GCLIB_ADDRESS_BLK with destination string & type*/
	strcpy(gclib_makecallp->destination.address,port[index].destination_num);
	/* set the called number plan to Transparent type  */
	gclib_makecallp->destination.address_type = GCADDRTYPE_TRANSPARENT;

	/* set GCLIB_ADDRESS_BLK with local string & type*/
	/* retrieve local host name and address */
	if (gethostname(str, MAX_STRING_SIZE) != 0) {
		printandlog(index, GC_APIERR, NULL, "Failed to retrieve Host Name", 0);
		exitdemo(1);
	}

	if ((phostent = (struct hostent *)gethostbyname(str)) == 0) { 
		printandlog(index, GC_APIERR, NULL, "Failed to retrieve Host IP address", 0);
		exitdemo(1);
	}

	memcpy(&addr, phostent->h_addr_list[0], sizeof(struct in_addr));

	sprintf(gclib_makecallp->origination.address, "%s_%s@%s", port[index].netdevice, str, inet_ntoa(addr));
	/* set the called number plan to Transparent type  */
	gclib_makecallp->origination.address_type = GCADDRTYPE_TRANSPARENT;

	if (gc_OpenEx(&port[index].ldev, port[index].devname, EV_ASYNC, (void *)&port[index]) != GC_SUCCESS) {
		sprintf(str, "gc_OpenEx(devicename=%s, mode=EV_ASYNC) Failed", port[index].devname);
		printandlog(index, GC_APIERR, NULL, str, 0);
		exitdemo(1);
	}
	
	sprintf(str, "gc_OpenEx(devicename=%s, mode=EV_ASYNC) Success", port[index].devname);
	printandlog(index, GC_APICALL, NULL, str, 0);
} /* End of function */

/****************************************************************
*			NAME: gc_set_channel_codecs(index)
* DESCRIPTION: configure codecs for a VoIP (H.323 or SIP) channel 
*		  INPUT: index into port[] for this channel
*		RETURNS: exits with exitdemo if not successful
*	  CAUTIONS: none
****************************************************************/
static void gc_set_channel_codecs(index)
{
	int						i, j, scope;
	IP_CAPABILITY			ipcap;
	GC_PARM_BLK			  *parmblkp = NULL;
	char						str[MAX_STRING_SIZE];


	printandlog(index, MISC, NULL, "Specifying codecs", 0);

	for (i= 0; i< port[index].num_codecs; i++) {

		memset(&ipcap, 0, sizeof(IP_CAPABILITY));

		/* First obtain value of transmit codec type from translation table */
		for (j = 0; j< (sizeof(codecxlat)/sizeof(CODECXLAT)); j++) {
			if (!strcmp(port[index].codec[i].txcodec, codecxlat[j].codecname)) {
				ipcap.capability = codecxlat[j].codecvalue;
				break;
			}
		}

		/* if not found, log error and exit */
		if (ipcap.capability == 0) {
			printf("Invalid codec type entry in %s\n", cfgfile);
			exitdemo(1);
		}

		/* Specify TX codec, unless type is T.38 UDP which is bi-directional */
		/* and handled separately as data codec.                             */
		if (ipcap.capability != GCCAP_DATA_t38UDPFax) {
			ipcap.type = GCCAPTYPE_AUDIO;
			ipcap.direction = IP_CAP_DIR_LCLTRANSMIT;
			ipcap.extra.audio.frames_per_pkt = port[index].codec[i].txrate;
			/* Set VAD if and only if supported by codec type */
			if (codecxlat[j].isVADsupported == YES) {
				/* If codec supports VAD, use setting from config file. */
				/* Only check for "ON" for simplicity sake. */
				ipcap.extra.audio.VAD = (strcmp(port[index].codec[i].txVAD,"ON") == 0); 
			} else { /* VAD not supported */
				ipcap.extra.audio.VAD = NO;
			}

			/* append the GC_PARM_BLK with the respective TX codec */
			gc_util_insert_parm_ref(&parmblkp, GCSET_CHAN_CAPABILITY, IPPARM_LOCAL_CAPABILITY, 
											sizeof(IP_CAPABILITY), &ipcap);

			/* Re-initialize for determining receive codec */
			ipcap.capability = 0;
			/* Next obtain value of receive codec type from translation table */
			for (j = 0; j< (sizeof(codecxlat)/sizeof(CODECXLAT)); j++) {
				if (!strcmp(port[index].codec[i].rxcodec, codecxlat[j].codecname)) {
					ipcap.capability = codecxlat[j].codecvalue;
					break;
				}
			}

			/* If not found, log error and exit */
			if (ipcap.capability == 0) {
				printf("Invalid codec type entry in %s\n", cfgfile);
				exitdemo(1);
			}

			ipcap.direction = IP_CAP_DIR_LCLRECEIVE;
			ipcap.extra.audio.frames_per_pkt = port[index].codec[i].rxrate;
			/* Set VAD if and only if supported by codec type */
			if (codecxlat[j].isVADsupported) {
				/* string validation minimized for simplicity sake */
				ipcap.extra.audio.VAD = (strcmp(port[index].codec[i].rxVAD,"ON") == 0); 
			} else { /* VAD not supported */
				ipcap.extra.audio.VAD = NO;
			}

			/* append the GC_PARM_BLK with the respective RX codec */
			gc_util_insert_parm_ref(&parmblkp, GCSET_CHAN_CAPABILITY, IPPARM_LOCAL_CAPABILITY, 
									sizeof(IP_CAPABILITY), &ipcap);
					

		} else {		/* else codec is T.38 UDP which is bi-directional data codec */
			ipcap.type = GCCAPTYPE_RDATA;
			ipcap.direction = IP_CAP_DIR_LCLTXRX;
			ipcap.extra.data.max_bit_rate = 144;

			/* append the GC_PARM_BLK with the T.38 codec */
			gc_util_insert_parm_ref(&parmblkp, GCSET_CHAN_CAPABILITY, IPPARM_LOCAL_CAPABILITY, 
											sizeof(IP_CAPABILITY), &ipcap);
		}	
	}	/* end for loop for all codecs specified for device */

	/* To interoperate with Microsoft Phoenix Version 5.0 Soft SIP phone,*/
	/* we will set coders on SIP calls on a per call basis.  Again, this */
	/* is not required for other SIP clients.                            */
	if(port[index].techtype == H323) {
		scope = GC_ALLCALLS;
	} else {  /* else SIP */
		scope = GC_SINGLECALL;
	}

	if (gc_SetUserInfo(GCTGT_GCLIB_CHAN, port[index].ldev, parmblkp, scope) != GC_SUCCESS) {
							sprintf(str, "gc_SetUserInfo(linedev=%ld) Failed when specifying coders", port[index].ldev);
							printandlog(index, GC_APIERR, NULL, str, 0);
							exitdemo(1);
	}

	gc_util_delete_parm_blk(parmblkp);
} /* End of function */

/****************************************************************
*			NAME: gc_demo_open_ss7_channel(index)
* DESCRIPTION: initialize a port used by an SS7 protocol
*		  INPUT: index into port[] for this channel
*		RETURNS: exits with exitdemo if not successful
*	  CAUTIONS: none
****************************************************************/
static void gc_demo_open_ss7_channel(int index)
{
	char		str[MAX_STRING_SIZE];
	char		temp_devname[MAX_DEVNAME];
	printandlog(index, MISC, NULL, "SS7 device being opened", 0);

	/* Initial channel state is blocked */
	port[index].blocked = YES; 

	if ( (strstr(port[index].protname, "ss7") != NULL) ||
       (strstr(port[index].protname, "SS7") != NULL) ) {
		port[index].techtype = SS7;
	} else {
		port[index].techtype = BICC;
	}

	port[index].makecall_timeout = 60;							/* timeout is supported, try 1 minute */
	
	/* 
	NOTE:
	The SS7 MakeCallBlock is not initialized currently, as Gc_MakeCallBlk->cclib is set to 
	Null to use default values. For Live Testing, initialize the SS7MakeCallBlock and set 
	its address to Gc_MakeCallBlk->cclib field.
	*/

   sprintf(temp_devname,":L_SS7%s", port[index].devname);
	if (gc_OpenEx(&port[index].ldev, temp_devname, EV_SYNC, (void *)&port[index]) != GC_SUCCESS) {
		sprintf(str, "gc_OpenEx(devicename=%s, mode=EV_SYNC) Failed", temp_devname);
		printandlog(index, GC_APIERR, NULL, str, 0);
		exitdemo(1);
	}
	
	sprintf(str, "gc_OpenEx(devicename=%s, mode=EV_SYNC) Success", temp_devname);
	printandlog(index, GC_APICALL, NULL, str, 0);

	set_calling_num(&port[index], "7654321");

	GC_PARM gcParm;

	if(port[index].techtype==BICC) {
		gcParm.intvalue = 1;  // true
    if (gc_SetParm (port[index].ldev, GCPR_UNKNOWN_ISUP_MSGS, gcParm) == GC_SUCCESS) {
    		printandlog (index, GC_APICALL, NULL, "gc_SetParm(GCPR_UNKNOWN_ISUP_MSGS,) Success", 0);
		} else {
    		printandlog (index, GC_APIERR, NULL, "gc_SetParm(GCPR_UNKNOWN_ISUP_MSGS,) Failed", 0);
  	}

	}

  gcParm.longvalue=8;
  if(gc_SetParm(port[index].ldev,GCPR_RECEIVE_INFO_BUF,gcParm)==GC_SUCCESS) {
      printandlog(index, GC_APICALL, NULL, "gc_SetParm(GCPR_RECEIVE_INFO_BUF) to 8 successful", 0);
  } else {
      printandlog(index, GC_APIERR, NULL, "gc_SetParm(GCPR_RECEIVE_INFO_BUF) to 8 failed", 0);
	}
} /* End of function */ 

/***************************************************************************
 *			NAME: void set_calling_num(struct channel *pline, char *phone_num)
 *DESCRIPTION: Sets the calling number for the outbound direction
 *		  INPUT: pline - pointer to entry in port table
 *					phone_num - pointer to the phone number
 *		RETURNS: None.
  ***************************************************************************/
static void set_calling_num(struct channel *pline, char *phone_num)
{
	char		str[MAX_STRING_SIZE];

	if (pline->direction == DIR_OUT) {
		if (gc_SetCallingNum(pline->ldev, phone_num) < 0) {
			sprintf(str, "gc_SetCallingNum(linedev=%ld, phone_num = %s) Failed", pline->ldev, phone_num);
			printandlog(pline->index, GC_APIERR, NULL, str, 0);
			exitdemo(1);
		}
		sprintf(str, "gc_SetCallingNum(linedev=%ld, phone_num = %s) Success\n", pline->ldev, phone_num);
		printandlog(pline->index, MISC, NULL, str, 0);
	}
}

//*****************************************************************************
// Function: int ::_kbhit()
// Description: Detect keyboard hit
// Return: int
// Parameters: none
//*****************************************************************************
int _kbhit()
{
   static const int STDIN = 0;
   static bool initialized = false;

   if ( !initialized )
   {
      // Use termios to turn off line buffering
      struct termios term;
      tcgetattr(STDIN, &term);
      term.c_lflag &= ~ICANON;
      tcsetattr(STDIN, TCSANOW, &term);
      setbuf(stdin, NULL);
      initialized = true;
   }

   int bytesWaiting;
   ioctl(STDIN, FIONREAD, &bytesWaiting);
   return bytesWaiting;
}


/***************************************************************************
 *			NAME: void intr_hdlr()
 *DESCRIPTION: Handler called when one of the following signals is
 *					received: SIGHUP, SIGINT, SIGQUIT, SIGTERM.
 *					This function stops I/O activity on all channels and
 *					closes all the channels.
 *		  INPUT: None
 *		RETURNS: None.
  ***************************************************************************/
static void intr_hdlr(void)
{
	if (interrupted == NO) {
		printandlog(ALL_DEVICES, MISC, NULL, " *******Received User Interrupted Signal *******\n", 0);

		interrupted = YES;
	}
	/* will handle graceful shutdown outside this handler in main's forever loop */
}

/********************************************************************************
 *			NAME: load_config_file()
 *DESCRIPTION: This function reads the configuration file line
 *					by line. Ignores commented lines. For valid lines, 
 *					it fills the devname field of corresponding element 
 *					of port (global array of struct channel) as it is
 *					to be passed as an argument to gc_OpenEx(). It also
 *					fills the direction field in port[index]. 
 * 
 *		  INPUT: None
 *		RETURNS: None - exits if error
 *	  CAUTIONS: Assumes config file is in a "reasonable" format
 *					Not all fields are case insensitive
 *					The parsing in this procedure is simplified for demo purposes,
 *					i.e. error handling is weak to non-existent.
 *					Your application will need to implement parsing that is appropriate 
 *					for your application
 ********************************************************************************/
static void load_config_file(void)
{
	static	char	line[MAX_STRING_SIZE], direction[MAX_STRING_SIZE];
	static	char	netdevice[MAX_DEVNAME], protname[MAX_DEVNAME], mediaDeviceName[MAX_DEVNAME];
	static	char	devname[MAX_DEVNAME], destination_num[GC_ADDRSIZE];
	static	char	txcodec[MAX_CODECNAME], rxcodec[MAX_CODECNAME];
	static	char	txVAD[MAX_VADSTR], rxVAD[MAX_VADSTR];
	int ipmChannelNumber=0;
	int brd, ch;
   int      intercall_delay=0;
	int		txrate = 0;
	int		rxrate = 0;
	int		index, num_params, num_codec;
	int lineNumber=0;
	brd=ch=0;
	FILE		*cfg_fp;
	fpos_t	fpos;
	
	/* Open Configuration file in Read mode */
	cfg_fp = fopen(cfgfile, "r");
	
	if (cfg_fp == NULL) {
		printf("Unable to open %s for read\n", cfgfile);
		perror("Reason is: ");
		exit(1);
	}

	while (fgets(line, MAX_STRING_SIZE, cfg_fp)) {
		lineNumber++;
		/* strip trailing white space */
		index = strlen(line) - 1;
		while (index && (line[index] <= ' ')) {
			line[index] = '\0';
			index--;
		}

		/* ignore blank lines */
		if (index == 0) {
			continue;
		}

		/* if 1st character is not alpha numeric, then treat the line as a comment & ignore it */
		if ((line[0] < 'A') || (line[0] > 'z')) {	/* simplifed non-alpha check */
			continue; /* Ignore the line and read next */
		}
		num_devices++; /* Increment Global Variable */

		if(num_devices > MAXCHAN)
		{
			printf(" [ERROR]:		 Number of devices in cfg file exceeded maximum. Increase MAXCHAN and recompile\n");
			num_devices--;
			fclose (cfg_fp);
			exitdemo(1);
		}

		/* Parse DTI Device, Protocol Name, Direction and Voice Device Name */ 
		num_params = sscanf(line, "%s%s%s%s%s%d%s%d%s%s%d%s",
			netdevice, 
			protname, 
			direction, 
			mediaDeviceName, 
			destination_num,
         &intercall_delay,
			/* following coder parameters apply to VoIP devices only: */
			txcodec,
			&txrate,
			txVAD,
			rxcodec,
			&rxrate,
			rxVAD);

#ifdef USE_VIDEO
			if( (strcasecmp(protname,"ISDN")!=0) && (strcasecmp(protname,"SS7")!=0) && (strcasecmp(protname,"SIP")!=0) && (strcasecmp(protname,"BICC")!=0)) {
				printf("This application do not support Video Call using Protocol [%s]. Check line number -> %d\n",protname,lineNumber);
				fclose (cfg_fp);
				exitdemo(1);
			}
#endif

			if ((netdevice[0] == '\0') || (protname[0] == '\0') || (direction == '\0')) {
			/* Invalid entry in the configuration file */
			fclose (cfg_fp);
			printf("Invalid entry in %s\n", cfgfile);
			printf("Line is: %s\n", line);
			exitdemo(1);
		}
		
		/* Fill devname to be passed as arg in gc_OpenEx() 
		in port array (global) of struct channel */

		if (strcmp(netdevice, "NONE") != 0) {
		 	if(strcasecmp(protname,"BICC")==0) { 
					sprintf(devname, ":N_%s:P_SS7",netdevice);
			} else {
					sprintf(devname, ":N_%s:P_%s", netdevice, protname);
			}
		} else {
			sprintf(devname, ":P_%s", protname);
		}
#ifndef USE_VIDEO
		/*Fill :V_ only if voice device name is not specified as NONE in cfg file */
		if (tolower(mediaDeviceName[0]) != 'n') {
			/* IP uses :M instead of :V */
			if (mediaDeviceName[0] == 'i') {
				strcat(devname, ":M_");
			}
			strcat(devname, mediaDeviceName);
		}
#endif

		/* Fill Direction in port array (global) of struct channel */
		if (tolower(direction[0]) == 'i') {
			port[num_devices - 1].direction = DIR_IN;
		} else {
			port[num_devices - 1].direction = DIR_OUT;
		}

		/* Fill device index in port array (global) of struct channel */
		port[num_devices - 1].index = num_devices - 1;

		/* Copying all the local variable values to Global Variables */
		strcpy(port[num_devices - 1].devname, devname);
		strcpy(port[num_devices - 1].netdevice, netdevice);
		if ( (!strcasecmp(protname, "H323")) || (!strcasecmp(protname, "SIP")) || (!strcasecmp(protname, "BICC")) ) {
			if(strcasecmp(protname, "BICC")==0) {
				num_bicc_devices++;
			} else {
				num_sip_devices++;
			}
			strcpy(port[num_devices - 1].deviceInfo[IPM].device_name, mediaDeviceName);

			sscanf(mediaDeviceName,"ipmB1C%d",&ipmChannelNumber);
			if(ipmChannelNumber>max_ip_channel_to_start)
			{	
				max_ip_channel_to_start=ipmChannelNumber;
			}

			if(num_devices>=2) {
				sscanf(port[num_devices - 2].dxDeviceName,"dxxxB%dC%d",&brd,&ch);
				if(ch==4) {
					brd++;
					ch=1;
				} else {
					ch+=1;
				}
			} else {
				brd=ch=1;
			}
			sprintf(port[num_devices - 1].dxDeviceName,"dxxxB%dC%d",brd,ch);
		 	printf("SIP:: Application will use %s voice device for %s netowrk channel\n",port[num_devices - 1].dxDeviceName,port[num_devices - 1].netdevice);
		} else {
			strcpy(port[num_devices - 1].dxDeviceName, mediaDeviceName);
		}
		strcpy(port[num_devices - 1].protname, protname);
		strcpy(port[num_devices - 1].destination_num, destination_num);
      port[num_devices - 1].intercall_delay = intercall_delay;

		/* For VoIP devices (H.323 or SIP) copy the codec parameters */ 
		if ((!strcasecmp(protname, "H323")) || (!strcasecmp(protname, "SIP"))) {
			if (num_params == 12) {
				strcpy(port[num_devices - 1].codec[0].txcodec, txcodec);
				port[num_devices - 1].codec[0].txrate = txrate;
				strcpy(port[num_devices - 1].codec[0].txVAD, txVAD);
				strcpy(port[num_devices - 1].codec[0].rxcodec, rxcodec);
				port[num_devices - 1].codec[0].rxrate = rxrate;
				strcpy(port[num_devices - 1].codec[0].rxVAD, rxVAD);

				/* Get current position in config file for possible rewinding later */
				fgetpos(cfg_fp, &fpos);
				num_codec = 1;

				while (fgets(line, MAX_STRING_SIZE, cfg_fp)) {
					/* Strip trailing white space */
					index = strlen(line) - 1;
					while (index && (line[index] <= ' ')) {
						line[index] = '\0';
						index--;
					}

					/* ignore comments or blank lines */
					if ((line[0] == '#') || (index == 0)) {
						continue; /* Ignore the line and read next */
					}

					num_params = sscanf(line, "%s%d%s%s%d%s",
						txcodec,
						&txrate,
						txVAD,
						rxcodec,
						&rxrate,
						rxVAD);

					if (num_params !=6) {
						break;
					}

					num_codec++;
					if  (num_codec > MAXCODECS) {
							printf("More than 10 codecs specified for device - please modify gc_basic_call_model_voice_video_3g_cnf_nbup.cfg to correct\n");
							exitdemo(1);
					}

					strcpy(port[num_devices - 1].codec[num_codec - 1].txcodec, txcodec);
					port[num_devices - 1].codec[num_codec - 1].txrate = txrate;
					strcpy(port[num_devices - 1].codec[num_codec - 1].txVAD, txVAD);
					strcpy(port[num_devices - 1].codec[num_codec - 1].rxcodec, rxcodec);
					port[num_devices - 1].codec[num_codec - 1].rxrate = rxrate;
					strcpy(port[num_devices - 1].codec[num_codec - 1].rxVAD, rxVAD);

					/* Get current position in config file for possible rewinding later */
					fgetpos(cfg_fp, &fpos);
				}  /* end while there are multiple codecs for device to parse */

				/* Reset the previous file position for other line device entries to process */
				fsetpos(cfg_fp, &fpos);

			} else {  /* Not all codec parameters were in config file */
				fclose (cfg_fp);
				printf("Invalid or missing codec parameter entry in %s\n", cfgfile);
				printf("Line is: %s\n", line);
				exitdemo(1);
			}

			/* Store number of codecs specified for device */
			port[num_devices - 1].num_codecs = num_codec;

		}  /* end if VoIP device */
	
	} /* End of read all lines loop */
	
	fclose (cfg_fp);
} /* End of Function load_config_file() */


/********************************************************************************
 *		 NAME: printandlog ()
 * DESCRIPTION: prints on the screen and to the log file. 
 *				 Put time stamp to the log file. If the
 *				 message to be logged is specific to a device,
 *				 it is logged to the device specific log file.
 * 
 *		INPUT: index is the device index
 *				 log_type valid values are ERR, MISC, EVENT, GC_APICALL, GC_APIERR and GC_RESULT_INFO
 *				 metaeventp = pointer to metaevent data structure - only used for GC_RESULT_INFO
 *				 msg_str message string to be printed
 *				 callindex is the CRN index - used for retrieving callstate.
 *	 RETURNS: NA
 ********************************************************************************/
static void printandlog(int index, int log_type, METAEVENT *metaeventp, char *msg_str, int callindex)
{	
	struct tm			*Time;
	time_t				Timet;
	int					i;

#ifdef _WIN32	/* Defined in Windows */  
	struct _timeb	  Timeb;										/* used to get milliseconds */
#else			/* Defined in Unix	*/
	struct timeb		Timeb;	  
#endif
	 
	char					TimeStr[MAX_STRING_SIZE];			/* will hold the time string when built */
	char					buffer[MAX_STRING_SIZE];			/* will have everything but the time */
	char					timestamp[MAX_STRING_SIZE];		/* holds the formatted time */
	char					gc_apierr[MAX_STRING_SIZE];		/* iff have GC API error - most of time will be NULL */
	char					gc_result_info[MAX_STRING_SIZE]; /* iff have GC result info request - most of time will be null */
		 
	/* initialize temporary buffers */
	buffer[0] = '\0';
	gc_apierr[0] = gc_result_info[0] = timestamp[0] = '\0';
	 
	/* Put time stamp	 - everything but the milliseconds */
	TimeStr[0] = '\0';
	time(&Timet);
	Time = localtime(&Timet);

#ifdef _WIN32	/* Defined in Windows */  
	_ftime(&Timeb);												/* needed to get the milliseconds */
#else			/* Defined in Unix	 */  
	ftime(&Timeb); 
#endif

	if (log_type == MISC_WITH_NL) {
		strcpy(timestamp, "\n");
	} 

	sprintf(TimeStr, "%02d/%02d/%02d %02d:%02d:%02d.%03d", Time->tm_mon+1, Time->tm_mday, (1900+Time->tm_year), Time->tm_hour, Time->tm_min, Time->tm_sec, Timeb.millitm);
	strcat(timestamp, TimeStr);

	// add index into buffer
	sprintf(buffer,"[%14s] [%04d] ",devEvtTypeStr[log_type],index);

	switch(log_type)
	{
		case GC_APIERR:
				format_error_info(gc_apierr);
				break;

		case GC_RESULT_INFO:
				format_result_info(index, metaeventp, gc_result_info);
				break;

		case STATE:
				append_call_state_name(buffer, index, callindex);
				break;
		default:
			break;
	} /* End of Switch */
	 
	strcat(buffer, msg_str);
	 

	/* now do the actual logging */
	/* 1st to screen */
	printf("%s	%s\n",timestamp,buffer);

	/* print GC error or GC result information if required */
	if (gc_apierr[0]) {
		printf("%s\n", gc_apierr);
	}
	if (gc_result_info[0]) {
		printf("%s\n", gc_result_info);
	}

	/* then to log file(s) */
	if (index == ALL_DEVICES) {
		for (i = 0; i < num_devices; i++) {
			if (port[i].log_fp) {
				fprintf(port[i].log_fp, "%s ", timestamp);
				fprintf(port[i].log_fp, "%s\n", buffer);
            fflush(port[i].log_fp);
			}
		}
	} else if (index==BOARD_DEVICE) {
		fprintf(boardLogFd,"%s",timestamp);
		fprintf(boardLogFd,"%s\n",buffer);
		fflush(boardLogFd);
	} else if (port[index].log_fp) {
		fprintf(port[index].log_fp, "%s", timestamp);
		fprintf(port[index].log_fp, "%s\n", buffer);

		/* print GC error or GC result information if required */
		if (gc_apierr[0]) {
			fprintf(port[index].log_fp, "%s\n", gc_apierr);
		}
		if (gc_result_info[0]) {
			fprintf(port[index].log_fp, "%s\n", gc_result_info);
		}
      fflush(port[index].log_fp);
	}
}

/********************************************************************************
 *			NAME: format_error_info()
 *DESCRIPTION: formats error information using gc_ErrorInfo()
 *		  INPUT: buffer - buffer to format into
 *		RETURNS: None
 *	  CAUTIONS: No integrity checking done on dest_buffer overflow
 ********************************************************************************/
static void format_error_info(char *dest_buffer)
{
	GC_INFO			  t_info;	  /* Structure which stores information about GC related errors */ 

	gc_ErrorInfo(&t_info);
			 
	strcpy(dest_buffer, "					 GC ERROR: ");
	strcat(dest_buffer, t_info.gcMsg);
	strcat(dest_buffer, "\n");


	strcat(dest_buffer, "					 CC Name: ");
	strcat(dest_buffer, t_info.ccLibName);
	strcat(dest_buffer, "\n");


	strcat(dest_buffer, "					 CC ERROR: ");
	strcat(dest_buffer, t_info.ccMsg);
	strcat(dest_buffer, "\n");

}

/********************************************************************************
 *			NAME: format_result_info()
 *DESCRIPTION: formats result information using gc_ResultInfo()
 *		  INPUT: index - index into port data structure
 *             metaeventp - pointer to metaevent structure
 *					buffer - buffer to format into
 *		RETURNS: None
 *	  CAUTIONS: No integrity checking done on dest_buffer overflow
 ********************************************************************************/
static void format_result_info(int index, METAEVENT *metaeventp, char *dest_buffer)
{
	char					str[MAX_STRING_SIZE];
	int					rc;
	GC_INFO				t_info;	  /* Structure which stores information about GC related errors */ 

	rc = gc_ResultInfo(metaeventp, &t_info);
	if (rc < 0) {
		sprintf(str, "gc_ResultInfo() call failed");
		printandlog(index, GC_APIERR, NULL, str, 0);
		exitdemo(1);
	}
			 
	strcpy(dest_buffer, "					 GC Result: ");
	strcat(dest_buffer, t_info.gcMsg);
	strcat(dest_buffer, "\n");


	strcat(dest_buffer, "					 CC Name: ");
	strcat(dest_buffer, t_info.ccLibName);
	strcat(dest_buffer, "\n");


	strcat(dest_buffer, "					 CC Result: ");
	strcat(dest_buffer, t_info.ccMsg);
	strcat(dest_buffer, "\n");
}

/********************************************************************************
 *			NAME: append_call_state_name ()
 *DESCRIPTION: Appends the call state to a given buffer for a specified channel index
 *		  INPUT: buffer - buffer to append to
 *					index - index into port structure (from 0)
 *		RETURNS: None
 *	  CAUTIONS: No integrity checking done on buffer overflow
 ********************************************************************************/
static void append_call_state_name(char *buffer, int index, int callindex)
{
	struct channel		*pline;
	char					str[100];

	if ((callindex < 0) || (callindex > MAX_CALLS)) {
		sprintf(str, "Invalid callindex(%d)", callindex);
		strcat(buffer, str);
		return;
	}

	pline = &port[index];
	switch(pline->call[callindex].call_state)
	{ 
		case GCST_NULL:
			strcat(buffer, "GCST_NULL");
			break;

		case GCST_IDLE:
			strcat(buffer, "GCST_IDLE");
			break;

		case GCST_OFFERED:
			strcat(buffer, "GCST_OFFERED");
			break;

		case GCST_ALERTING:
			strcat(buffer, "GCST_ALERTING");
			break;

        case GCST_PROCEEDING:
			strcat(buffer, "GCST_PROCEEDING");
			break;

		case GCST_ACCEPTED:
			strcat(buffer, "GCST_ACCEPTED");
			break;
				
		case GCST_CONNECTED:
			strcat(buffer, "GCST_CONNECTED");
			break;

		case GCST_DISCONNECTED:
			strcat(buffer, "GCST_DISCONNECTED");
			break;

		default:
			sprintf(str, "UNKNOWN GC CALL STATE (%d)", pline->call[callindex].call_state);
			strcat(buffer, str);
			break;
	}
}

/*******************************************************************************
 *			NAME: existCRN()
 *DESCRIPTION: Search the CRN array for a valid CRN.
 *		  INPUT: CRN crn - CRN associated with the call.
 *						Note: if crn == -1 then this means return the index of the 1st valid
 *						crn in the call array.
 *		  			struct channel *pline - pointer to data structure for the channel
 *		RETURNS: if CRN exists, then return callindex; otherwise return OUT_OF_RANGE.		
 ******************************************************************************/
static int existCRN(CRN crn, struct channel *pline)
{
	int callindex;

	if (crn == -1) {
		/* For drop_outbound_calls_if_required() case */
		/* The function does not have access to metaevent structure. */
		for (callindex = 0; callindex < MAX_CALLS; callindex++) {
			/* if crn is found, then break */
			if (pline->call[callindex].crn > 0) {
				break;
			}
		}
	} else {
		/* All other cases. */
		for (callindex = 0; callindex < MAX_CALLS; callindex++) {
			if (pline->call[callindex].crn == crn) {
				/* crn is found, break out of for loop and continue. */
				break;
			}
		}
	}

	/* If the callindex < MAX_CALLS, then return CRN. */
	/* Otherwise return (-1) indicating CRN was not found. */
	if (callindex < MAX_CALLS) {
		/* CRN found */
		return (callindex);
	}
	else {
		/* CRN not found */
		return (-1);
	}
}

/********************************************************************************
 *			NAME: initiate_exit()
 * DESCRIPTION: Initiate termination of active calls.
 * 			This will be done via gc_ResetLineDev()
 * 			which will clean up active calls
 *		  INPUT: none
 *		RETURNS: none
 *
 *******************************************************************************/
static void initiate_exit()
{
	int				index;
	char				str[MAX_STRING_SIZE]; 
	struct channel	*pline;
	static int		exit_active = NO;

	if (exit_active == YES) {
		return;
	}
	printandlog(ALL_DEVICES, MISC, NULL, "********** Exit initiated ***********\n", 0);

	exit_active = YES;
	reset_linedevs_left_to_complete=-1;
	for(index = 0; index < num_devices; index++) {
		pline = &port[index];
		if (pline->ldev) {
			if (pline->resetlinedev_active == NO) {
				if (gc_ResetLineDev(port[index].ldev, EV_ASYNC) == GC_SUCCESS) {
					reset_linedevs_left_to_complete++;
					pline->resetlinedev_active = YES;
					sprintf(str, "gc_ResetLineDev(linedev=%ld, mode=EV_ASYNC) Success",
						 port[index].ldev);
					printandlog(index, GC_APICALL, NULL, str, 0);
				} else {
					sprintf(str, "gc_ResetLineDev(linedev=%ld, mode=EV_ASYNC) failure ignored while exiting (reload boards after exit)",
						port[index].ldev);
					printandlog(index, GC_APIERR, NULL, str, 0);
				}
			} else {
				reset_linedevs_left_to_complete++;			/* reset linedev already active */
			}															/* just wait for it to complete */
		}
	}
	reset_linedevs_left_to_complete++;		// SKS since intial value is -1
	if (reset_linedevs_left_to_complete) {
		sprintf(str, "Please wait while %d gc_ResetLineDev()s are being done...",
			reset_linedevs_left_to_complete);
		printandlog(pline->index, MISC, NULL, str, 0);
	}
}

/********************************************************************************
 *			NAME: exitdemo()
 *DESCRIPTION: Closes all the logfiles,
 *					Closes all the devices
 *					Disables event handler
 *					Stops GlobalCall
 *					And Exits
 *		  INPUT: exitlevel 0 success, 1 some error 
 *					Exit from program with the exitlevel as specified by exitlevel
 *		RETURNS: exits		
 ********************************************************************************/
static void exitdemo(int exitlevel)
{
	int					index;
	char					str[MAX_STRING_SIZE]; 
	struct channel		*pline;
	time_t				finish_time;
	double				duration, min_time = 0;
	static int			recursive = 0;					/* called recursively. e.g. Multiple ^c. If yes, skip clean-up logic and exit */

	printandlog(ALL_DEVICES, MISC_WITH_NL, NULL, "********* Program Exiting *********", 0);
	if (recursive == 0) {
		recursive++;
		for (index = 0; index < num_devices; index++) {
			pline = &port[index];
		 
			/* The total calls run on Inbound & Outbound side */
			if (pline->numb_calls>0) {
				if (pline->direction == DIR_IN) {
					sprintf(str, "The total no of Inbound calls on this device is: %d", pline->numb_calls);
					printandlog(pline->index, MISC, NULL, str, 0);
				} else { 
					sprintf(str, "The total no of Outbound calls on this device is: %d", pline->numb_calls);
					printandlog(pline->index, MISC, NULL, str, 0);
				}
			}
			if(pline->numb_calls>0) {
				/* Note down the finish time for the test */
				time(&finish_time);
				/* Calculate the duration taken by the test */
				duration = difftime(finish_time,	 start_time);
				min_time = duration / 60;
				sprintf(str, "The total duration taken by the test is: %5.2f minutes", min_time);
				printandlog(pline->index, MISC, NULL, str, 0);
			}
		}
	}

	/* code that can never be called recursively */

	/* close all log files */
	for (index = 0; index < num_devices; index++) {
		pline = &port[index];
		if (pline->log_fp != NULL) {
			fclose(pline->log_fp);
			pline->log_fp = NULL;			
		}
	}
	if(boardLogFd!=NULL) {
		fclose(boardLogFd);
		boardLogFd=NULL;
	}
	if (gc_started == YES) {
		for(index = 0; index < num_devices; index++) {
			pline = &port[index];

			if (pline->ldev) {
				/* nothing can do about errors, so ignore them */			
				gc_Close(pline->ldev);
				pline->ldev = 0;
			}
			/* if GC makecall block being used, then free it */
			if (port[index].makecallblk.gclib) {
				/* free the parm block if used using the utility functions */
				/* CAUTION: this must be done before freeing the GC makecall block */
				if (port[index].makecallblk.gclib->ext_datap) {
					gc_util_delete_parm_blk(port[index].makecallblk.gclib->ext_datap);
				}

				/* now free the GC makecall block */
				free(port[index].makecallblk.gclib);
				port[index].makecallblk.gclib = NULL;
			}
		}
		gc_Stop();
	}
	exit(exitlevel);
} /* End of function exitdemo() */
/*****<< End of File >>*****/
